﻿using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using MVR.FileManagementSecure;
using System.Text.RegularExpressions;

namespace JayJayWon
{
    public class UIAssist : MVRScript {
        const string pluginName = "UIAssist"; const string pluginAuthor = "JayJayWon"; const string pluginVersion = "1.3.0"; const float pluginVersionFloat = 1.03f;

        public static float _gazeFOVNarrow = 10f;
        public static float _gazeFOVWide = 60f;
        public string _presetRef = "1";
        public static List<string> VAMSaveRestoreActions = new List<string>() { "SaveToStore1", "SaveToStore2", "SaveToStore3", "RestoreAllFromStore1", "RestoreAllFromStore2", "RestoreAllFromStore3", "RestorePhysicsFromStore1", "RestorePhysicsFromStore2", "RestorePhysicsFromStore3", "RestoreAppearanceFromStore1", "RestoreAppearanceFromStore2", "RestoreAppearanceFromStore3", "RestoreAllFromDefaults", "RestorePhysicalFromDefaults", "RestoreAppearanceFromDefaults" };
        public string GetLocalPluginPath()
        {
            string pluginId = storeId.Split('_')[0];
            string pathToScriptFile = manager.GetJSON(true, true)["plugins"][pluginId].Value;
            string pathToScriptFolder = pathToScriptFile.Substring(0, pathToScriptFile.LastIndexOfAny(new char[] { '/', '\\' }) + 1);
            pathToScriptFolder = pathToScriptFolder.Replace('/', '\\');
            if (pathToScriptFolder.Contains(":")) pathToScriptFolder = pathToScriptFolder.Substring(pathToScriptFolder.IndexOf(':') + 1);
            char[] charsToTrim = { '/', '\\' };
            pathToScriptFolder = pathToScriptFolder.Trim(charsToTrim);
            bool exists = FileManagerSecure.DirectoryExists(pathToScriptFolder);
            if (!exists) pathToScriptFolder = "Custom\\Scripts";
            return pathToScriptFolder;
        }
        bool DefaultUIAPFileExists() {
            string[] fileNamesArray = SuperController.singleton.GetFilesAtPath(GetLocalPluginPath());
            List<string> fileNameList = new List<string>();
            foreach (string filePath in fileNamesArray) {
                string fileName = filePath.Substring(filePath.LastIndexOfAny(new char[] { '/', '\\' }) + 1);
                fileNameList.Add(fileName);
            }

            if (fileNameList.Contains("default.uiap")) { return (true); }
            return (false);
        }

        public UIDynamicTextField CreateTextInput(JSONStorableString jss, bool rightSide = false) {
            UIDynamicTextField textfield = CreateTextField(jss, rightSide); textfield.height = 1f; textfield.backgroundColor = Color.white; var input = textfield.gameObject.AddComponent<InputField>(); input.textComponent = textfield.UItext; input.lineType = InputField.LineType.MultiLineNewline; input.textComponent.resizeTextMaxSize = 30; input.textComponent.resizeTextForBestFit = true; jss.inputField = input; return (textfield);
        }
        public UIDynamicPopup CreateSetupPopup(JSONStorableStringChooser stringChooserJSON, bool rightUI = false, float panelHeight = 400f, UIPopup.OnOpenPopup callbackMethod = null, string label = null, bool scrollable = true) {
            UIDynamicPopup dynamicPopup;
            if (scrollable) dynamicPopup = CreateScrollablePopup(stringChooserJSON, rightUI);
            else dynamicPopup = CreatePopup(stringChooserJSON, rightUI);
            dynamicPopup.popupPanelHeight = panelHeight;
           
            if (callbackMethod != null) { dynamicPopup.popup.onOpenPopupHandlers += callbackMethod; }
            if (label != null) { dynamicPopup.label = label; }
            return (dynamicPopup);
        }
        //		Transform trans ;
        public override void Init() {
            try {
                _isActivePluginJSON = new JSONStorableBool("activeUIAssist", false);
                RegisterBool(_isActivePluginJSON);
                _isActivePluginJSON.isStorable = false;
                _isActivePluginJSON.isRestorable = false;

                _isActivePlugin = false;
                if (containingAtom.name == "CoreControl") {
                    if (GetActiveUIAssistCount() < 2) {
                        List<string> handChoices = new List<string>();
                        handChoices.Add("Left Hand");
                        handChoices.Add("Right Hand");
                        handChoices.Add("Disabled");

                        _vrHandControlJSON = new JSONStorableStringChooser("handControl", handChoices, "Left Hand", "VR Watch Wrist", SwitchVRHand);
                        RegisterStringChooser(_vrHandControlJSON);

                        _vrPinModeJSON = new JSONStorableStringChooser("pinMode", null, "Static", "VR Pin Mode", SwitchVRPinMode);
                        _gazeMenuSelectModeJSON = new JSONStorableStringChooser("gazeMenuSelectMode", null, "Wrist UI Only", "Gaze UI Control");
                        _toolboxJSON = new JSONStorableStringChooser("toolBoxMode", null, "Select...", "Toolbox", SwitchToolBox);

                        _vrHUDHorizontalOffsetJSON = new JSONStorableFloat("vrHUDHorizontalOffset", 0.5f, 0.3f, 0.8f);
                        _vrHUDVerticalOffsetJSON = new JSONStorableFloat("vrHUDVerticalOffset", 0.5f, 0.2f, 0.8f);
                        _vrHUDDepthOffsetJSON = new JSONStorableFloat("vrHUDDepth", 0.6f, 0.1f, 2f);

                        _isActivePluginJSON.val = true;
                        _isActivePlugin = true;
                        _pluginVersionJSON = new JSONStorableString("PluginVersion", pluginVersion);
                        RegisterString(_pluginVersionJSON);
                        _pluginVersionJSON.isStorable = true;
                        _pluginVersionJSON.isRestorable = false;

                        _pluginVersionFloatJSON = new JSONStorableFloat("PluginVersionID", pluginVersionFloat, 1f, 10f);
                        RegisterFloat(_pluginVersionFloatJSON);
                        _pluginVersionFloatJSON.isStorable = true;
                        _pluginVersionFloatJSON.isRestorable = false;

                        _isPatreonOnlyVersionJSON = new JSONStorableBool("PatreonOnlyVersion", Constants._patreonContentEnabled);
                        RegisterBool(_isPatreonOnlyVersionJSON);
                        _isPatreonOnlyVersionJSON.isStorable = true;
                        _isPatreonOnlyVersionJSON.isRestorable = false;

                        _isPrimaryUIAJSON = new JSONStorableBool("PrimaryUIA", false);
                        RegisterBool(_isPrimaryUIAJSON);
                        _isPrimaryUIAJSON.isStorable = false;
                        _isPrimaryUIAJSON.isRestorable = false;

                        _uiMainCanvas = new WristUI(this);
                        _uiButtonCanvas = new WristUI(this);
                        _defaultButtonSkin = new ButtonSkin(2, this);
                        _defaultButtonSkinDisplay = new ButtonSkinDisplay(_defaultButtonSkin, this, true);

                        _isLoaded = false;
                        _mainButton = _uiMainCanvas.CreateUIButtonInCanvas(System.DateTime.Now.ToString("HH:mm"), 0f, 0f, 0f, 35f, 20f);
                        _mainButton.button.onClick.AddListener(() => { ToggleUIActivateButton(); });
                        _leftButton = _uiMainCanvas.CreateUIButtonInCanvas("<<", -0.05f, 0f, 0f, 15f, 15f);
                        _leftButton.gameObject.SetActive(false);
                        _leftButton.button.onClick.AddListener(() => { ScrollLeftButton(); });
                        _rightButton = _uiMainCanvas.CreateUIButtonInCanvas(">>", 0.05f, 0f, 0f, 15f, 15f);
                        _rightButton.gameObject.SetActive(false);
                        _rightButton.button.onClick.AddListener(() => { ScrollRightButton(); });

                        _mainPinButton = _uiMainCanvas.CreateUIButtonInCanvas("X", -0.1f, 0f, 0f, 15f, 15f);
                        _mainPinButton.gameObject.SetActive(false);
                        _mainPinButton.button.onClick.AddListener(() => { TogglePin(); });

                        _uiButtonList = new List<UIDynamicButton>();
                        _buttonGridSelectorJSON = new JSONStorableStringChooser("gridSelection", GetGridChoices(), "1", "", SwitchDisplayGrid);

                        _displayGazeSelectionsJSON = new JSONStorableBool("displayGazeSelections", true, SwitchDisplayGazeSelections);
                        _displayScreenLabelsJSON = new JSONStorableBool("displayScreenLabels", false, SwitchDisplayScreenLabels);
                        _gazeActivateVRScreenJSON = new JSONStorableBool("gazeActivateVRScreen", true);

                        string pluginDetails = "";
                        if (Constants._patreonContentEnabled) { pluginDetails = pluginName + " v" + pluginVersion + "\nby " + pluginAuthor + "\nPatron version"; }
                        else { pluginDetails = pluginDetails = pluginName + " v" + pluginVersion + "\nby " + pluginAuthor + "\nFree version"; }

                        _pluginDetailsJSON = new JSONStorableString("PluginDetails", pluginDetails);

                        _toolbox = new ToolBox(this);

                        _atomRenameHandle = new SuperController.OnAtomUIDRename(this.AtomNameUpdate);
                        SuperController.singleton.onAtomUIDRenameHandlers += _atomRenameHandle;

                        List<string> test = new List<string>();
                        test.Add("Temp");
                        PatreonFeatures.GridLayoutFeature(test);
                        if (test != null && test.Count == 7) _secondValidation = true;

                        //					trans = SuperController.singleton.leapHandRight.Find("R_Arm_Cut").Find("R_Wrist").Find("R_Finger_Index_A").Find("R_Finger_Index_B").Find("R_Finger_Index_C");
                    }
                    else SuperController.LogMessage("A maximum of two UIAssist plugins can be active");
                }
                else SuperController.LogMessage("UIAssist can only be loaded as a Session Plugin");
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        protected void AtomNameUpdate(string oldName, string newName) {
            try {
                if (_lastViewedAtom == oldName) _lastViewedAtom = newName;
                if (_lastViewedPerson == oldName) _lastViewedPerson = newName;
                if (_lastViewedFemale == oldName) _lastViewedFemale = newName;
                if (_lastViewedMale == oldName) _lastViewedMale = newName;

                foreach (ButtonGrid buttonGrid in _buttonGridsList) {
                    foreach (Button button in buttonGrid._buttonList) {

                        foreach (string key in button._atomButtonStates.Keys.ToList()) {

                            if (key == oldName) {
                                button._atomButtonStates[newName] = button._atomButtonStates[key];
                                button._atomButtonStates.Remove(key);
                            }
                        }
                        if (button._targetCategory == "Specific Atom" && button._targetName == oldName) button._targetName = newName;
                    }
                }
                RefreshRightUIDisplay();
                if (_atomActiveClothingDictionary.ContainsKey(oldName))
                {
                    _atomActiveClothingDictionary.Add(newName, _atomActiveClothingDictionary[oldName]);
                    _atomActiveClothingDictionary.Remove(oldName);
                }
                _toolbox._heelAdjustTool.AtomNameUpdate(oldName, newName);


            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        protected void VRHandSelectionPopupOpen()
        {
            CloseAllLeftUIPopups(_vrHandControlPopup);
        }
        protected void RefreshPinModeChoices()
        {
            CloseAllLeftUIPopups(_vrPinModePopup);
            List<string> pinModeChoices = new List<string>();
            if (_vrHandControlJSON.val != "Disabled") pinModeChoices.Add("Static");
            pinModeChoices.Add("HUD (Left)");
            pinModeChoices.Add("HUD (Right)");
            pinModeChoices.Add("HUD (Custom)");
            if (_vrHandControlJSON.val != "Disabled") pinModeChoices.Add("Disabled");
            _vrPinModeJSON.choices = pinModeChoices;
        }
        protected void RefreshGazeControlModeChoices()
        {
            CloseAllLeftUIPopups(_gazeMenuSelectModePopup);
            List<string> gazeControlModeChoices = new List<string>();
            gazeControlModeChoices.Add("Wrist UI Only");
            if (_vrPinModeJSON.val == "Static") gazeControlModeChoices.Add("Wrist & Pinned UI");
            gazeControlModeChoices.Add("Disabled");
            _gazeMenuSelectModeJSON.choices = gazeControlModeChoices;
        }
        protected void ToolBoxPopupOpen()
        {
            CloseAllLeftUIPopups(_toolboxPopup);
            List<string> toolboxChoices = new List<string>();
            toolboxChoices.Add("Auto Heel Adjust");
            if (_isPrimaryUIAJSON.val)
            {
 //               toolboxChoices.Add("Auto Plugin Loader");
 //               toolboxChoices.Add("Auto Session Plugin Loader");
            }


            _toolboxJSON.choices = toolboxChoices;
        }
        protected void TogglePin() {
            try {
                _uiPinned = !_uiPinned;
                _autoDePinEnabled = false;

                if (_vrHandControlJSON.val == "Right Hand") {
                    _fixedPinPosition = SuperController.singleton.rightHand.position;
                    _fixedPinRotation = SuperController.singleton.rightHand.rotation;

                }
                else {
                    _fixedPinPosition = SuperController.singleton.leftHand.position;
                    _fixedPinRotation = SuperController.singleton.leftHand.rotation;
                }

            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            try {
                JSONClass jc = base.GetJSON(includePhysical, includeAppearance, true);
                if (jc == null) { jc = new JSONClass(); }
                
                jc = _toolbox._heelAdjustTool.GetJSON(jc);
                return jc;
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); return (null); }
        }
        public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true,  bool setMissingToDefault = true)
        {
            try
            {
                base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
                _toolbox._heelAdjustTool.RestoreJSON(jc);
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }

        }

        public JSONClass GetJSON(JSONClass jc = null) {
            if (jc == null) { jc = new JSONClass(); }

            jc["PluginVersion"] = pluginVersion;
            jc["PluginVersionID"].AsFloat = pluginVersionFloat;
            jc["PatreonOnlyVersion"].AsBool = Constants._patreonContentEnabled;

            jc["vrHandControl"] = _vrHandControlJSON.val;
            jc["vrPinMode"] = _vrPinModeJSON.val;
            jc["gazeMenuSelectMode"] = _gazeMenuSelectModeJSON.val;

            jc["isLoaded"].AsBool = _isLoaded;
            jc["displayGazeSelection"].AsBool = _displayGazeSelectionsJSON.val;
            jc["displayScreenLabels"].AsBool = _displayScreenLabelsJSON.val;
            jc["gazeActivateVRScreen"].AsBool = _gazeActivateVRScreenJSON.val;

            jc["vrHUDHorizontalOffset"].AsFloat = _vrHUDHorizontalOffsetJSON.val;
            jc["vrHUDVerticalOffset"].AsFloat = _vrHUDVerticalOffsetJSON.val;
            jc["vrHUDDepth"].AsFloat = _vrHUDDepthOffsetJSON.val;
           

            JSONArray gridArray = new JSONArray();
            _buttonGridsList.ForEach((grid) =>
            {
                gridArray.Add(grid.GetJSON());
            });
            
            jc["ButtonGrids"] = gridArray;
            
            jc["DefaultButtonSkin"] = _defaultButtonSkin.GetJSON();
            jc["ToolBox"] = _toolbox.GetJSON();
            
            return jc;
        }

        public JSONStorable GetOtherActiveUIAPlugin()
        {
            JSONStorable otherUIA =null;
            foreach (string storableName in containingAtom.GetStorableIDs())
            {
                if (storableName.StartsWith("plugin#"))
                {
                    string pluginType = storableName.Substring(storableName.IndexOf('_') + 1);
                    if (pluginType == "JayJayWon.UIAssist")
                    {
                        JSONStorable pluginStorable = containingAtom.GetStorableByID(storableName);
                        JSONStorableBool activeJSON = pluginStorable.GetBoolJSONParam("activeUIAssist");

                        if (activeJSON.val)
                        {
                            otherUIA = pluginStorable;
                            break;
                        }
                    }

                }
            }
            return (otherUIA);
        }
        private int GetActiveUIAssistCount() {
            int pluginCount = 0;
            foreach (string storableName in containingAtom.GetStorableIDs()) {
                if (storableName.StartsWith("plugin#")) {
                    string pluginType = storableName.Substring(storableName.IndexOf('_') + 1);
                    if (pluginType == "JayJayWon.UIAssist") {
                        JSONStorable pluginStorable = containingAtom.GetStorableByID(storableName);
                        JSONStorableBool activeJSON = pluginStorable.GetBoolJSONParam("activeUIAssist");

                        if (activeJSON.val) pluginCount++;
                    }

                }
            }
            return (pluginCount);
        }
        private void UpdatePrimaryUIA()
        {
            bool foundOtherActiveUIA = false;
            foreach (string storableName in containingAtom.GetStorableIDs())
            {
                if (storableName.StartsWith("plugin#") && storableName!=this.name)
                {
                    string pluginType = storableName.Substring(storableName.IndexOf('_') + 1);
                    if (pluginType == "JayJayWon.UIAssist")
                    {
                        JSONStorable pluginStorable = containingAtom.GetStorableByID(storableName);
                        JSONStorableBool activeJSON = pluginStorable.GetBoolJSONParam("activeUIAssist");
                        if (activeJSON.val)
                        {
                            JSONStorableBool primaryUIAJSON = pluginStorable.GetBoolJSONParam("PrimaryUIA");
                            _isPrimaryUIAJSON.val = !primaryUIAJSON.val;
                            foundOtherActiveUIA = true;
                            break;
                        }
                    }

                }
            }
            if (!foundOtherActiveUIA) _isPrimaryUIAJSON.val = true;
           
        }


        public void LoadJSON(JSONClass jc) {
            _isRestoring = true;
            if (jc["vrHandControl"] != null) _vrHandControlJSON.val = jc["vrHandControl"];
            if (jc["vrPinMode"] != null) _vrPinModeJSON.val = jc["vrPinMode"];
            if (jc["gazeMenuSelectMode"] != null) _gazeMenuSelectModeJSON.val = jc["gazeMenuSelectMode"];

            if (jc["isLoaded"] != null) _isLoaded = jc["isLoaded"].AsBool;
            if (jc["displayGazeSelection"] != null) _displayGazeSelectionsJSON.val = jc["displayGazeSelection"].AsBool;
            if (jc["displayScreenLabels"] != null) _displayScreenLabelsJSON.val = jc["displayScreenLabels"].AsBool;
            if (jc["gazeActivateVRScreen"] != null) _gazeActivateVRScreenJSON.val = jc["gazeActivateVRScreen"].AsBool;

            if (jc["vrHUDHorizontalOffset"] != null) _vrHUDHorizontalOffsetJSON.val = jc["vrHUDHorizontalOffset"].AsFloat;
            if (jc["vrHUDVerticalOffset"] != null) _vrHUDVerticalOffsetJSON.val = jc["vrHUDVerticalOffset"].AsFloat;
            if (jc["vrHUDDepth"] != null) _vrHUDDepthOffsetJSON.val = jc["vrHUDDepth"].AsFloat;

            if (jc["ButtonGrids"] != null)
            {
                JSONArray gridArrayJSON = jc["ButtonGrids"].AsArray;
                for (int i = 0; i < gridArrayJSON.Count; i++) {
                    JSONClass gridJSON = gridArrayJSON[i].AsObject;
                    ButtonGrid grid;
                    if (i >= _buttonGridsList.Count) {
                        grid = new ButtonGrid(0, 0, this);
                        _buttonGridsList.Add(grid);
                    }
                    else { grid = _buttonGridsList[i]; }
                    grid.LoadJSON(gridJSON);
                }
            }
            if (jc["DefaultButtonSkin"] != null) { _defaultButtonSkin.LoadJSON((JSONClass)jc["DefaultButtonSkin"]); }
            if (jc["ToolBox"] != null) _toolbox.LoadJSON((JSONClass)jc["ToolBox"]);
            _isRestoring = false;
        }

        private List<string> GetGridChoices() {
            List<string> gridChoices = new List<string>();
            for (int i = 1; i <= _buttonGridsList.Count; i++) {
                gridChoices.Add(i.ToString());
            }
            gridChoices.Add("Add Screen...");
            return (gridChoices);
        }

        private void SwitchDisplayGrid(string grid) {
            try {
                int gridRef = -1;
                if (grid == "Add Screen...") {
                    ButtonGrid newBG = new ButtonGrid(3, 3, this);
                    _buttonGridsList.Add(newBG);
                    gridRef = _buttonGridsList.Count;
                    _buttonGridSelectorJSON.val = gridRef.ToString();
                }
                else {
                    gridRef = int.Parse(grid) - 1;
                    _buttonGridDisplay._selectedButtonGrid = _buttonGridsList[gridRef];
                    if (_buttonGridDisplay._buttonSelectJSON.val == "1") { _buttonGridDisplay.SwitchSelectedButton("1"); }
                    else { _buttonGridDisplay._buttonSelectJSON.val = "1"; }
                    RefreshRightUIDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchDisplayGazeSelections(bool display) {
            try {
                if (!_isRestoring) { RefreshUIButtonGrid(); }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchDisplayScreenLabels(bool display) {
            try {
                if (!_isRestoring) {
                    RefreshUIButtonGrid();
                    RefreshRightUIDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void RefreshUIButtonGrid() {
            if (_uiActivated && _currentGridScreen >= 0) {
                DestroyUIButtons();
                CreateUIButtonGrid();
            }
        }
        private void CreateUIButtonGrid() {
            ButtonGrid buttonGrid = _buttonGridsList[_currentGridScreen];
            int columns = buttonGrid._gridColumns;
            int rows = buttonGrid._gridRows;
            string size = buttonGrid._buttonSize;
            int buttonNum = 0;
            int height = 0;
            int width = 0;
            switch (size) {
                case "Small":
                    height = 30;
                    width = 30;
                    break;
                case "Medium":
                    height = 45;
                    width = 45;
                    break;
                case "Large":
                    height = 60;
                    width = 60;
                    break;
            }

            _uiButtonList.Clear();
            float buttonHorizontalSpacing = 0f;
            float buttonVerticalSpacing = 0f;
            float horizontalStartOffset = 0f;
            float verticalStartOffset = 0.06f;

            Vector3[] rectCornersArray = new Vector3[4];
            List<string> test = new List<string>();
            test.Add("Temp");
            PatreonFeatures.GridLayoutFeature(test);
            if (!Constants._patreonContentEnabled || test == null) {
                rows = Mathf.Min(rows, 3);
                columns = Mathf.Min(columns, 3);
            }

            for (int row = 0; row < rows; row++) {
                for (int col = 0; col < columns; col++) {
                    buttonNum++;
                    UIDynamicButton gridButton = _uiButtonCanvas.CreateUIButtonInCanvas(buttonNum.ToString(), 0f, 0f, 0f, width, height);
                    int temp = buttonNum;
                    gridButton.button.onClick.AddListener(delegate { ButtonActivated(temp); });

                    RectTransform rt = gridButton.transform.GetComponent<RectTransform>();
                    rt.GetWorldCorners(rectCornersArray);
                    buttonHorizontalSpacing = Vector3.Distance(rectCornersArray[0], rectCornersArray[3]);
                    buttonVerticalSpacing = Vector3.Distance(rectCornersArray[0], rectCornersArray[1]);
                    horizontalStartOffset = buttonHorizontalSpacing * ((columns - 1) / 2f);

                    rt = gridButton.buttonText.gameObject.GetComponent<RectTransform>();
                    rt.anchoredPosition = new Vector2(0, 0);
                    rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width * 0.8f);
                    rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height * 0.8f);

                    gridButton.transform.Translate(-horizontalStartOffset + (col * buttonHorizontalSpacing), verticalStartOffset + ((row + 0.5f) * buttonVerticalSpacing), 0f, Space.Self);
                    _uiButtonList.Add(gridButton);
                }
            }

            _gazeStatusTextField = _uiButtonCanvas.CreateTextField(0, 0, 0, Mathf.Max(columns * width, 120), 40);

            _gazeStatusTextField.transform.Translate(-horizontalStartOffset - (buttonHorizontalSpacing / 2f), (buttonVerticalSpacing * buttonGrid._maxButtonRow) + (verticalStartOffset * 0), 0f, Space.Self);

            _gazeStatusTextField.UItext.fontSize = 6;
            _gazeStatusTextField.backgroundColor = new Color(173f / 255f, 216f / 255f, 230f / 255f, 1f);

            UpdateButtonTextures();
        }
        private void ButtonActivated(int buttonRef) {
            try {
                Button button = _buttonGridsList[_currentGridScreen]._buttonList[buttonRef - 1];
                button.PlayAction(_lastViewedAtom, _lastViewedPerson, _lastViewedFemale, _lastViewedFemale);
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void DestroyUIButtons() {
            foreach (UIDynamicButton uiButton in _uiButtonList) {
                if (uiButton != null) { GameObject.Destroy(uiButton.gameObject); }
            }
            if (_gazeStatusTextField != null) { GameObject.Destroy(_gazeStatusTextField.gameObject); }
        }
        private void ToggleUIActivateButton() {
            try {
                if (_uiActivated) {
                    DestroyUIButtons();
                    _uiActivated = false;
                    _isUIGazeDeactivated = false;

                }
                else {
                    if (_currentGridScreen < 0 || _buttonGridsList[_currentGridScreen]._activeButtonCount == 0) {
                        _currentGridScreen = -1;
                        for (int i = _buttonGridsList.Count - 1; i >= 0; i--) {
                            if (_buttonGridsList[i]._activeButtonCount > 0) { _currentGridScreen = i; }
                        }
                    }
                    if (_currentGridScreen >= 0) {
                        CreateUIButtonGrid();

                        _uiActivated = true;
                    }
                }
              
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private int GetNextLeftScreen() {
            int nextGridScreen = -1;
            for (int i = 0; i < _currentGridScreen; i++) {
                if (_buttonGridsList[i]._activeButtonCount > 0) { nextGridScreen = i; }
            }
            return (nextGridScreen);
        }
        private int GetNextRightScreen() {
            int nextGridScreen = -1;
            for (int i = _buttonGridsList.Count - 1; i > _currentGridScreen; i--) {
                if (_buttonGridsList[i]._activeButtonCount > 0) { nextGridScreen = i; }
            }
            return (nextGridScreen);
        }
        private void ScrollLeftButton() {
            try {
                _currentGridScreen = GetNextLeftScreen();
                DestroyUIButtons();
                if (_currentGridScreen > -1) {
                    CreateUIButtonGrid();
                    _uiActivated = true;
                }
                else { _uiActivated = false; }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void ScrollRightButton() {
            try {
                _currentGridScreen = GetNextRightScreen();
                DestroyUIButtons();
                if (_currentGridScreen > -1) {
                    CreateUIButtonGrid();
                    _uiActivated = true;
                }
                else { _uiActivated = false; }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void CreateDefaultButtonGrids() {
            ButtonGrid defaultBG = new ButtonGrid(3, 3, this);
            _buttonGridsList.Add(defaultBG);
            defaultBG._buttonList[0].UpdateButtonCategory("Clothing");
            defaultBG._buttonList[1].UpdateButtonCategory("Presets");
            defaultBG._buttonList[1].UpdateButtonType("Load Clothing Preset");
            defaultBG._buttonList[1].SetPresetFile("Custom/Atom/Person/Clothing/MVR/Preset_VaM_Seductress_Sim.vap");
            defaultBG._buttonList[1]._autoLabel = false;
            defaultBG._buttonList[1].UpdateAtom("Last viewed Female");
            defaultBG._buttonList[2].UpdateButtonCategory("Presets");
            defaultBG._buttonList[2].SetPresetFile("Custom/Atom/Person/Appearance/MVR/Preset_Ren_Tara.vap");
            defaultBG._buttonList[2]._autoLabel = false;
            defaultBG._buttonList[2].UpdateAtom("Last viewed Female");
            defaultBG.RecalcActiveButtonCount();
            defaultBG.RecalcGazeSelections();
        }
        public void RefreshLeftUIDisplay() {
            CloseAllLeftUIPopups(null);

            if (_loadProfileButton != null) { RemoveButton(_loadProfileButton); }
            if (_saveProfileButton != null) { RemoveButton(_saveProfileButton); }
            if (_setDefaultProfileButton != null) { RemoveButton(_setDefaultProfileButton); }
            if (_vrHandControlPopup != null) { RemovePopup(_vrHandControlPopup); }
            if (_vrPinModePopup != null) { RemovePopup(_vrPinModePopup); }
            if (_toolboxPopup != null) { RemovePopup(_toolboxPopup); }

            if (_gazeMenuSelectModePopup != null) { RemovePopup(_gazeMenuSelectModePopup); }
            if (_vrHUDHorizontalOffsetSlider != null) { RemoveSlider(_vrHUDHorizontalOffsetSlider); }
            if (_vrHUDVerticalOffsetSlider != null) { RemoveSlider(_vrHUDVerticalOffsetSlider); }
            if (_vrHUDDepthOffsetSlider != null) { RemoveSlider(_vrHUDDepthOffsetSlider); }


            if (_displayGazeSelectionsToggle != null) { RemoveToggle(_displayGazeSelectionsToggle); }
            if (_displayScreenLabelsToggle != null) { RemoveToggle(_displayScreenLabelsToggle); }
            if (_gazeActivateVRScreenToggle != null) { RemoveToggle(_gazeActivateVRScreenToggle); }
            if (_pluginVersionTextField != null) { RemoveTextField(_pluginVersionTextField); }
            if (_spacer1 != null) { RemoveSpacer(_spacer1); }
            if (_spacer2 != null) { RemoveSpacer(_spacer2); }
            if (_spacer2b != null) { RemoveSpacer(_spacer2b); }
            if (_spacer3 != null) { RemoveSpacer(_spacer3); }
            if (_spacer4 != null) { RemoveSpacer(_spacer4); }

            _isDisplayUpdating = true;
            RefreshPinModeChoices();
            if (!_vrPinModeJSON.choices.Contains(_vrPinModeJSON.val)) _vrPinModeJSON.val = _vrPinModeJSON.choices[0];
            RefreshGazeControlModeChoices();
            if (!_gazeMenuSelectModeJSON.choices.Contains(_gazeMenuSelectModeJSON.val)) _gazeMenuSelectModeJSON.val = _gazeMenuSelectModeJSON.choices[0];

            _isDisplayUpdating = false;

            _loadProfileButton = CreateButton("Load UI Profile ...");
            List<ShortCut> shortcuts = FileManagerSecure.GetShortCutsForDirectory("Custom\\Scripts\\JayJayWon", false, false, true, true);
            _loadProfileButton.button.onClick.AddListener(() => { SuperController.singleton.GetMediaPathDialog(LoadUIAPSelected, "uiap", GetLocalPluginPath(), false, true, false, null, false, shortcuts); });
            _saveProfileButton = CreateButton("Save UI Profile ...");
            _saveProfileButton.button.onClick.AddListener(() => {
                SuperController.singleton.GetMediaPathDialog(SaveUIAPSelected, "uiap", GetLocalPluginPath(), false, true, false);
                // Update the browser to be a Save browser
                uFileBrowser.FileBrowser browser = SuperController.singleton.mediaFileBrowserUI;
                browser.SetTextEntry(true);
                browser.fileEntryField.text = String.Format("{0}.{1}", ((int)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalSeconds).ToString(), "uiap");
                browser.ActivateFileNameField();

            });
            _setDefaultProfileButton = CreateButton("Set current Profile as Default");
            _setDefaultProfileButton.button.onClick.AddListener(() => { SaveUIAPSelected(GetLocalPluginPath() + "\\default.uiap"); });
            _spacer1 = CreateSpacer();
            _spacer1.height = 10;
            _vrHandControlPopup = CreateSetupPopup(_vrHandControlJSON, false, 200f, VRHandSelectionPopupOpen, null, false);
            if (_vrHandControlJSON.val != "Disabled")
            {
                _gazeActivateVRScreenToggle = CreateToggle(_gazeActivateVRScreenJSON);
                _gazeActivateVRScreenToggle.label = "Gaze Open/Close Screens";
            }
            _vrPinModePopup = CreateSetupPopup(_vrPinModeJSON, false, 200f, RefreshPinModeChoices, null, false);
            if (_vrPinModeJSON.val == "HUD (Custom)")
            {
                _vrHUDHorizontalOffsetSlider = CreateSlider(_vrHUDHorizontalOffsetJSON);
                _vrHUDHorizontalOffsetSlider.label = "HUD Horizontal position";
                _vrHUDVerticalOffsetSlider = CreateSlider(_vrHUDVerticalOffsetJSON);
                _vrHUDVerticalOffsetSlider.label = "HUD Vertical position";
                _vrHUDDepthOffsetSlider = CreateSlider(_vrHUDDepthOffsetJSON);
                _vrHUDDepthOffsetSlider.label = "HUD Distance";
            }
            if (_vrHandControlJSON.val != "Disabled") {
                _gazeMenuSelectModePopup = CreateSetupPopup(_gazeMenuSelectModeJSON, false, 200f, RefreshGazeControlModeChoices, null, false);
            }

            _spacer2 = CreateSpacer();
            _spacer2.height = 10;
            _displayGazeSelectionsToggle = CreateToggle(_displayGazeSelectionsJSON);
            _displayGazeSelectionsToggle.label = "Show Gaze Targets on Wrist UI";

            List<string> test = new List<string>();
            test.Add("Temp");
            PatreonFeatures.GridLayoutFeature(test);
            if (Constants._patreonContentEnabled && test != null && test.Count == 7)
            {
                _displayScreenLabelsToggle = CreateToggle(_displayScreenLabelsJSON);
                _displayScreenLabelsToggle.label = "Show Screen Labels on Wrist UI";
            }

            _spacer2b = CreateSpacer();
            _spacer2b.height = 10;
            _toolboxPopup = CreateSetupPopup(_toolboxJSON, false, 200f, ToolBoxPopupOpen, null, false);
            _toolboxPopup.labelTextColor = Color.blue;
            if (_toolboxJSON.val != "Select...") _toolbox.RefreshToolboxDisplay(_toolboxJSON.val);

            if (Constants._patreonContentEnabled && test != null && test.Count == 7)
            {
                _spacer3 = CreateSpacer();
                _spacer3.height = 10;
                _defaultButtonSkinDisplay.RefreshDisplay();
            }
            _spacer4 = CreateSpacer();
            _spacer4.height = 10;

            _pluginVersionTextField = CreateTextField(_pluginDetailsJSON, false);
            _pluginVersionTextField.height = 1;


        }
        public void RefreshRightUIDisplay() {
            CloseAllRightUIPopups(null);
            if (_buttonGridSelectorPopup != null) { RemovePopup(_buttonGridSelectorPopup); }

            _buttonGridSelectorPopup = CreateSetupPopup(_buttonGridSelectorJSON, true, 500f, ButtonGridSelectorPopupOpen, "Screen Selector");
            _buttonGridSelectorPopup.labelTextColor = Color.blue;
            _buttonGridDisplay.RefreshDisplay();

        }
        public void LoadUIAPSelected(string file) {
            try {
                if (file != "") {
                    JSONClass jc = (JSONClass)SuperController.singleton.LoadJSON(file);
                    _buttonGridsList.Clear();
                    LoadJSON(jc);
                    foreach (ButtonGrid buttonGrid in _buttonGridsList) {
                        buttonGrid.RecalcActiveButtonCount();
                        buttonGrid.RecalcGazeSelections();
                    }
                    _currentGridScreen = 0;
                    SwitchDisplayGrid("1");
                    RefreshLeftUIDisplay();
                    RefreshRightUIDisplay();
                }

            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        public void SaveUIAPSelected(string file) {
            try {
                if (file != "") {
                    if (!file.EndsWith(".uiap")) { file = file + ".uiap"; }
                    SuperController.singleton.SaveJSON(GetJSON(), file);
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void DeleteCurrentScreen() {
            try {
                int gridRef = int.Parse(_buttonGridSelectorJSON.val);
                _buttonGridsList.RemoveAt(gridRef - 1);
                if (gridRef > _buttonGridsList.Count) { _buttonGridSelectorJSON.val = _buttonGridsList.Count.ToString(); }
                else { SwitchDisplayGrid(gridRef.ToString()); }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void ButtonGridSelectorPopupOpen() {
            _buttonGridSelectorJSON.choices = GetGridChoices();
            CloseAllRightUIPopups(_buttonGridSelectorPopup); }

        public void CloseAllRightUIPopups(UIDynamicPopup exception) {
            if (_buttonGridSelectorPopup != null && _buttonGridSelectorPopup != exception) { _buttonGridSelectorPopup.popup.visible = false; }
            _buttonGridDisplay.CloseAllPopup(exception);
        }

        public void CloseAllLeftUIPopups(UIDynamicPopup exception)
        {
            if (_vrHandControlPopup != null && _vrHandControlPopup != exception) { _vrHandControlPopup.popup.visible = false; }
            if (_vrPinModePopup != null && _vrPinModePopup != exception) { _vrPinModePopup.popup.visible = false; }
            if (_gazeMenuSelectModePopup != null && _gazeMenuSelectModePopup != exception) { _gazeMenuSelectModePopup.popup.visible = false; }
            if (_toolboxPopup != null && _toolboxPopup != exception) { _toolboxPopup.popup.visible = false; }

            if (_defaultButtonSkinDisplay != null && _defaultButtonSkinDisplay._buttonSkinParamSelectorPopup != null && _defaultButtonSkinDisplay._buttonSkinParamSelectorPopup != exception) { _defaultButtonSkinDisplay._buttonSkinParamSelectorPopup.popup.visible = false; }
            if (_defaultButtonSkinDisplay != null && _defaultButtonSkinDisplay._skinStringChooserParamPopup != null && _defaultButtonSkinDisplay._skinStringChooserParamPopup != exception) { _defaultButtonSkinDisplay._skinStringChooserParamPopup.popup.visible = false; }
        }

        void Start() {
            try {

                if (_isActivePlugin) {
                    if (!_isLoaded) { CreateDefaultButtonGrids(); _isLoaded = true; }
                    _buttonGridSelectorJSON.choices = GetGridChoices();
                    _buttonGridDisplay = new ButtonGridDisplay(_buttonGridsList[0], this);
                    if (DefaultUIAPFileExists()) { LoadUIAPSelected(GetLocalPluginPath() + "\\default.uiap"); }
                    foreach (ButtonGrid buttonGrid in _buttonGridsList) {
                        buttonGrid.RecalcActiveButtonCount();
                        buttonGrid.RecalcGazeSelections();
                    }
                    RefreshLeftUIDisplay();
                    RefreshRightUIDisplay();
                    _lastGazeTargetUpateTime = Time.time;
                }

            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        void DisableHeelAdjust()
        {
            if (_isPrimaryUIAJSON.val)
            {
                foreach (Atom atom in SuperController.singleton.GetAtoms())
                {
                    if (atom.type == "Person")
                    {
                        _toolbox._heelAdjustTool.FootPoseReset(atom);
                    }
                }
            }
        }
        void OnEnable()
        {
            try
            {
                if (_uiMainCanvas != null) { _uiMainCanvas._canvas.enabled = true; }
                if (_uiButtonCanvas != null) { _uiButtonCanvas._canvas.enabled = true; }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        void OnDisable()
        {
            try
            {
                DisableHeelAdjust();
                if (_uiMainCanvas != null) { _uiMainCanvas._canvas.enabled = false; }
                if (_uiButtonCanvas != null) { _uiButtonCanvas._canvas.enabled = false; }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        void OnDestroy() {
            try {
                DisableHeelAdjust();
                if (_uiMainCanvas != null) { _uiMainCanvas.OnDestroy(); }
                if (_uiButtonCanvas != null) { _uiButtonCanvas.OnDestroy(); }
                SuperController.singleton.onAtomUIDRenameHandlers -= _atomRenameHandle;
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void UpdateButtonTextures() {
            ButtonGrid buttonGrid = _buttonGridsList[_currentGridScreen];
            int buttonCount = _uiButtonList.Count;
            List<string> test = new List<string>();
            test.Add("Temp");
            PatreonFeatures.GridLayoutFeature(test);
            if (!Constants._patreonContentEnabled || test == null) {
                buttonCount = Mathf.Min(buttonCount, 9);
            }

            for (int buttonRef = 0; buttonRef < buttonCount; buttonRef++) {
                UIDynamicButton uiButton = _uiButtonList[buttonRef];
                Button button = buttonGrid._buttonList[buttonRef];
                Texture2D thumbnailTexture = null;
                if (button._buttonType != "Blank" && button._buttonTexture != null && button._useButtonTexture && button._buttonType.Length > 6 && button._buttonType.Substring(button._buttonType.Length - 6, 6) == "Preset") {
                    thumbnailTexture = button._buttonTexture;
                }
                else {
                    int buttonState = 1;
                    string targetAtom = "";
                    if (button._buttonType != "Blank" && button._targetCategory == "Gaze Selected Atom") {
                        switch (button._targetName) {
                            case "Last viewed Person":
                                targetAtom = _lastViewedPerson;
                                break;
                            case "Last viewed Female":
                                targetAtom = _lastViewedFemale;
                                break;
                            case "Last viewed Male":
                                targetAtom = _lastViewedMale;
                                break;
                            case "Last viewed Atom":
                                targetAtom = _lastViewedAtom;
                                break;
                        }
                        if (targetAtom != "") { buttonState = button.GetButtonState(targetAtom); }
                    }
                    else { buttonState = button.GetButtonState(""); }
                    if (buttonState == 1) {
                        if (button._buttonSkin == null || button._buttonSkin._customButtonOnTextureDefault) { thumbnailTexture = _defaultButtonSkin._buttonOnTexture; }
                        else { thumbnailTexture = button._buttonSkin._buttonOnTexture; }
                    }
                    else {
                        if (button._buttonSkin == null || button._buttonSkin._customButtonOffTextureDefault) { thumbnailTexture = _defaultButtonSkin._buttonOffTexture; }
                        else { thumbnailTexture = button._buttonSkin._buttonOffTexture; }

                    }
                    if (!_secondValidation) thumbnailTexture = null;
                }
                if (thumbnailTexture != null) {
                    Sprite mySprite = Sprite.Create(thumbnailTexture, new Rect(0.0f, 0.0f, thumbnailTexture.width, thumbnailTexture.height), new Vector2(0.5f, 0.5f), 100.0f);
                    uiButton.buttonImage.sprite = mySprite;
                }
            }
        }
        private void UpdateButtonLabels(bool updateGazeButtons = true, bool updateGridButtons = true, bool enableGazeToggleScreens = false) {
            ButtonGrid buttonGrid=null;
            if (_currentGridScreen>=0) buttonGrid = _buttonGridsList[_currentGridScreen];
            int buttonCount = _uiButtonList.Count;
            int nearestButtonRef = -1;
            float nearestButtonDistance = -1f;
            Vector3[] rectCornersArray = new Vector3[4];
            Transform cameraT = SuperController.singleton.lookCamera.transform;
            Vector3 eyePosition = cameraT.position - (cameraT.up * 0.1f);

            Ray ray = new Ray(eyePosition, cameraT.forward);

            List<string> test = new List<string>();
            test.Add("Temp");
            PatreonFeatures.GridLayoutFeature(test);
            if (!Constants._patreonContentEnabled || test == null) {
                buttonCount = Mathf.Min(buttonCount, 9);
            }
            if (updateGridButtons) {
                for (int buttonRef = 0; buttonRef < buttonCount; buttonRef++) {
                    UIDynamicButton uiButton = _uiButtonList[buttonRef];
                    Button button = buttonGrid._buttonList[buttonRef];
                    ButtonSkin buttonSkin = button._buttonSkin;
                    int buttonState = 1;
                    string targetAtom = "";
                    if (button._buttonType == "Blank") {
                        uiButton.gameObject.SetActive(false);
                    }
                    else {
                        uiButton.gameObject.SetActive(true);
                        if (updateGazeButtons || enableGazeToggleScreens)
                        {
                            RectTransform rtButton = uiButton.transform.GetComponent<RectTransform>();
                            rtButton.GetWorldCorners(rectCornersArray);
                            Vector3 buttonCentre1 = rectCornersArray[0] + ((rectCornersArray[3] - rectCornersArray[0]) * 0.5f);
                            float distance = Vector3.Cross(ray.direction, buttonCentre1 - ray.origin).magnitude;
                            if (distance < nearestButtonDistance || nearestButtonDistance < 0f)
                            {
                                nearestButtonDistance = distance;
                                nearestButtonRef = buttonRef;
                            }
                        }
                        if (button._targetCategory == "Gaze Selected Atom" && button.IsAtomTargetable()) {
                            switch (button._targetName) {
                                case "Last viewed Person":
                                    targetAtom = _lastViewedPerson;
                                    break;
                                case "Last viewed Female":
                                    targetAtom = _lastViewedFemale;
                                    break;
                                case "Last viewed Male":
                                    targetAtom = _lastViewedMale;
                                    break;
                                case "Last viewed Atom":
                                    targetAtom = _lastViewedAtom;
                                    break;
                            }
                            if (targetAtom != "") { buttonState = button.GetButtonState(targetAtom); }
                        }
                        else { buttonState = button.GetButtonState(""); }

                        string targetFont = "Arial";
                        if (buttonSkin == null || buttonSkin._buttonFontDefault || !_secondValidation) { targetFont = _defaultButtonSkin._buttonFont; }
                        else { targetFont = buttonSkin._buttonFont; }

                        Font selectedFont = (Font)Resources.GetBuiltinResource(typeof(Font), "Arial.ttf");
                        foreach (Font font in Resources.FindObjectsOfTypeAll(typeof(Font)) as Font[]) {
                            if (font.name == targetFont) selectedFont = font;
                        }
                        uiButton.buttonText.font = selectedFont;

                        HSVColor hsvTextColor = new HSVColor();
                        HSVColor hsvButtonColor = new HSVColor();
                        if (buttonState == 1) {
                            if (button._autoLabel) uiButton.label = button.GetAutoLabel(1, targetAtom);
                            else uiButton.label = button.GetManualLabel(1, targetAtom);
                            if (buttonSkin == null || buttonSkin._textOnColorDefault || !_secondValidation) hsvTextColor = _defaultButtonSkin._textOnColor;
                            else hsvTextColor = buttonSkin._textOnColor;
                            if (buttonSkin == null || buttonSkin._buttonOnColorDefault || !_secondValidation) hsvButtonColor = _defaultButtonSkin._buttonOnColor;
                            else hsvButtonColor = buttonSkin._buttonOnColor;

                        }
                        else if (buttonState == 2) {

                            if (button._autoLabel) uiButton.label = button.GetAutoLabel(2, targetAtom);
                            else uiButton.label = button.GetManualLabel(2, targetAtom);
                            if (buttonSkin == null || buttonSkin._textOffColorDefault || !_secondValidation) hsvTextColor = _defaultButtonSkin._textOffColor;
                            else hsvTextColor = buttonSkin._textOffColor;
                            if (buttonSkin == null || buttonSkin._buttonOffColorDefault || !_secondValidation) hsvButtonColor = _defaultButtonSkin._buttonOffColor;
                            else hsvButtonColor = buttonSkin._buttonOffColor;
                        }
                        if (buttonSkin == null || buttonSkin._fontSizeDefault || !_secondValidation) {
                            if (buttonGrid._buttonSize == "Small") {
                                uiButton.buttonText.resizeTextMaxSize = (int)_defaultButtonSkin._maxSmallFontSize;
                                uiButton.buttonText.resizeTextMinSize = (int)_defaultButtonSkin._minSmallFontSize;
                            }
                            else if (buttonGrid._buttonSize == "Medium") {
                                uiButton.buttonText.resizeTextMaxSize = (int)_defaultButtonSkin._maxMediumFontSize;
                                uiButton.buttonText.resizeTextMinSize = (int)_defaultButtonSkin._minMediumFontSize;
                            }
                            else if (buttonGrid._buttonSize == "Large") {
                                uiButton.buttonText.resizeTextMaxSize = (int)_defaultButtonSkin._maxLargeFontSize;
                                uiButton.buttonText.resizeTextMinSize = (int)_defaultButtonSkin._minLargeFontSize;
                            }
                        }
                        else {
                            uiButton.buttonText.resizeTextMaxSize = (int)buttonSkin._maxFontSize;
                            uiButton.buttonText.resizeTextMinSize = (int)buttonSkin._minFontSize;
                        }
                        uiButton.buttonText.resizeTextForBestFit = true;
                        uiButton.buttonText.color = Color.HSVToRGB(hsvTextColor.H, hsvTextColor.S, hsvTextColor.V);
                        uiButton.buttonColor = Color.HSVToRGB(hsvButtonColor.H, hsvButtonColor.S, hsvButtonColor.V);

                        string textAlignment = "";
                        if (buttonSkin != null) textAlignment = buttonSkin._textAlignment;
                        if (buttonSkin == null || buttonSkin._textAlignmentDefault || !_secondValidation) { textAlignment = _defaultButtonSkin._textAlignment; }
                        if (textAlignment == "LowerCenter") { uiButton.buttonText.alignment = TextAnchor.LowerCenter; }
                        else if (textAlignment == "MiddleCenter") { uiButton.buttonText.alignment = TextAnchor.MiddleCenter; }
                        else if (textAlignment == "UpperCenter") { uiButton.buttonText.alignment = TextAnchor.UpperCenter; }
                        else if (textAlignment == "LowerLeft") { uiButton.buttonText.alignment = TextAnchor.LowerLeft; }
                        else if (textAlignment == "MiddleLeft") { uiButton.buttonText.alignment = TextAnchor.MiddleLeft; }
                        else if (textAlignment == "UpperLeft") { uiButton.buttonText.alignment = TextAnchor.UpperLeft; }
                        else if (textAlignment == "LowerRight") { uiButton.buttonText.alignment = TextAnchor.LowerRight; }
                        else if (textAlignment == "MiddleRight") { uiButton.buttonText.alignment = TextAnchor.MiddleRight; }
                        else if (textAlignment == "UpperRight") { uiButton.buttonText.alignment = TextAnchor.UpperRight; }
                    }
                }
            }
            float leftUIButtonDistance = -1f;
            float rightUIButtonDistance = -1f;
            float mainUIButtonDistance = -1f;
            float pinUIButtonDistance = -1f;

            if (updateGazeButtons || enableGazeToggleScreens)
            {
                RectTransform rt = _mainButton.transform.GetComponent<RectTransform>();
                rt.GetWorldCorners(rectCornersArray);
                Vector3 buttonCentre = rectCornersArray[0] + ((rectCornersArray[3] - rectCornersArray[0]) * 0.5f);
                mainUIButtonDistance = Vector3.Cross(ray.direction, buttonCentre - ray.origin).magnitude;

                if (_leftButton.gameObject.activeSelf)
                {
                    rt = _leftButton.transform.GetComponent<RectTransform>();
                    rt.GetWorldCorners(rectCornersArray);
                    buttonCentre = rectCornersArray[0] + ((rectCornersArray[3] - rectCornersArray[0]) * 0.5f);
                    leftUIButtonDistance = Vector3.Cross(ray.direction, buttonCentre - ray.origin).magnitude;
                }
                if (_rightButton.gameObject.activeSelf)
                {
                    rt = _rightButton.transform.GetComponent<RectTransform>();
                    rt.GetWorldCorners(rectCornersArray);
                    buttonCentre = rectCornersArray[0] + ((rectCornersArray[3] - rectCornersArray[0]) * 0.5f);
                    rightUIButtonDistance = Vector3.Cross(ray.direction, buttonCentre - ray.origin).magnitude;
                }
                if (_mainPinButton.gameObject.activeSelf)
                {
                    rt = _mainPinButton.transform.GetComponent<RectTransform>();
                    rt.GetWorldCorners(rectCornersArray);
                    buttonCentre = rectCornersArray[0] + ((rectCornersArray[3] - rectCornersArray[0]) * 0.5f);
                    pinUIButtonDistance = Vector3.Cross(ray.direction, buttonCentre - ray.origin).magnitude;
                }
            }

            if (_mainPinButton.gameObject.activeSelf && _uiPinned) _mainPinButton.label = "O";
            else if (_mainPinButton.gameObject.activeSelf && !_uiPinned) _mainPinButton.label = "X";

            _gazeUIButton = -1;
            _gazeMainButton = false;
            _gazeLeftButton = false;
            _gazeRightButton = false;
            _gazePinButton = false;
            _leftButton.buttonColor = Color.grey;
            _rightButton.buttonColor = Color.grey;
            _mainPinButton.buttonColor = Color.grey;
            _mainButton.buttonColor = Color.grey;
            if (updateGazeButtons)
            {
                Color gazeButtonColor = new Color();
                gazeButtonColor = Color.HSVToRGB(_defaultButtonSkin._buttonGazeSelectColor.H, _defaultButtonSkin._buttonGazeSelectColor.S, _defaultButtonSkin._buttonGazeSelectColor.V);

                if (nearestButtonDistance > -1f && nearestButtonDistance < mainUIButtonDistance && (leftUIButtonDistance == -1f || nearestButtonDistance < leftUIButtonDistance) && (rightUIButtonDistance == -1f || nearestButtonDistance < rightUIButtonDistance) && (pinUIButtonDistance == -1f || nearestButtonDistance < pinUIButtonDistance))
                {
                    if (nearestButtonDistance < 0.1f)
                    {
                        _uiButtonList[nearestButtonRef].buttonColor = gazeButtonColor;
                        if (updateGazeButtons) _gazeUIButton = nearestButtonRef;
                    }
                }
                else if (leftUIButtonDistance > -1f && leftUIButtonDistance < mainUIButtonDistance && (rightUIButtonDistance == -1f || leftUIButtonDistance < rightUIButtonDistance) && (pinUIButtonDistance == -1f || leftUIButtonDistance < pinUIButtonDistance))
                {
                    if (leftUIButtonDistance < 0.1f)
                    {
                        _leftButton.buttonColor = gazeButtonColor;
                        if (updateGazeButtons) _gazeLeftButton = true;
                    }
                }
                else if (rightUIButtonDistance > -1f && rightUIButtonDistance < mainUIButtonDistance && (pinUIButtonDistance == -1f || rightUIButtonDistance < pinUIButtonDistance))
                {
                    if (rightUIButtonDistance < 0.1f)
                    {
                        _rightButton.buttonColor = gazeButtonColor;
                        if (updateGazeButtons) _gazeRightButton = true;
                    }
                }
                else if (pinUIButtonDistance > -1f && pinUIButtonDistance < mainUIButtonDistance)
                {
                    if (pinUIButtonDistance < 0.1f)
                    {
                        _mainPinButton.buttonColor = gazeButtonColor;
                        if (updateGazeButtons) _gazePinButton = true;
                    }
                }
                else
                {
                    if (mainUIButtonDistance < 0.1f)
                    {
                        _mainButton.buttonColor = gazeButtonColor;
                        if (updateGazeButtons) _gazeMainButton = true;
                    }
                }
            }

            if (updateGridButtons) {

                if (_displayGazeSelectionsJSON.val && (buttonGrid._gazeSelectionFemales || buttonGrid._gazeSelectionMales || buttonGrid._gazeSelectionPerson || buttonGrid._gazeSelectionAtoms)) {
                    string text = "";
                    int height = 14;
                    _gazeStatusTextField.gameObject.SetActive(true);
                    if (buttonGrid._gazeSelectionFemales) { text = text + "\n   Target Female: " + _lastViewedFemale; height = height + 7; }
                    if (buttonGrid._gazeSelectionMales) { text = text + "\n   Target Male: " + _lastViewedMale; height = height + 7; }
                    if (buttonGrid._gazeSelectionPerson) { text = text + "\n   Target Person: " + _lastViewedPerson; height = height + 7; }
                    if (buttonGrid._gazeSelectionAtoms) { text = text + "\n   Target Atom: " + _lastViewedAtom; height = height + 7; }

                    _gazeStatusTextField.text = text;
                    _gazeStatusTextField.height = height;
                }
                else if (_gazeStatusTextField != null) { _gazeStatusTextField.gameObject.SetActive(false); }
            }

            Vector3 mainButtonToEyeDirection = _mainPinButton.transform.position - cameraT.position;
            float mainButtonToEyeAngle = Vector3.Angle(_mainPinButton.transform.forward, mainButtonToEyeDirection);
            if (!_uiActivated && mainUIButtonDistance < 0.2f && enableGazeToggleScreens && mainButtonToEyeAngle < 35f && _isUIGazeDeactivated) ToggleUIActivateButton();
            if (_uiActivated && ((mainUIButtonDistance > 0.25f && nearestButtonDistance > 0.1f) || mainButtonToEyeAngle > 90f) && enableGazeToggleScreens)
            {
                ToggleUIActivateButton();
                _isUIGazeDeactivated = true;
            }
            if (!_uiActivated && ((mainUIButtonDistance > 0.25f && nearestButtonDistance > 0.1f) || mainButtonToEyeAngle > 90f)) _isUIGazeDeactivated = true;
        }
        private void UpdateLeftRightButtons() {
            if (_uiActivated) {
                if (GetNextLeftScreen() >= 0) { _leftButton.gameObject.SetActive(true); }
                else { _leftButton.gameObject.SetActive(false); }
                if (GetNextRightScreen() >= 0) { _rightButton.gameObject.SetActive(true); }
                else { _rightButton.gameObject.SetActive(false); }
            }
            else {
                _leftButton.gameObject.SetActive(false);
                _rightButton.gameObject.SetActive(false);
            }
        }
        private void UpdateGazeTargets() {
            List<GazeTarget> gazeTargetsNarrowFOVList = new List<GazeTarget>();
            List<GazeTarget> gazeTargetsWideFOVList = new List<GazeTarget>();
            ButtonGrid buttonGrid = _buttonGridsList[_currentGridScreen];

            foreach (Atom atom in SuperController.singleton.GetAtoms()) {
                bool male = false;
                bool female = false;
                if (atom.type == "Person" && atom.on &&(atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("male") || atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("jarlee") || atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("jarjulian"))) male = true;
                else if (atom.type == "Person" && atom.on) { female = true; }
                if (buttonGrid._gazeSelectionAtoms || (buttonGrid._gazeSelectionFemales && female) || (buttonGrid._gazeSelectionMales && male) || ((buttonGrid._gazeSelectionPerson && (male || female)))) {
                    if (atom.name != "CoreControl" && atom.name != "[CameraRig]") {
                        foreach (Rigidbody rb in atom.realRigidbodies) {
                            Vector3 rbPosition = rb.transform.position;
                            Transform playerHeadTransform = CameraTarget.centerTarget.transform;
                            Vector3 playerLookingVector = playerHeadTransform.forward;
                            Vector3 directionToTarget = rbPosition - playerHeadTransform.position;
                            float angle = Vector3.Angle(playerLookingVector, directionToTarget);
                            float distance = Vector3.Distance(rbPosition, playerHeadTransform.position);
                            if (angle <= _gazeFOVNarrow / 2f) {
                                GazeTarget gazeTarget = new GazeTarget();
                                gazeTarget.atomName = atom.name;
                                gazeTarget.nodeName = rb.name;
                                gazeTarget.angle = angle;
                                gazeTarget.distance = distance;
                                gazeTargetsNarrowFOVList.Add(gazeTarget);
                            }
                            if (angle <= _gazeFOVWide / 2f) {
                                GazeTarget gazeTarget = new GazeTarget();
                                gazeTarget.atomName = atom.name;
                                gazeTarget.nodeName = rb.name;
                                gazeTarget.angle = angle;
                                gazeTarget.distance = distance;
                                gazeTargetsWideFOVList.Add(gazeTarget);
                            }
                        }
                    }
                }
            }
            string atomName = "";
            if (gazeTargetsNarrowFOVList.Count > 0) {
                float distance = -1f;
                foreach (GazeTarget gazeTarget in gazeTargetsNarrowFOVList) {
                    if (distance == -1f || gazeTarget.distance < distance) {
                        atomName = gazeTarget.atomName;
                        distance = gazeTarget.distance;
                    }
                }
            }
            else if (gazeTargetsWideFOVList.Count > 0) {
                float angle = -1f;
                foreach (GazeTarget gazeTarget in gazeTargetsWideFOVList) {
                    if (angle == -1f || gazeTarget.angle < angle) {
                        atomName = gazeTarget.atomName;
                        angle = gazeTarget.angle;
                    }
                }
            }
            if (atomName != "") {
                _lastViewedAtom = atomName;
                Atom atom = SuperController.singleton.GetAtomByUid(atomName);
                if (atom.type == "Person") {
                    _lastViewedPerson = atomName;
                    if (atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("male") || atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("jarlee") || atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("jarjulian"))  _lastViewedMale = atomName;
                    else  _lastViewedFemale = atomName;
                }
            }
        }
        public void SwitchVRHand(string pluginHand)
        {
            UpdateOtherUIAssistVRHand(pluginHand);
            if (!_isDisplayUpdating)
            {
                RefreshLeftUIDisplay();
            }
        }
        public void SwitchToolBox(string toolbox)
        {
            if (!_isDisplayUpdating)
            {
                RefreshLeftUIDisplay();
            }
        }
        public void SwitchVRPinMode(string pinMode)
        {
            if (!_isDisplayUpdating)
            {
                RefreshLeftUIDisplay();
            }
        }

        private void UpdateOtherUIAssistVRHand(string pluginHand) {
            foreach (string storableName in containingAtom.GetStorableIDs()) {
                if (storableName.StartsWith("plugin#")) {
                    string pluginType = storableName.Substring(storableName.IndexOf('_') + 1);

                    if (pluginType == "JayJayWon.UIAssist" && storableName != this.name) {
                        JSONStorable pluginStorable = containingAtom.GetStorableByID(storableName);
                        JSONStorableBool activeJSON = pluginStorable.GetBoolJSONParam("activeUIAssist");
                        if (activeJSON.val) {
                            JSONStorableStringChooser otherPluginVRHand = pluginStorable.GetStringChooserJSONParam("handControl");
                            if (otherPluginVRHand.val == _vrHandControlJSON.val) {
                                if (_vrHandControlJSON.val == "Left Hand") { otherPluginVRHand.val = "Right Hand"; }
                                else { otherPluginVRHand.val = "Left Hand"; }
                            }
                            return;
                        }
                    }

                }
            }
        }

        private void UpdateActiveClothingLists()
        {
            _atomActiveClothingDictionary.Clear();
            foreach (Atom atom in SuperController.singleton.GetAtoms())
            {
                if (atom.type == "Person")
                {
                    ActiveClothingLists activeClothingLists = new ActiveClothingLists();
                    activeClothingLists.UpdateActiveClothing(atom);
                    _atomActiveClothingDictionary.Add(atom.name, activeClothingLists);

                    if (_toolbox._heelAdjustTool._heelAdjustActiveJSON.val && _isPrimaryUIAJSON.val) _toolbox._heelAdjustTool.LoadFootPosesForClothing(atom, activeClothingLists._activeHAClothingItems);
                }

            }
        }

        void Update() {
            try {
                if (_isActivePlugin) {
                    string uiDisplayMode = "";
                    bool gazeControlActive = false;

                    if (XRSettings.enabled == false) uiDisplayMode = "HUD";
                    else if (_vrHandControlJSON.val == "Disabled" || (_uiPinned && (_vrPinModeJSON.val == "HUD (Left)" || _vrPinModeJSON.val == "HUD (Right)" || _vrPinModeJSON.val == "HUD (Custom)"))) uiDisplayMode = "VRHUD";
                    else if (_vrPinModeJSON.val == "Disabled")
                    {
                        uiDisplayMode = "WRIST";
                        if (_gazeMenuSelectModeJSON.val != "Disabled") gazeControlActive = true;
                    }
                    else if (_uiPinned)
                    {
                        uiDisplayMode = "STATIC";
                        if (_gazeMenuSelectModeJSON.val == "Wrist & Pinned UI") gazeControlActive = true;
                    }
                    else
                    {
                        uiDisplayMode = "WRIST";
                        if (_gazeMenuSelectModeJSON.val != "Disabled") gazeControlActive = true;
                    }

                    // Check if pin Mode is available
                    bool pinModeAvailable = false;
                    if (uiDisplayMode != "HUD" && _vrPinModeJSON.val != "Disabled" && _vrHandControlJSON.val != "Disabled" && _uiActivated)
                    {
                        _mainPinButton.gameObject.SetActive(true);
                        pinModeAvailable = true;
                    }
                    else _mainPinButton.gameObject.SetActive(false);

                    bool detachToggle = false;
                    float currentTime = Time.time;
                    bool triggerUIButton = false;

                    if (_timeLastHAFileListUpdated < 0f || (currentTime - _timeLastHAFileListUpdated > Constants._TimeBetweenHAFileUpdates))
                    {
                        if (_toolbox._heelAdjustTool._heelAdjustActiveJSON.val)_toolbox._heelAdjustTool.UpdateExistingHAFiles();
                        _timeLastHAFileListUpdated = currentTime;
                    }

                    if (_timeLastClothingUpdate < 0f || ((currentTime - _timeLastClothingUpdate) > Constants._TimeBetweenClothingUpdates))
                    {
                        UpdatePrimaryUIA();
                        _toolbox._heelAdjustTool.ResetAllSavedAtomRotationStates();
                        UpdateActiveClothingLists();
                        _timeLastClothingUpdate = currentTime;
                    }

                    // Check if Grab activated and trigger gaze action 
                    if (SuperController.singleton.GetLeftGrab() && _vrHandControlJSON.val == "Left Hand" && gazeControlActive) triggerUIButton = true;
                    if (SuperController.singleton.GetRightGrab() && _vrHandControlJSON.val == "Right Hand" && gazeControlActive) triggerUIButton = true;
                    if (triggerUIButton) {
                        if (_gazeUIButton > -1) ButtonActivated(_gazeUIButton + 1);
                        else if (_gazeMainButton) ToggleUIActivateButton();
                        else if (_gazeLeftButton) ScrollLeftButton();
                        else if (_gazeRightButton) ScrollRightButton();
                        else if (_gazePinButton) TogglePin();
                    }

                    // Check if a double click Select is starting completing - to allow Toggle the pin
                    if (pinModeAvailable)
                    {
                        if (SuperController.singleton.GetLeftSelect())
                        {
                            if (_timeLeftSelect == -1f) _timeLeftSelect = currentTime;
                            else
                            {
                                if (currentTime - _timeLeftSelect > 0.4f) _timeLeftSelect = currentTime;
                                else
                                {
                                    _timeLeftSelect = -1f;
                                    if (_vrHandControlJSON.val == "Left Hand") detachToggle = true;
                                }
                            }
                        }
                        if (SuperController.singleton.GetRightSelect())
                        {
                            if (_timeRightSelect == -1f) _timeRightSelect = currentTime;
                            else
                            {
                                if (currentTime - _timeRightSelect > 0.4f) _timeRightSelect = currentTime;
                                else
                                {
                                    _timeRightSelect = -1f;
                                    if (_vrHandControlJSON.val == "Right Hand") detachToggle = true;
                                }
                            }
                        }
                        if (detachToggle) TogglePin();
                    }

                    // Set the 'read only' parameters that are visible to other plugins
                    _isActivePluginJSON.val = true;
                    _pluginVersionJSON.val = pluginVersion;
                    _pluginVersionFloatJSON.val = pluginVersionFloat;
                    _isPatreonOnlyVersionJSON.val = Constants._patreonContentEnabled;

                    // Update Gaze targets every 0.1 seconds
                    if (currentTime - _lastGazeTargetUpateTime > 0.1f) {
                        UpdateOtherUIAssistVRHand(_vrHandControlJSON.val);
                        if (_uiActivated) {
                            UpdateGazeTargets();
                            _lastGazeTargetUpateTime = currentTime;
                        }
                    }

                    UpdateButtonLabels(gazeControlActive, _uiActivated, (_gazeActivateVRScreenJSON.val && !(uiDisplayMode == "HUD") && !(uiDisplayMode == "VRHUD")));

                    Transform vrHandTransform;
                    bool vrRightHand = _vrHandControlJSON.val == "Right Hand";
                    if (vrRightHand) vrHandTransform = SuperController.singleton.rightHand;
                    else vrHandTransform = SuperController.singleton.leftHand;
                    Vector3 canvasPosition;
                    Quaternion canvasRotation;
                    float horizontalHUDOffset = 0.9f;
                    float verticalHUDOffset = 0.06f;
                    float depthHUDOffset = 0.6f;

                    if (uiDisplayMode == "STATIC") {
                        canvasPosition = _fixedPinPosition;
                        canvasRotation = _fixedPinRotation;
                        float distance = Vector3.Distance(_fixedPinPosition, vrHandTransform.position);
                        if (distance > 0.3f) _autoDePinEnabled = true;
                        if (distance < 0.1f & _autoDePinEnabled) TogglePin();
                    }
                    else {
                        canvasPosition = vrHandTransform.position;
                        canvasRotation = vrHandTransform.rotation;
                    }
                    if (uiDisplayMode == "VRHUD")
                    {
                        if (_vrPinModeJSON.val == "HUD (Right)")
                        {
                            horizontalHUDOffset = 0.57f;
                            verticalHUDOffset = 0.3f;
                        }
                        else if (_vrPinModeJSON.val == "HUD (Left)")
                        {
                            horizontalHUDOffset = 0.43f;
                            verticalHUDOffset = 0.3f;
                        }
                        else if (_vrPinModeJSON.val == "HUD (Custom)")
                        {
                            horizontalHUDOffset = _vrHUDHorizontalOffsetJSON.val;
                            verticalHUDOffset = _vrHUDVerticalOffsetJSON.val;
                            depthHUDOffset = _vrHUDDepthOffsetJSON.val;
                        }
                    }
                    _uiMainCanvas.Update(canvasPosition, canvasRotation, uiDisplayMode, uiDisplayMode == "STATIC", vrRightHand, verticalHUDOffset, horizontalHUDOffset, depthHUDOffset);
                    _uiButtonCanvas.Update(canvasPosition, canvasRotation, uiDisplayMode, true, vrRightHand, verticalHUDOffset, horizontalHUDOffset, depthHUDOffset);

                    UpdateLeftRightButtons();
                    if (_uiActivated && _buttonGridsList[_currentGridScreen]._gridLabel != "" && _displayScreenLabelsJSON.val && Constants._patreonContentEnabled) _mainButton.label = _buttonGridsList[_currentGridScreen]._gridLabel;
                    else _mainButton.label = System.DateTime.Now.ToString("HH:mm");
                    //Vector3 fingerTip = trans.position + (trans.up * 0.03f);
                    //float distance = Vector3.Distance(fingerTip,_mainButton.button.transform.position);
                    //_mainButton.label = distance.ToString("0.00");
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        protected WristUI _uiMainCanvas;
        protected WristUI _uiButtonCanvas;
        private UIDynamicButton _mainButton;
        private UIDynamicButton _leftButton;
        private UIDynamicButton _rightButton;
        private UIDynamicTextField _gazeStatusTextField;
        private List<UIDynamicButton> _uiButtonList;
        private UIDynamicButton _mainPinButton;
        private Vector3 _fixedPinPosition;
        private Quaternion _fixedPinRotation;

        private UIDynamicButton _loadProfileButton;
        private UIDynamicButton _saveProfileButton;
        private UIDynamicButton _setDefaultProfileButton;
        private UIDynamicPopup _vrHandControlPopup;
        private UIDynamicPopup _vrPinModePopup;
        private UIDynamicPopup _gazeMenuSelectModePopup;
        private UIDynamicPopup _toolboxPopup;
        private UIDynamicSlider _vrHUDHorizontalOffsetSlider;
        private UIDynamicSlider _vrHUDVerticalOffsetSlider;
        private UIDynamicSlider _vrHUDDepthOffsetSlider;

        private UIDynamicToggle _displayGazeSelectionsToggle;
        private UIDynamicToggle _displayScreenLabelsToggle;
        private UIDynamicToggle _gazeActivateVRScreenToggle;
        private UIDynamicTextField _pluginVersionTextField;
        private UIDynamic _spacer1;
        private UIDynamic _spacer2;
        private UIDynamic _spacer2b;
        private UIDynamic _spacer3;
        private UIDynamic _spacer4;

        public bool _uiActivated = false;
        public bool _isRestoring = false;
        private bool _uiPinned = false;
        private int _gazeUIButton = -1;
        private bool _gazeMainButton = false;
        private bool _gazeLeftButton = false;
        private bool _gazeRightButton = false;
        private bool _gazePinButton = false;
        private bool _autoDePinEnabled = false;

        private int _currentGridScreen = 0;
        private string _lastViewedMale = "";
        private string _lastViewedFemale = "";
        private string _lastViewedPerson = "";
        private string _lastViewedAtom = "";
        private float _lastGazeTargetUpateTime;
        private bool _secondValidation = false;
        private float _timeLeftSelect = -1f;
        private float _timeRightSelect = -1f;
        private float _timeLastClothingUpdate = -1f;
        private float _timeLastHAFileListUpdated = -1f;
        private bool _isUIGazeDeactivated = true;
        public bool _suppressScaleLoad = false;

        bool _isDisplayUpdating = false;

        public JSONStorableStringChooser _buttonGridSelectorJSON;
        public bool _isLoaded;
        public JSONStorableBool _isPrimaryUIAJSON;

        public bool _isActivePlugin = false;
        public JSONStorableBool _isActivePluginJSON;
        public JSONStorableFloat _pluginVersionFloatJSON;
        public JSONStorableString _pluginVersionJSON;
        public JSONStorableBool _isPatreonOnlyVersionJSON;

        public JSONStorableStringChooser _vrHandControlJSON;
        public JSONStorableStringChooser _vrPinModeJSON;
        public JSONStorableStringChooser _gazeMenuSelectModeJSON;
        public JSONStorableStringChooser _toolboxJSON;

        public JSONStorableFloat _vrHUDHorizontalOffsetJSON;
        public JSONStorableFloat _vrHUDVerticalOffsetJSON;
        public JSONStorableFloat _vrHUDDepthOffsetJSON;
        public JSONStorableBool _gazeActivateVRScreenJSON;

        public JSONStorableBool _displayGazeSelectionsJSON;
        public JSONStorableBool _displayScreenLabelsJSON;
        public JSONStorableString _pluginDetailsJSON;
        private UIDynamicPopup _buttonGridSelectorPopup;

        public ButtonSkin _defaultButtonSkin;
        public ButtonSkinDisplay _defaultButtonSkinDisplay;
        public List<ButtonGrid> _buttonGridsList = new List<ButtonGrid>();
        public ButtonGridDisplay _buttonGridDisplay;

        public ToolBox _toolbox;

        public Dictionary<string, ActiveClothingLists> _atomActiveClothingDictionary = new Dictionary<string, ActiveClothingLists>();

        SuperController.OnAtomUIDRename _atomRenameHandle;
    }

    public class GazeTarget {
        public string atomName;
        public string nodeName;
        public float angle;
        public float distance;
    }

    public class ActiveClothingLists
    {
        public List<string> _activeHAClothingItems;
        public Dictionary<string, List<string>> _activeClothingTags;

        private string StripNonAlphaNum(string s)
        {
            return Regex.Replace(s, @"[^A-Za-z0-9]+", "");
        }
        private string GetHAItemNameFromCustomClothingUID(string uid)
        {
            string itemName = uid;
            if (uid.Contains("/"))
            {
                itemName = uid.Substring(uid.LastIndexOf("/") + 1);
                if (uid.EndsWith(".vam"))
                {
                    itemName = itemName.Substring(0, itemName.Length - 4);
                }
            }
            return (itemName);
        }
        public void UpdateActiveClothing(Atom atom)
        {
            _activeHAClothingItems = new List<string>();
            _activeClothingTags = new Dictionary<string, List<string>>();

            foreach (Transform child in atom.gameObject.transform.Find("rescale2/geometry/FemaleClothes"))
            {
                if (child.gameObject.name == "FemaleClothingPrefab(Clone)")
                {
                    foreach (Transform subchild in atom.gameObject.transform.Find("rescale2/geometry/FemaleClothes/FemaleClothingPrefab(Clone)"))
                    {
                        DAZClothingItem dci = subchild.GetComponent<DAZClothingItem>();
                        if (dci.isActiveAndEnabled)
                        {                         
                            string[] tags = dci.tagsArray;
                            if (tags.Length > 0)
                            {
                                List<string> newTagList = tags.ToList();
                                _activeClothingTags.Add(dci.uid, newTagList);
                            }

                            string haClothingItemName = StripNonAlphaNum(dci.creatorName) + "-" + StripNonAlphaNum(dci.uid);
                            _activeHAClothingItems.Add(haClothingItemName);
                        }
                    }
                }
                else
                {
                    if (child.GetComponent<DAZClothingItem>().isActiveAndEnabled)
                    {
                        DAZClothingItem dci = child.GetComponent<DAZClothingItem>();
                        if (dci.isActiveAndEnabled)
                        {
                            string[] tags = dci.tagsArray;
                            if (tags.Length > 0)
                            {
                                List<string> newTagList = tags.ToList();
                                _activeClothingTags.Add(dci.uid, newTagList);
                            }
                            string haClothingItemName = StripNonAlphaNum(dci.creatorName) + "-" + StripNonAlphaNum(GetHAItemNameFromCustomClothingUID(dci.uid));
                            _activeHAClothingItems.Add(haClothingItemName);
                        }
                    }
                }
            }
        }

    }

    public class ButtonGrid {
        public int _gridRows = 0;
        public int _gridColumns = 0;
        public string _buttonSize = "Medium";
        public int _activeButtonCount = 0;
        public int _maxButtonRow = 0;
        public bool _gazeSelectionAtoms = false;
        public bool _gazeSelectionFemales = false;
        public bool _gazeSelectionMales = false;
        public bool _gazeSelectionPerson = false;
        public string _gridLabel = "";

        public List<Button> _buttonList = new List<Button>();
        private UIAssist _mvrScript;

        public ButtonGrid(int columns = 0, int rows = 0, UIAssist mvr = null) {
            try {
                _gridRows = rows;
                _gridColumns = columns;
                _mvrScript = mvr;
                UpdateGridButtons();
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        public Button GetButtonBefore(int buttonRef) {
            Button button = null;
            if (buttonRef > 1) {
                for (int i = buttonRef - 2; i >= 0; i--) {
                    if (_buttonList[i]._buttonType != "Blank") {
                        button = _buttonList[i];
                        break;
                    }
                }
            }
            return (button);
        }

        public void UpdateGridButtons() {
            if (_buttonList.Count < (_gridRows * _gridColumns)) {
                for (int i = _buttonList.Count; i < (_gridRows * _gridColumns); i++) {
                    Button newButton = new Button(i + 1, _mvrScript);
                    _buttonList.Add(newButton);
                }
            }
        }
        public void RecalcActiveButtonCount() {
            _activeButtonCount = 0;
            _maxButtonRow = 0;
            foreach (Button button in _buttonList) {
                if (button._buttonType != "Blank" && button._buttonRef <= (_gridRows * _gridColumns)) {
                    _activeButtonCount++;
                    _maxButtonRow = (button._buttonRef / _gridColumns);
                    if (button._buttonRef % _gridColumns > 0) { _maxButtonRow++; }
                }
            }
        }
        public void RecalcGazeSelections() {

            _gazeSelectionAtoms = false;
            _gazeSelectionMales = false;
            _gazeSelectionFemales = false;
            _gazeSelectionPerson = false;
            foreach (Button button in _buttonList) {
                if (button._targetCategory == "Gaze Selected Atom" && button.IsAtomTargetable()) {
                    if (button._targetName == "Last viewed Atom") {
                        _gazeSelectionAtoms = true;
                    }
                    if (button._targetName == "Last viewed Person") {
                        _gazeSelectionPerson = true;
                    }
                    if (button._targetName == "Last viewed Male") {
                        _gazeSelectionMales = true;
                    }
                    if (button._targetName == "Last viewed Female") {
                        _gazeSelectionFemales = true;
                    }
                }
            }

        }

        public JSONClass GetJSON()
        {
            JSONClass node = new JSONClass();
            node["gridColumns"].AsInt = _gridColumns;
            node["gridRows"].AsInt = _gridRows;
            node["buttonSize"] = _buttonSize;
            node["activeButtonCount"].AsInt = _activeButtonCount;
            node["maxButtonRow"].AsInt = _maxButtonRow;
            node["gridLabel"] = _gridLabel;

            JSONArray buttonArray = new JSONArray();
            _buttonList.ForEach((button) =>
            {
                buttonArray.Add(button.GetJSON());
            });

            node["Buttons"] = buttonArray;

            return node;
        }
        public void LoadJSON(JSONClass gridJSON) {
            if (gridJSON["gridColumns"] != null) _gridColumns = gridJSON["gridColumns"].AsInt;
            if (gridJSON["gridRows"] != null) _gridRows = gridJSON["gridRows"].AsInt;
            if (gridJSON["buttonSize"] != null) _buttonSize = gridJSON["buttonSize"];
            if (gridJSON["activeButtonCount"] != null) _activeButtonCount = gridJSON["activeButtonCount"].AsInt;
            if (gridJSON["maxButtonRow"] != null) _maxButtonRow = gridJSON["maxButtonRow"].AsInt;
            if (gridJSON["gridLabel"] != null) _gridLabel = gridJSON["gridLabel"];

            _buttonList.Clear();
            if (gridJSON["Buttons"] != null)
            {
                JSONArray buttonArrayJSON = gridJSON["Buttons"].AsArray;
                for (int i = 0; i < buttonArrayJSON.Count; i++) {
                    JSONClass buttonJSON = buttonArrayJSON[i].AsObject;
                    Button button = new Button(i + 1, _mvrScript);
                    button.LoadJSON(buttonJSON);
                    _buttonList.Add(button);
                }
            }
        }
    }

    public class ButtonGridDisplay {
        public JSONStorableStringChooser _buttonLayoutJSON;
        public JSONStorableStringChooser _buttonSelectJSON;
        public JSONStorableStringChooser _buttonSizeJSON;
        public JSONStorableString _gridLabelJSON;

        public ButtonGrid _selectedButtonGrid;

        private UIDynamicTextField _gridLabelTextField;
        private UIAssist _mvrScript;
        private UIDynamicPopup _buttonLayoutPopup;
        private UIDynamicPopup _buttonSelectPopup;
        private UIDynamicPopup _buttonSizePopup;
        private UIDynamic _spacer1;
        public ButtonDisplay _buttonDisplay;
        bool _isDisplayUpdating = false;

        public ButtonGridDisplay(ButtonGrid bg, UIAssist mvrScript) {
            try {
                _selectedButtonGrid = bg;
                _mvrScript = mvrScript;
                _buttonLayoutJSON = new JSONStorableStringChooser("buttonLayout", GetButtonGridLayoutChoices(), "", "", SwitchButtonLayout);
                _buttonSelectJSON = new JSONStorableStringChooser("buttonSelection", null, "1", "", SwitchSelectedButton);
                List<string> buttonSizeOptions = new List<string>();
                buttonSizeOptions.Add("Small");
                buttonSizeOptions.Add("Medium");
                buttonSizeOptions.Add("Large");
                _buttonSizeJSON = new JSONStorableStringChooser("buttonSize", buttonSizeOptions, "Medium", "", SwitchButtonSize);
                _gridLabelJSON = new JSONStorableString("gridLabel", "", SwitchGridLabel);
                _buttonDisplay = new ButtonDisplay(bg._buttonList[0], mvrScript);
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public List<string> GetButtonGridLayoutChoices() {
            List<string> buttonGridLayoutChoices = new List<string>();
            buttonGridLayoutChoices.Add("2x2"); buttonGridLayoutChoices.Add("2x3"); buttonGridLayoutChoices.Add("3x2"); buttonGridLayoutChoices.Add("3x3");
            if (Constants._patreonContentEnabled) { PatreonFeatures.GridLayoutFeature(buttonGridLayoutChoices);

            }
            return (buttonGridLayoutChoices);
        }
        public void RefreshButtonSelectionChoices() {
            List<string> buttonSelectionChoices = new List<string>();
            _mvrScript.CloseAllRightUIPopups(_buttonSelectPopup);
            for (int i = 1; i <= (_selectedButtonGrid._gridRows * _selectedButtonGrid._gridColumns); i++) {
                buttonSelectionChoices.Add(i.ToString());
            }
            _buttonSelectJSON.choices = buttonSelectionChoices;
        }

        public void SwitchGridLabel(string newLabel) {
            try {
                if (!_isDisplayUpdating) { _selectedButtonGrid._gridLabel = newLabel; }
                if (_mvrScript._uiActivated) {
                    _mvrScript.RefreshUIButtonGrid();
                    _mvrScript.RefreshRightUIDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void SwitchSelectedButton(string newButton) {
            try {
                _buttonDisplay._selectedButton = _selectedButtonGrid._buttonList[int.Parse(newButton) - 1];
                if (_buttonDisplay._buttonSkinDisplay != null) _buttonDisplay._buttonSkinDisplay._selectedButtonSkin = _selectedButtonGrid._buttonList[int.Parse(newButton) - 1]._buttonSkin;
                if (_buttonDisplay._pluginVariablesDisplayOn != null)
                {
                    _buttonDisplay._pluginVariablesDisplayOn._selectButtonPluginVariables = _selectedButtonGrid._buttonList[int.Parse(newButton) - 1]._pluginVariablesButtonOn;
                    _buttonDisplay._pluginVariablesDisplayOn._pluginVariableSelectorJSON.val = "Select Parameter...";
                }
                if (_buttonDisplay._pluginVariablesDisplayOff != null)
                {
                    _buttonDisplay._pluginVariablesDisplayOff._selectButtonPluginVariables = _selectedButtonGrid._buttonList[int.Parse(newButton) - 1]._pluginVariablesButtonOff;
                    _buttonDisplay._pluginVariablesDisplayOff._pluginVariableSelectorJSON.val = "Select Parameter...";
                }
                RefreshDisplay();
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void SwitchButtonSize(string newButtonSize) {
            try {
                if (!_isDisplayUpdating) { _selectedButtonGrid._buttonSize = newButtonSize; }
                if (_mvrScript._uiActivated) {
                    _mvrScript.RefreshUIButtonGrid();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void SwitchButtonLayout(string newLayout) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButtonGrid._gridColumns = int.Parse(newLayout.Substring(0, 1));
                    _selectedButtonGrid._gridRows = int.Parse(newLayout.Substring(2, 1));
                    _selectedButtonGrid.UpdateGridButtons();
                    _selectedButtonGrid.RecalcActiveButtonCount();
                    _selectedButtonGrid.RecalcGazeSelections();
                    RefreshDisplayValues();
                }
                if (_mvrScript._uiActivated) {
                    _mvrScript.RefreshUIButtonGrid();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void RefreshDisplay() {
            CloseAllPopup(null);
            if (_buttonLayoutPopup != null) { _mvrScript.RemovePopup(_buttonLayoutPopup); }
            if (_buttonSelectPopup != null) { _mvrScript.RemovePopup(_buttonSelectPopup); }
            if (_buttonSizePopup != null) { _mvrScript.RemovePopup(_buttonSizePopup); }
            if (_gridLabelTextField != null) { _mvrScript.RemoveTextField(_gridLabelTextField); }
            if (_spacer1 != null) { _mvrScript.RemoveSpacer(_spacer1); }
            List<string> test = new List<string>();
            test.Add("Temp");
            PatreonFeatures.GridLayoutFeature(test);
            if (_mvrScript._displayScreenLabelsJSON.val && Constants._patreonContentEnabled && test != null && test.Count == 7) _gridLabelTextField = _mvrScript.CreateTextInput(_gridLabelJSON, true);
            _buttonLayoutPopup = _mvrScript.CreateSetupPopup(_buttonLayoutJSON, true, 500f, ButtonLayoutPopupOpen, "Screen " + _mvrScript._buttonGridSelectorJSON.val + " Layout");
            _buttonSizePopup = _mvrScript.CreateSetupPopup(_buttonSizeJSON, true, 500f, ButtonSizePopupOpen, "Screen " + _mvrScript._buttonGridSelectorJSON.val + " Button Size");
            _spacer1 = _mvrScript.CreateSpacer(true);
            _spacer1.height = 10;
            _buttonSelectPopup = _mvrScript.CreateSetupPopup(_buttonSelectJSON, true, 500f, RefreshButtonSelectionChoices, "Button Selector");
            _buttonSelectPopup.labelTextColor = Color.blue;

            RefreshDisplayValues();
            _buttonDisplay.RefreshDisplay();
        }
        public void ButtonLayoutPopupOpen() { _mvrScript.CloseAllRightUIPopups(_buttonLayoutPopup); }
        public void ButtonSizePopupOpen() { _mvrScript.CloseAllRightUIPopups(_buttonSizePopup); }
        public void CloseAllPopup(UIDynamicPopup exception) {
            if (_buttonLayoutPopup != null && _buttonLayoutPopup != exception) { _buttonLayoutPopup.popup.visible = false; }
            if (_buttonSelectPopup != null && _buttonSelectPopup != exception) { _buttonSelectPopup.popup.visible = false; }
            if (_buttonSizePopup != null && _buttonSizePopup != exception) { _buttonSizePopup.popup.visible = false; }

            _buttonDisplay.CloseAllPopup(exception);
        }
        public void RefreshDisplayValues() {
            _isDisplayUpdating = true;
            int cols = _selectedButtonGrid._gridColumns;
            int rows = _selectedButtonGrid._gridRows;
            if (_selectedButtonGrid._gridLabel == "") _gridLabelJSON.val = "<Screen " + _mvrScript._buttonGridSelectorJSON.val + " Label>";
            else _gridLabelJSON.val = _selectedButtonGrid._gridLabel;
            _buttonSizeJSON.val = _selectedButtonGrid._buttonSize;
            _buttonLayoutJSON.val = cols.ToString() + " x " + rows.ToString();
            if (int.Parse(_buttonSelectJSON.val) > rows * cols) { _buttonSelectJSON.val = (rows * cols).ToString(); }

            _isDisplayUpdating = false;
        }
    }

    public class Button {
        public int _buttonRef;
        public string _buttonCategory = "";
        public string _buttonType = "";
        private int _buttonState = 1;
        public int _maxButtonStates = 1;
        public string _targetCategory = "";
        public string _targetName = "";
        public string _targetParamNameOn = "";
        public string _targetParamNameOff = "";
        public bool _suppressClothingLoad = false;
        public bool _suppressPersonScaleLoad = false;
        public bool _autoLabel = true;
        public string _pluginType = "";
        public bool _pluginForceMergeLoad = true;
        public bool _closeGameUIonPlayMode = true;
        public bool _openGameUIonEditMode = true;

        public string _presetFilePath = "";
        public string _buttonLabelOn = "";
        public string _buttonLabelOff = "";
        public Texture2D _buttonTexture;
        public bool _useButtonTexture = true;
        public ButtonSkin _buttonSkin;

        public PluginVariables _pluginVariablesButtonOn;
        public PluginVariables _pluginVariablesButtonOff;

        public List<string> _mergeClothingGeometryIDs;

        private UIAssist _mvrScript;

        public List<ButtonPluginLoad> _pluginLoadList;
        public Dictionary<string, int> _atomButtonStates;
        public List<string> _tagsToRemoveOnMerge;
        public Button(int buttonRef, UIAssist mvr)
        {
            try
            {
                _buttonRef = buttonRef;
                _buttonCategory = "Miscellaneous";
                _buttonType = "Blank";
                _targetCategory = "Gaze Selected Atom";
                _mvrScript = mvr;
                _pluginLoadList = new List<ButtonPluginLoad>();
                ButtonPluginLoad tempPIL = new ButtonPluginLoad();
                _pluginLoadList.Add(tempPIL);
                _atomButtonStates = new Dictionary<string, int>();

                _tagsToRemoveOnMerge = new List<string>();
                if (Constants._patreonContentEnabled) _buttonSkin = new ButtonSkin(_maxButtonStates, mvr);
                _pluginVariablesButtonOn = new PluginVariables(this);
                _pluginVariablesButtonOff = new PluginVariables(this);

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        public bool IsAtomTargetable()
        {
            return (_buttonType != "Suppress Scale Load" && _buttonType != "Freeze Motion/Sound" && _buttonType != "Toggle Play/Edit Mode" && _buttonType != "Blank");
        }

        public void CopyFrom(Button bn) {
            if (bn != null) {
                _buttonCategory = bn._buttonCategory;
                _buttonType = bn._buttonType;
                _buttonState = 1;
                _maxButtonStates = bn._maxButtonStates;
                _targetCategory = bn._targetCategory;
                _targetName = bn._targetName;
                _targetParamNameOn = bn._targetParamNameOn;
                _targetParamNameOff = bn._targetParamNameOff;
                _suppressClothingLoad = bn._suppressClothingLoad;
                _suppressPersonScaleLoad = bn._suppressPersonScaleLoad;
                _autoLabel = bn._autoLabel;
                _pluginType = bn._pluginType;
                _pluginForceMergeLoad = bn._pluginForceMergeLoad;
                _presetFilePath = bn._presetFilePath;
                _buttonLabelOn = bn._buttonLabelOn;
                _buttonLabelOff = bn._buttonLabelOff;
                _buttonTexture = bn._buttonTexture;
                _useButtonTexture = bn._useButtonTexture;
                _closeGameUIonPlayMode = bn._closeGameUIonPlayMode;
                _openGameUIonEditMode = bn._openGameUIonEditMode;

                _buttonSkin.CopyFrom(bn._buttonSkin);
                _mergeClothingGeometryIDs = null;
                _atomButtonStates = new Dictionary<string, int>();
                _tagsToRemoveOnMerge = new List<string>();
                foreach (string tag in bn._tagsToRemoveOnMerge) _tagsToRemoveOnMerge.Add(tag);
                _pluginLoadList.Clear();

                for (int i = 0; i < bn._pluginLoadList.Count; i++)
                {
                    ButtonPluginLoad newPluginLoad = new ButtonPluginLoad();
                    newPluginLoad.CopyFrom(bn._pluginLoadList[i]);
                    _pluginLoadList.Add(newPluginLoad);
                }
                _pluginVariablesButtonOn.CopyFrom(bn._pluginVariablesButtonOn);
                _pluginVariablesButtonOff.CopyFrom(bn._pluginVariablesButtonOff);
            }

        }

        public string GetManualLabel(int buttonState, string targetAtomName) {
            string buttonLabel = "";
            if (buttonState == 1) buttonLabel = _buttonLabelOn;
            else buttonLabel = _buttonLabelOff;
            if (_targetCategory == "Gaze Selected Atom" && buttonLabel.ToUpper().Contains("<GAZE>")) {
                int index = buttonLabel.ToUpper().IndexOf("<GAZE>");
                if (targetAtomName == "") buttonLabel = buttonLabel.Substring(0, index) + "<TARGET>" + buttonLabel.Substring(index + 6);
                else buttonLabel = buttonLabel.Substring(0, index) + targetAtomName + buttonLabel.Substring(index + 6);
            }

            return (buttonLabel);
        }
        public string GetAutoLabel(int buttonState, string targetAtomName) {
            string buttonLabel = "";

            if (_buttonType=="Freeze Motion/Sound")
            {
                if (buttonState==1) return ("Feeze Motion / Sound");
                else return ("Unfeeze Motion / Sound");
            }
            else if (_buttonType== "Toggle Play/Edit Mode")
            {
                if (buttonState == 1) return ("Edit Mode");
                else return ("Play Mode");
            }
            if (_buttonCategory == "Presets" || _buttonCategory == "Clothing" || _buttonCategory == "Appearance")
            {
                if (_buttonType.EndsWith("Preset")) {
                    buttonLabel = _buttonType.Substring(0, _buttonType.Length - 7);
                    if (buttonLabel.EndsWith("App")) buttonLabel = buttonLabel + "earance";
                    string presetName = " <PRESET>";
                    if (_presetFilePath != "") {
                        int filenameIndex = _presetFilePath.LastIndexOf('/');
                        presetName = " '" + _presetFilePath.Substring(filenameIndex + 8, _presetFilePath.Length - (filenameIndex + 8) - 4) + "'";
                    }
                    buttonLabel = buttonLabel + presetName;
                }
                else buttonLabel = _buttonType;
                if (buttonLabel.StartsWith("Undress") && buttonState == 1) { buttonLabel = "Set Undress for " + buttonLabel.Substring(8); }
                if (buttonLabel.StartsWith("Undress") && buttonState == 2) { buttonLabel = "Unset Undress for " + buttonLabel.Substring(8); }
                if (buttonLabel.StartsWith("Merge") && buttonState == 2) { buttonLabel = "Unmerge " + buttonLabel.Substring(6); }
                if (_buttonType == "Suppress Scale Load" && buttonState == 2) buttonLabel = "Unsuppress Scale Load";

            }
            else if (_buttonCategory == "Plugins") {
                string targetParamOn = _targetParamNameOn;
                string targetParamOff = _targetParamNameOff;
                if (targetParamOn == "No Actions available") targetParamOn = "<ACTION>";
                if (targetParamOff == "No Actions available") targetParamOff = "<ACTION>";
                string pluginType = _pluginType;
                if (pluginType.Contains(".")) {
                    pluginType = pluginType.Substring(pluginType.IndexOf('.') + 1);
                }

                if (_buttonType == "Load Plugin(s)") buttonLabel = "Load Plugins";
                else if (_buttonType == "Plugin Action") buttonLabel = pluginType + ": " + targetParamOn;
                else if (_buttonType == "Plugin Action Toggle" && buttonState == 1) buttonLabel = pluginType + ": " + targetParamOn;
                else if (_buttonType == "Plugin Action Toggle" && buttonState == 2) buttonLabel = pluginType + ": " + targetParamOff;
                else if (_buttonType == "Plugin Bool Toggle" && buttonState == 1) buttonLabel = "Set " + pluginType + ": " + targetParamOn + " to TRUE";
                else if (_buttonType == "Plugin Bool Toggle" && buttonState == 2) buttonLabel = "Set " + pluginType + ": " + targetParamOn + " to FALSE";
                else if (_buttonType == "Set Plugin Bool True") buttonLabel = "Set " + pluginType + ": " + targetParamOn + " to TRUE";
                else if (_buttonType == "Set Plugin Bool False") buttonLabel = "Set " + pluginType + ": " + targetParamOn + " to FALSE";
                else if (_buttonType == "Unload All Plugins") buttonLabel = "Unload all Plugins";
            }

            string preposition = " ";
            if (buttonLabel.StartsWith("Load") || buttonLabel.StartsWith("Merge") || buttonLabel.StartsWith("Unmerge")) preposition = " to ";
            if (buttonLabel.StartsWith("Remove") || buttonLabel.StartsWith("Undress") || buttonLabel.StartsWith("Set") || buttonLabel.StartsWith("Unset")) preposition = " from ";
            if (buttonLabel.StartsWith("Reset") || buttonLabel.StartsWith("Set") || buttonLabel.EndsWith("Scale")) preposition = " on ";
            if (_buttonCategory == "Plugins") preposition = " for ";

            if (_buttonType == "Suppress Scale Load")  buttonLabel = buttonLabel;
            else if ((_targetCategory == "Gaze Selected Atom" && targetAtomName == "") || _targetName == "<None available>") buttonLabel = buttonLabel + preposition + "<TARGET>";
            else if (_targetCategory == "Gaze Selected Atom") buttonLabel = buttonLabel + preposition + "'" + targetAtomName + "'";
            else buttonLabel = buttonLabel + preposition + "'" + _targetName + "'";

            return (buttonLabel);
        }

        public JSONClass GetJSON()
        {
            JSONClass node = new JSONClass();
            node["buttonCategory"] = _buttonCategory;
            node["buttonType"] = _buttonType;
            if (_buttonType != "Blank") {
                node["targetCategory"] = _targetCategory;
                node["targetName"] = _targetName;
                node["targetParamNameOn"] = _targetParamNameOn;
                node["targetParamNameOff"] = _targetParamNameOff;
                node["pluginType"] = _pluginType;
                node["presetFilePath"] = _presetFilePath;
                node["buttonLabelOn"] = _buttonLabelOn;
                node["buttonLabelOff"] = _buttonLabelOff;

                node["suppressClothingLoad"].AsBool = _suppressClothingLoad;
                node["suppressPersonScaleLoad"].AsBool = _suppressPersonScaleLoad;
                node["autoLabel"].AsBool = _autoLabel;
                node["useButtonTexture"].AsBool = _useButtonTexture;
                node["pluginForceMergeLoad"].AsBool = _pluginForceMergeLoad;
                node["closeGameUIonPlayMode"].AsBool = _closeGameUIonPlayMode;
                node["openGameUIonEditMode"].AsBool = _openGameUIonEditMode;

                node["maxButtonStates"].AsInt = _maxButtonStates;

                if (Constants._patreonContentEnabled) node["ButtonSkin"] = _buttonSkin.GetJSON();

                node["pluginVariablesOn"] = _pluginVariablesButtonOn.GetJSON();
                node["pluginVariablesOff"] = _pluginVariablesButtonOff.GetJSON();

                if (_buttonType == "Load Plugin(s)") {
                    JSONArray pluginLoadArray = new JSONArray();
                    _pluginLoadList.ForEach((pluginLoad) =>
                    {
                        if (pluginLoad._filePath != "") pluginLoadArray.Add(pluginLoad.GetJSON());
                    });

                    node["pluginLoads"] = pluginLoadArray;
                }
                if (_tagsToRemoveOnMerge.Count > 0)
                {
                    JSONArray tagsToRemoveArray = new JSONArray();
                    foreach (string tag in _tagsToRemoveOnMerge) node["tagsToRemove"][-1] = tag;

                }
            }
            return node;
        }
        public void LoadJSON(JSONClass buttonJSON) {
            if (buttonJSON["buttonCategory"] != null) _buttonCategory = buttonJSON["buttonCategory"];
            if (buttonJSON["buttonType"] != null) _buttonType = buttonJSON["buttonType"];
            if (buttonJSON["targetCategory"] != null) _targetCategory = buttonJSON["targetCategory"];
            if (buttonJSON["targetName"] != null) _targetName = buttonJSON["targetName"];
            if (buttonJSON["targetParamNameOn"] != null) _targetParamNameOn = buttonJSON["targetParamNameOn"];
            if (buttonJSON["targetParamNameOff"] != null) _targetParamNameOff = buttonJSON["targetParamNameOff"];
            if (buttonJSON["pluginType"] != null) _pluginType = buttonJSON["pluginType"];
            if (buttonJSON["presetFilePath"] != null) _presetFilePath = buttonJSON["presetFilePath"];
            if (buttonJSON["buttonLabelOn"] != null) _buttonLabelOn = buttonJSON["buttonLabelOn"];
            if (buttonJSON["buttonLabelOff"] != null) _buttonLabelOff = buttonJSON["buttonLabelOff"];

            if (buttonJSON["suppressClothingLoad"] != null) _suppressClothingLoad = buttonJSON["suppressClothingLoad"].AsBool;
            if (buttonJSON["suppressPersonScaleLoad"] != null) _suppressPersonScaleLoad = buttonJSON["suppressPersonScaleLoad"].AsBool;
            if (buttonJSON["autoLabel"] != null) _autoLabel = buttonJSON["autoLabel"].AsBool;
            if (buttonJSON["useButtonTexture"] != null) _useButtonTexture = buttonJSON["useButtonTexture"].AsBool;
            if (buttonJSON["pluginForceMergeLoad"] != null) _pluginForceMergeLoad = buttonJSON["pluginForceMergeLoad"].AsBool;
            if (buttonJSON["closeGameUIonPlayMode"] != null) _closeGameUIonPlayMode = buttonJSON["closeGameUIonPlayMode"].AsBool;
            if (buttonJSON["openGameUIonEditMode"] != null) _openGameUIonEditMode = buttonJSON["openGameUIonEditMode"].AsBool;

            if (buttonJSON["maxButtonStates"] != null) _maxButtonStates = buttonJSON["maxButtonStates"].AsInt;
            if (buttonJSON["ButtonSkin"] != null && Constants._patreonContentEnabled) _buttonSkin.LoadJSON((JSONClass)buttonJSON["ButtonSkin"]);
            if (buttonJSON["pluginVariablesOn"] != null) _pluginVariablesButtonOn.LoadJSON((JSONClass)buttonJSON["pluginVariablesOn"]);
            if (buttonJSON["pluginVariablesOff"] != null) _pluginVariablesButtonOff.LoadJSON((JSONClass)buttonJSON["pluginVariablesOff"]);

            if (buttonJSON["pluginLoads"] != null)
            {
                JSONArray pluginLoadsArrayJSON = buttonJSON["pluginLoads"].AsArray;
                for (int i = 0; i < pluginLoadsArrayJSON.Count; i++) {
                    JSONClass pluginLoadJSON = pluginLoadsArrayJSON[i].AsObject;
                    ButtonPluginLoad pluginLoad;
                    if (i >= _pluginLoadList.Count) { pluginLoad = new ButtonPluginLoad(); _pluginLoadList.Add(pluginLoad); }
                    else { pluginLoad = _pluginLoadList[i]; }

                    pluginLoad.LoadJSON(pluginLoadJSON);
                }
            }
            if (_presetFilePath != "") QueueLoadTexture(_presetFilePath.Substring(0, _presetFilePath.Length - 3) + "jpg"); _mergeClothingGeometryIDs = null;
            if (buttonJSON["tagsToRemove"] != null)
            {
                _tagsToRemoveOnMerge = new List<string>();
                JSONArray tagsToRemoveArray = buttonJSON["tagsToRemove"].AsArray;
                for (int i = 0; i < tagsToRemoveArray.Count; i++)
                {
                    _tagsToRemoveOnMerge.Add(tagsToRemoveArray[i]);
                }
            }

        }
        public void ToggleButtonState(string atomName) {
            if (_maxButtonStates == 2) {
                if (_targetCategory == "Gaze Selected Atom" && atomName!=null) {
                    int currentState = 0;
                    if (_atomButtonStates.TryGetValue(atomName, out currentState)) {
                        if (currentState == 1) { _atomButtonStates[atomName] = 2; }
                        else { _atomButtonStates[atomName] = 1; }
                    }
                    else {
                        _atomButtonStates.Add(atomName, 2);
                    }
                }
                else {
                    if (_buttonState == 1) { _buttonState = 2; }
                    else { _buttonState = 1; }
                }
                _mvrScript.RefreshUIButtonGrid();
            }
            else _buttonState = 1;
        }
        public int GetButtonState(string atomName) {
            int buttonState = 1;
            if (_buttonType=="Suppress Scale Load")
            {
                if (_mvrScript._suppressScaleLoad) buttonState = 2;
                else buttonState = 1;
                
            }
            else if (_buttonType == "Freeze Motion/Sound")
            {
                if (SuperController.singleton.freezeAnimation) _buttonState = 2;
                else _buttonState = 1;
                buttonState = _buttonState;
            }
            else if (_buttonType == "Toggle Play/Edit Mode")
            {
                if (SuperController.singleton.gameMode== SuperController.GameMode.Edit) _buttonState = 2;
                else _buttonState = 1;
                return(_buttonState);
            }
            else if (_targetCategory == "Gaze Selected Atom" && atomName != "") {
                int currentState = 1;
                if (_atomButtonStates.TryGetValue(atomName, out currentState)) {
                    buttonState = currentState;
                }
                else {
                    _atomButtonStates.Add(atomName, 1);
                    buttonState = 1;
                }
            }
            else {
                buttonState = _buttonState;
            }
           
            if (_buttonType == "Merge Clothing Preset" && _presetFilePath != "") {
                List<string> atomNames = new List<string>();
                if (_targetCategory == "Atom Group") {
                    foreach (Atom atomFromGroup in SuperController.singleton.GetAtoms()) {
                        bool isMale = false;
                        if (atomFromGroup.type == "Person" && (atomFromGroup.GetComponentInChildren<DAZCharacter>().name.StartsWith("male") || atomFromGroup.GetComponentInChildren<DAZCharacter>().name.StartsWith("jarlee") || atomFromGroup.GetComponentInChildren<DAZCharacter>().name.StartsWith("jarjulian"))) isMale = true;

                        if (atomFromGroup.name != "CoreControl" && atomFromGroup.name != "[CameraRig]") {
                            switch (_targetName) {
                                case "All Atoms":
                                    atomNames.Add(atomFromGroup.name);
                                    break;
                                case "All People":
                                    if (atomFromGroup.type == "Person") { atomNames.Add(atomFromGroup.name); }
                                    break;
                                case "All Females":
                                    if (atomFromGroup.type == "Person" && !isMale) { atomNames.Add(atomFromGroup.name); }
                                    break;
                                case "All Males":
                                    if (atomFromGroup.type == "Person" && isMale) { atomNames.Add(atomFromGroup.name); }
                                    break;
                            }
                        }
                    }
                }
                else if (_targetCategory == "Specific Atom" && SuperController.singleton.GetAtomUIDs().Contains(_targetName)) atomNames.Add(_targetName);
                else if (_targetCategory == "Gaze Selected Atom" && SuperController.singleton.GetAtomUIDs().Contains(atomName)) atomNames.Add(atomName);
                bool allAtomsClothingItemInExpectedState = false;
                foreach (string atomNameForMergeAction in atomNames) {
                    Atom atom = SuperController.singleton.GetAtomByUid(atomNameForMergeAction);
                    JSONStorable atomGeometryStorable = atom.GetStorableByID("geometry");
                    if (_mergeClothingGeometryIDs == null) { _mergeClothingGeometryIDs = new List<string>();
                        PatreonFeatures.PopulateMergeClothingGeometryIDs(_mvrScript, _mergeClothingGeometryIDs, _presetFilePath, atom);
                    }
                    bool clothingItemInExpectedState = false;
                    foreach (string clothingItem in _mergeClothingGeometryIDs) {
                        JSONStorableBool clothingItemStateJSON = atomGeometryStorable.GetBoolJSONParam(clothingItem);
                        if (clothingItemStateJSON != null)
                        {
                            bool clothingItemCurrentState = clothingItemStateJSON.val;
                            if (buttonState == 2 && clothingItemCurrentState) clothingItemInExpectedState = true;
                            if (buttonState == 1 && !clothingItemCurrentState) clothingItemInExpectedState = true;
                            if (clothingItemInExpectedState) break;
                        }
                        else { clothingItemInExpectedState = true; }
                    }
                    if (clothingItemInExpectedState) { allAtomsClothingItemInExpectedState = true; break; }
                }
                if (!allAtomsClothingItemInExpectedState) {
                    ToggleButtonState(atomName);
                    if (buttonState == 1) buttonState = 2;
                    else buttonState = 1;

                }
            }


            return (buttonState);
        }
        public List<string> GetButtonCategoryChoices() {
            List<string> buttonCategoryChoices = new List<string>();
            buttonCategoryChoices.Add("Appearance");
            buttonCategoryChoices.Add("Clothing");
            buttonCategoryChoices.Add("Presets");
            buttonCategoryChoices.Add("Plugins");
            buttonCategoryChoices.Add("Miscellaneous");
            return (buttonCategoryChoices);
        }
        public List<string> GetButtonTypeChoices() {
            List<string> buttonTypeChoices = new List<string>();
            switch (_buttonCategory) {
                case "Appearance":
                    buttonTypeChoices.Add("Load App Preset");
                    buttonTypeChoices.Add("Reset Scale");
                    buttonTypeChoices.Add("Reset Appearance");
 //                   buttonTypeChoices.Add("Set Scale");
 //                   buttonTypeChoices.Add("Increment Scale");
 //                   buttonTypeChoices.Add("Decrement Scale");
                    if (Constants._patreonContentEnabled)buttonTypeChoices.Add("Suppress Scale Load");
                    break;
                case "Clothing":
                    buttonTypeChoices.Add("Remove All Clothing");
                    buttonTypeChoices.Add("Undress All Clothing");
                    buttonTypeChoices.Add("Reset Sim All Clothing");
                    buttonTypeChoices.Add("Load Clothing Preset");
                    if (Constants._patreonContentEnabled) {
                        buttonTypeChoices.Add("Remove Clothing Preset");
                        buttonTypeChoices.Add("Merge Clothing Preset");
                        buttonTypeChoices.Add("Undress Cloth Preset");
                        buttonTypeChoices.Add("Reset Sim Cloth Preset");
                    }
                    break;
                case "Presets":
                    buttonTypeChoices.Add("Load App Preset");
                    buttonTypeChoices.Add("Load Clothing Preset");
                    buttonTypeChoices.Add("Load Pose Preset");
                    buttonTypeChoices.Add("Load Hair Preset");
                    buttonTypeChoices.Add("Load Glute Preset");
                    buttonTypeChoices.Add("Load Breast Preset");
                    buttonTypeChoices.Add("Load Morph Preset");
                    buttonTypeChoices.Add("Load Skin Preset");
                    buttonTypeChoices.Add("Load General Preset");
                    break;
                case "Plugins":
                    buttonTypeChoices.Add("Load Plugin(s)");
                    buttonTypeChoices.Add("Plugin Action");
                    buttonTypeChoices.Add("Plugin Action Toggle");
                    buttonTypeChoices.Add("Plugin Bool Toggle");
                    buttonTypeChoices.Add("Set Plugin Bool True");
                    buttonTypeChoices.Add("Set Plugin Bool False");
                    buttonTypeChoices.Add("Unload All Plugins");
                    break;
                case "Miscellaneous":
                    
                    buttonTypeChoices.Add("Freeze Motion/Sound");
                    buttonTypeChoices.Add("Toggle Play/Edit Mode");
                    buttonTypeChoices.Add("Blank");
                    break;
            }
            return (buttonTypeChoices);
        }
        public List<string> GetAtomCategoryChoices() {
            List<string> atomCategoryChoices = new List<string>();
            atomCategoryChoices.Add("Gaze Selected Atom");
            atomCategoryChoices.Add("Atom Group");
            atomCategoryChoices.Add("Specific Atom");
            return (atomCategoryChoices);
        }
        public List<string> GetTargetChoices() {
            List<string> atomChoices = new List<string>();
            switch (_targetCategory) {
                case "Gaze Selected Atom":
                    atomChoices.Add("Last viewed Person"); atomChoices.Add("Last viewed Female"); atomChoices.Add("Last viewed Male");
                    if (_buttonCategory == "Plugins" || _buttonType == "Load General Preset") { atomChoices.Add("Last viewed Atom"); }
                    break;
                case "Atom Group":
                    atomChoices.Add("All Females"); atomChoices.Add("All Males"); atomChoices.Add("All People");
                    if (_buttonCategory == "Plugins" || _buttonType == "Load General Preset") { atomChoices.Add("All Atoms"); }
                    break;
                case "Specific Atom":
                    foreach (string atomUID in SuperController.singleton.GetAtomUIDs()) {
                        Atom atom = SuperController.singleton.GetAtomByUid(atomUID);
                        if (atomUID != "[CameraRig]" && (atomUID != "CoreControl" || _buttonType != "Load Plugin(s)") && ((_buttonCategory != "Clothing" && _buttonType.Substring(_buttonType.Length - 6, 6) != "Preset") || atom.type == "Person" || _buttonType == "Load General Preset")) { atomChoices.Add(atomUID); }
                    }
                    break;
            }
            if (atomChoices.Count == 0) { atomChoices.Add("<None available>"); }
            return (atomChoices);
        }
        public List<string> GetPluginTypeChoices() {
            List<string> popupChoices = new List<string>();

            foreach (Atom atom in SuperController.singleton.GetAtoms()) {
                foreach (string storableName in atom.GetStorableIDs()) {
                    if (storableName.StartsWith("plugin#")) {
                        string pluginType = storableName.Substring(storableName.IndexOf('_') + 1);
                        if (!popupChoices.Contains(pluginType)) { popupChoices.Add(pluginType); }
                    }
                }
            }

            if (popupChoices.Count == 0) { popupChoices.Add("No plugins in Scene"); }
            popupChoices.Sort();
            return (popupChoices);
        }
        public string GetVarTypeFromPlugin(string varName)
        {
            string varType = "NONE";
            bool foundVar = false;

            foreach (Atom atom in SuperController.singleton.GetAtoms())
            {
                foreach (string storableName in atom.GetStorableIDs())
                {               
                    if (storableName.StartsWith("plugin#")&& storableName.EndsWith(_pluginType))
                    {
                        JSONStorable pluginStorable = atom.GetStorableByID(storableName);
                        if (pluginStorable.GetBoolParamNames().Contains(varName)) varType = "BOOL";
                        else if (pluginStorable.GetFloatParamNames().Contains(varName)) varType = "FLOAT";
                        else if (pluginStorable.GetStringParamNames().Contains(varName)) varType = "STRING";
                        else if (pluginStorable.GetStringChooserParamNames().Contains(varName)) varType = "STRINGCHOOSER";
                        else if (pluginStorable.GetUrlParamNames().Contains(varName)) varType = "URL";
                        else if (pluginStorable.GetColorParamNames().Contains(varName)) varType = "COLOR";
                        if (varType != "NONE")
                        {
                            foundVar = true;
                            break;
                        }
                    }
                }
                if (foundVar) break;
            }
            return (varType);
        }
        public float GetMinFloatValue(string varName)
        {
            float minFloat = 100000000000f;

            foreach (Atom atom in SuperController.singleton.GetAtoms())
            {
                foreach (string storableName in atom.GetStorableIDs())
                {
                    if (storableName.StartsWith("plugin#") && storableName.EndsWith(_pluginType))
                    {
                        JSONStorable pluginStorable = atom.GetStorableByID(storableName);

                        if (pluginStorable.GetFloatParamNames().Contains(varName))
                        {
                            JSONStorableFloat floatJSON = pluginStorable.GetFloatJSONParam(varName);
                            if (floatJSON.min < minFloat) minFloat = floatJSON.min;
                            
                        }

                    }
                }
            }
            if (minFloat == 100000000000f) minFloat = 0f;
            return (minFloat);
        }
        public float GetMaxFloatValue(string varName)
        {
            float maxFloat = -100000000000f;

            foreach (Atom atom in SuperController.singleton.GetAtoms())
            {
                foreach (string storableName in atom.GetStorableIDs())
                {
                    if (storableName.StartsWith("plugin#") && storableName.EndsWith(_pluginType))
                    {
                        JSONStorable pluginStorable = atom.GetStorableByID(storableName);

                        if (pluginStorable.GetFloatParamNames().Contains(varName))
                        {
                            JSONStorableFloat floatJSON = pluginStorable.GetFloatJSONParam(varName);
                            if (floatJSON.max > maxFloat) maxFloat = floatJSON.max;

                        }
                    }
                }
            }
            if (maxFloat == -100000000000f) maxFloat = 1f;
            return (maxFloat);
        }
        public List<string> GetStringChooserChoices(string varName)
        {
            List<string> choices = new List<string>();

            foreach (Atom atom in SuperController.singleton.GetAtoms())
            {
                foreach (string storableName in atom.GetStorableIDs())
                {
                    if (storableName.StartsWith("plugin#") && storableName.EndsWith(_pluginType))
                    {
                        JSONStorable pluginStorable = atom.GetStorableByID(storableName);
                        if (pluginStorable.GetStringChooserParamNames().Contains(varName))
                        {
                            List<string> pluginSCChoices = pluginStorable.GetStringChooserJSONParamChoices(varName);
                            if (pluginSCChoices != null)
                            {
                                foreach (string choice in pluginSCChoices)
                                {
                                    if (!choices.Contains(choice)) choices.Add(choice);
                                }
                            }
                        }
                    }
                }
            }
            return (choices);
        }

        public void UpdateButtonCategory(string category) {
            _buttonCategory = category;
            if (!GetButtonTypeChoices().Contains(_buttonType)) { UpdateButtonType(GetButtonTypeChoices()[0]); }
            if (!GetTargetChoices().Contains(_targetName)) { UpdateAtom(GetTargetChoices()[0]); }
        }
        public void UpdateButtonType(string type) {
            _buttonType = type;

            switch (_buttonType) {
                case "Plugin Action Toggle":
                case "Undress All Clothing":
                case "Merge Clothing Preset":
                case "Undress Cloth Preset":
                case "Plugin Bool Toggle":
                case "Suppress Scale Load":
                case "Freeze Motion/Sound":
                case "Toggle Play/Edit Mode":
                    _maxButtonStates = 2;
                    if (_buttonSkin != null) _buttonSkin._maxButtonStates = 2;
                    break;
                default:
                    _maxButtonStates = 1;
                    if (_buttonSkin != null) _buttonSkin._maxButtonStates = 1;
                    break;
            }
            if (_buttonType.Length > 6 && _buttonType.Substring(_buttonType.Length - 6, 6) == "Preset" && _buttonCategory != "Clothing") {
                _presetFilePath = "";
                _buttonTexture = null;
                _mergeClothingGeometryIDs = null;
            }

            List<string> targetChoices = new List<string>();
            targetChoices = GetTargetChoices();
            if (!targetChoices.Contains(_targetName)) {
                if (targetChoices.Contains(_mvrScript._defaultButtonSkin._defaultGazeTarget)) _targetName = _mvrScript._defaultButtonSkin._defaultGazeTarget;
                else _targetName = targetChoices[0];
            }

            if (_buttonCategory == "Plugins" && _buttonType != "Load Plugin(s)") {
                List<string> pluginTypeChoices = new List<string>();
                pluginTypeChoices = GetPluginTypeChoices();
                if (!pluginTypeChoices.Contains(_pluginType)) { _pluginType = pluginTypeChoices[0]; }
                List<string> targetParamNameChoices = new List<string>();
                if (_buttonType.Contains("Bool")) { targetParamNameChoices = GetTargetPluginParamNameChoices("BOOL"); }
                else { targetParamNameChoices = GetTargetPluginParamNameChoices("ACTION"); }
                if (!targetParamNameChoices.Contains(_targetParamNameOn)) { _targetParamNameOn = targetParamNameChoices[0]; }
                if (_buttonType.Contains("Toggle")) {
                    if (!targetParamNameChoices.Contains(_targetParamNameOff)) { _targetParamNameOff = targetParamNameChoices[0]; }
                }
                else { _targetParamNameOff = ""; }
            }
            if (_buttonType == "Blank") _targetName = "";
        }
        public void UpdateAtomCategory(string category) {
            _targetCategory = category;
            if (!GetTargetChoices().Contains(_targetName)) { UpdateAtom(GetTargetChoices()[0]); }
        }
        public void UpdateAtom(string atom) {
            _targetName = atom;
        }
        public void UpdateTagItemToRemove(string tag)
        {
            if (_tagsToRemoveOnMerge.Contains(tag)) _tagsToRemoveOnMerge.Remove(tag);
            else _tagsToRemoveOnMerge.Add(tag);
        }

        public void SetPresetFile(string fileName) {
            _presetFilePath = fileName;
            _mergeClothingGeometryIDs = null;
            QueueLoadTexture(fileName.Substring(0, fileName.Length - 3) + "jpg");
        }

        public List<string> GetTargetPluginParamNameChoices(string paramType) {
            List<string> choices = new List<string>();
            bool pluginOfTypeFound = false;

            foreach (Atom atom in SuperController.singleton.GetAtoms()) {
                foreach (string storableName in atom.GetStorableIDs()) {
                    if (storableName.EndsWith(_pluginType)) {
                        pluginOfTypeFound = true;
                        JSONStorable storable = atom.GetStorableByID(storableName);
                        if (paramType == "ACTION")
                        {
                            foreach (string actionName in storable.GetActionNames())
                            {
                                if (!UIAssist.VAMSaveRestoreActions.Contains(actionName) && !choices.Contains(actionName)) { choices.Add(actionName); }
                            }
                        }
                        else
                        {
                            foreach (string boolName in storable.GetBoolParamNames())
                            {
                                if (!choices.Contains(boolName)) { choices.Add(boolName); }
                            }
                            if (paramType == "ALL")
                            {
                                foreach (string floatName in storable.GetFloatParamNames())
                                {
                                    if (!choices.Contains(floatName)) { choices.Add(floatName); }
                                }
                                foreach (string stringName in storable.GetStringParamNames())
                                {
                                    if (!choices.Contains(stringName)) { choices.Add(stringName); }
                                }
                                foreach (string stringChooserName in storable.GetStringChooserParamNames())
                                {
                                    if (!choices.Contains(stringChooserName)) { choices.Add(stringChooserName); }
                                }
                                foreach (string urlName in storable.GetUrlParamNames())
                                {
 //                                   if (!choices.Contains(urlName)) { choices.Add(urlName); }
                                }
                                foreach (string colorName in storable.GetColorParamNames())
                                {
                                    if (!choices.Contains(colorName)) { choices.Add(colorName); }
                                }
                            }
                        }
                    }

                }
            }

            if (choices.Count < 1) {
                if (!pluginOfTypeFound) { choices.Add("No Plugin of type loaded"); }
                else {
                    if (paramType == "ACTION") { choices.Add("No Actions available"); }
                    else { choices.Add("No Bools available"); }
                }
            }
            choices.Sort();
            return (choices);
        }

        public JSONArray GetPluginDataByName(string atomName, string name) {
            Atom atom = SuperController.singleton.GetAtomByUid(atomName);
            JSONArray pluginData = new JSONArray();
            if (atom != null) {
                foreach (string storableName in atom.GetStorableIDs()) {

                    if (storableName.StartsWith(name)) {
                        JSONClass storableJSON = atom.GetStorableByID(storableName).GetJSON();
                        pluginData[-1] = storableJSON;

                    }
                }
            }
            return (pluginData);
        }
        public List<string> GetPluginRefsOfType(Atom atom, string type) {
            List<string> pluginRefs = new List<string>();
            MVRPluginManager manager = atom.GetStorableByID("PluginManager") as MVRPluginManager;
            if (manager != null) {
                JSONClass currentPlugins = null;
                currentPlugins = manager.GetJSON(true, true, true);
                if (currentPlugins != null) {
                    foreach (KeyValuePair<string, JSONNode> kvp in (JSONClass)currentPlugins["plugins"]) {
                        if (kvp.Value.ToString().Substring(kvp.Value.ToString().LastIndexOfAny(new char[] { '/', '\\' }) + 1) == type.Substring(type.LastIndexOfAny(new char[] { '/', '\\' }) + 1) + "\"") { pluginRefs.Add(kvp.Key); }
                    }
                }
            }
            return (pluginRefs);
        }
        public void QueueLoadTexture(string url)
        {
            if (ImageLoaderThreaded.singleton == null)
                return;
            if (string.IsNullOrEmpty(url))
                return;
            if (!FileManagerSecure.FileExists(url)) {
                _buttonTexture = null;
                _mvrScript.RefreshUIButtonGrid();
                return;
            }

            ImageLoaderThreaded.QueuedImage queuedImage = new ImageLoaderThreaded.QueuedImage();
            queuedImage.imgPath = url;
            queuedImage.forceReload = true;
            queuedImage.skipCache = true;
            queuedImage.compress = true;
            queuedImage.createMipMaps = false;
            queuedImage.isNormalMap = false;
            queuedImage.linear = false;
            queuedImage.createAlphaFromGrayscale = false;
            queuedImage.createNormalFromBump = false;
            queuedImage.bumpStrength = 0f;
            queuedImage.isThumbnail = false;
            queuedImage.fillBackground = false;
            queuedImage.invert = false;
            queuedImage.callback = (ImageLoaderThreaded.QueuedImage qi) =>
            {
                Texture2D tex = qi.tex;
                _buttonTexture = tex;
                _mvrScript.RefreshUIButtonGrid();
            };

            ImageLoaderThreaded.singleton.QueueImage(queuedImage);
        }

        public void PlayAction(string gazeAtom, string gazePerson, string gazeFemale, string gazeMale) {
            if (_buttonType == "Suppress Scale Load")
            {
                if (_mvrScript._suppressScaleLoad) _mvrScript._suppressScaleLoad = false;
                else _mvrScript._suppressScaleLoad = true;
            }
            else if (_buttonType == "Freeze Motion/Sound")
            {
                if (SuperController.singleton.freezeAnimation) SuperController.singleton.SetFreezeAnimation(false);
                else SuperController.singleton.SetFreezeAnimation(true);
            }
            else if (_buttonType == "Toggle Play/Edit Mode")
            {
                if (SuperController.singleton.gameMode == SuperController.GameMode.Play)
                {
                    SuperController.singleton.gameMode = SuperController.GameMode.Edit;
                    if (_openGameUIonEditMode) SuperController.singleton.ShowMainHUDAuto();
                }
                else
                {
                    SuperController.singleton.gameMode = SuperController.GameMode.Play;
                    if (_closeGameUIonPlayMode) SuperController.singleton.HideMainHUD();
                }              
            }
            else
            {
                if (_targetCategory == "Specific Atom")
                {
                    Atom atom = SuperController.singleton.GetAtomByUid(_targetName);
                    if (atom != null)
                    {
                        PlayActionTargetAtom(atom);
                        ToggleButtonState(atom.name);
                    }
                }
                else if (_targetCategory == "Gaze Selected Atom")
                {
                    string atomName = "";
                    switch (_targetName)
                    {
                        case "Last viewed Atom":
                            atomName = gazeAtom;
                            break;
                        case "Last viewed Person":
                            atomName = gazePerson;
                            break;
                        case "Last viewed Female":
                            atomName = gazeFemale;
                            break;
                        case "Last viewed Male":
                            atomName = gazeMale;
                            break;
                    }
                    Atom atom = SuperController.singleton.GetAtomByUid(atomName);
                    if (atom != null)
                    {
                        PlayActionTargetAtom(atom);
                        ToggleButtonState(atom.name);
                    }
                }
                else if (_targetCategory == "Atom Group")
                {
                    foreach (Atom atom in SuperController.singleton.GetAtoms())
                    {
                        if (atom.name != "CoreControl" && atom.name != "[CameraRig]")
                        {
                            bool isMale = false;
                            if (atom.type == "Person" && (atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("male") || atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("jarlee") || atom.GetComponentInChildren<DAZCharacter>().name.StartsWith("jarjulian"))) isMale = true;
                            switch (_targetName)
                            {
                                case "All Atoms":
                                    PlayActionTargetAtom(atom);
                                    break;
                                case "All People":
                                    if (atom.type == "Person") { PlayActionTargetAtom(atom); }
                                    break;
                                case "All Females":
                                    if (atom.type == "Person" && !isMale) { PlayActionTargetAtom(atom); }
                                    break;
                                case "All Males":
                                    if (atom.type == "Person" && isMale) { PlayActionTargetAtom(atom); }
                                    break;
                            }
                        }
                    }
                    ToggleButtonState(_targetName);
                }
            }
        }

        private void PlayActionTargetAtom(Atom atom) {
            JSONStorable js = null;
            string actionType = "";
            bool reverse = false;
            string actionMode = "";

            switch (_buttonType) {
                case "Load App Preset":
                    js = atom.GetStorableByID("AppearancePresets");
                    actionType = "LOADPRESET";
                    if (_suppressClothingLoad || _suppressPersonScaleLoad || _mvrScript._suppressScaleLoad) { actionMode = "SUPPRESS"; }
                    break;
                case "Load Pose Preset":
                    js = atom.GetStorableByID("PosePresets");
                    actionType = "LOADPRESET";
                    break;
                case "Load Hair Preset":
                    js = atom.GetStorableByID("HairPresets");
                    actionType = "LOADPRESET";
                    break;
                case "Load Glute Preset":
                    js = atom.GetStorableByID("FemaleGlutePhysicsPresets");
                    if (js == null) { js = atom.GetStorableByID("MaleGlutePhysicsPresets"); }
                    actionType = "LOADPRESET";
                    break;
                case "Load Breast Preset":
                    js = atom.GetStorableByID("FemaleBreastPhysicsPresets");
                    if (js == null) { js = atom.GetStorableByID("MaleBreastPhysicsPresets"); }
                    actionType = "LOADPRESET";
                    break;
                case "Load Morph Preset":
                    js = atom.GetStorableByID("MorphPresets");
                    actionType = "LOADPRESET";
                    break;
                case "Load Skin Preset":
                    js = atom.GetStorableByID("SkinPresets");
                    actionType = "LOADPRESET";
                    break;
                case "Load General Preset":
                    js = atom.GetStorableByID("Preset");
                    actionType = "LOADPRESET";
                    break;
                case "Load Clothing Preset":
                    js = atom.GetStorableByID("ClothingPresets");
                    actionType = "LOADPRESET";
                    break;
                case "Remove All Clothing":
                    actionType = "CLOTHING";
                    actionMode = "REMOVE";
                    break;
                case "Undress All Clothing":
                    actionType = "CLOTHING";
                    actionMode = "UNDRESS";
                    if (GetButtonState(atom.name) == 2) { reverse = true; }
                    break;
                case "Reset Sim All Clothing":
                    actionType = "CLOTHING";
                    actionMode = "RESETSIM";
                    break;
                case "Remove Clothing Preset":
                    actionType = "CLOTHINGPRESET";
                    actionMode = "REMOVE";
                    break;
                case "Merge Clothing Preset":
                    actionType = "CLOTHINGPRESET";
                    actionMode = "MERGE";
                    if (GetButtonState(atom.name) == 2) { reverse = true; }
                    break;
                case "Undress Cloth Preset":
                    actionType = "CLOTHINGPRESET";
                    actionMode = "UNDRESS";
                    if (GetButtonState(atom.name) == 2) { reverse = true; }
                    break;
                case "Reset Sim Cloth Preset":
                    actionType = "CLOTHINGPRESET";
                    actionMode = "RESETSIM";
                    break;
                case "Load Plugin(s)":
                    actionType = "LOADPLUGINS";
                    break;
                case "Unload All Plugins":
                    actionType = "UNLOADPLUGINS";
                    break;
                case "Plugin Action":
                    actionType = "PLUGINACTION";
                    actionMode = "SINGLE";
                    break;
                case "Plugin Action Toggle":
                    actionType = "PLUGINACTION";
                    actionMode = "TOGGLE";
                    break;
                case "Plugin Bool Toggle":
                    actionType = "PLUGINBOOL";
                    actionMode = "TOGGLE";
                    break;
                case "Set Plugin Bool True":
                    actionType = "PLUGINBOOL";
                    actionMode = "TRUE";
                    break;
                case "Set Plugin Bool False":
                    actionType = "PLUGINBOOL";
                    actionMode = "FALSE";
                    break;
                case "Reset Scale":
                    actionType = "RESET";
                    actionMode = "SCALE";
                    break;
                case "Reset Appearance":
                    actionType = "RESET";
                    actionMode = "APP";
                    break;
            }
            switch (actionType) {
                case "LOADPRESET":
                    if (js != null && _presetFilePath != "") {
                        string fileName = "";
                        if (actionMode == "SUPPRESS" && Constants._patreonContentEnabled) { fileName = PatreonFeatures.AppPresetSuppressLoad(_mvrScript, atom, _presetFilePath, _suppressClothingLoad, _suppressPersonScaleLoad || _mvrScript._suppressScaleLoad); }
                        else { fileName = _presetFilePath; }

                        JSONStorableUrl presetPathJSON = js.GetUrlJSONParam("presetBrowsePath");
                        presetPathJSON.val = "";
                        presetPathJSON.val = SuperController.singleton.NormalizePath(fileName);
                        js.CallAction("LoadPreset");
                        if (actionMode == "SUPPRESS") {
                           FileManagerSecure.DeleteFile("Custom/Atom/Person/Appearance/Preset_UIAssistTemp1.vap");
                           FileManagerSecure.DeleteFile("Custom/Atom/Person/Appearance/Preset_UIAssistTemp2.vap");
                        }

                    }
                    break;
                case "CLOTHING":
                    ClothingActions(atom, actionMode, reverse);
                    break;
                case "CLOTHINGPRESET":
                    if (Constants._patreonContentEnabled)
                    {
                        if (actionMode=="MERGE" && _tagsToRemoveOnMerge.Count > 0 && _mvrScript._atomActiveClothingDictionary.ContainsKey(atom.name))
                        {
                            PatreonFeatures.RemoveTagClothingItems(atom, _mvrScript._atomActiveClothingDictionary[atom.name]._activeClothingTags,_tagsToRemoveOnMerge);
                        }
                        PatreonFeatures.ClothingPresetActions(_mvrScript, _mergeClothingGeometryIDs, atom, actionMode, _presetFilePath, reverse);
                    }
                    break;
                case "LOADPLUGINS":
                    if (atom.name != "CoreControl") StartLoadPlugins(atom, _pluginForceMergeLoad);
                    break;
                case "UNLOADPLUGINS":
                    if (atom.name != "CoreControl") UnloadPlugins(atom);
                    break;
                case "PLUGINACTION":
                    PluginAction(atom, actionMode);
                    break;
                case "PLUGINBOOL":
                    PluginBool(atom, actionMode);
                    break;
                case "RESET":
                    ResetAppearance(atom, actionMode);
                    break;
            }
        }
        private void ResetAppearance(Atom atom, string mode)
        {
            JSONArray atomsArray = SuperController.singleton.loadJson["atoms"].AsArray;
            bool atomExistedAtLoad = false;
            JSONClass atomAtLoadJSON =null;

            foreach (JSONClass atomJSON in atomsArray)
            {
                if ((string)atomJSON["id"] == atom.name)
                {
                    atomExistedAtLoad = true;
                    atomAtLoadJSON = atomJSON;
                }
            }

            if (mode == "SCALE")
            {

                JSONStorable scaleStorableJSON = atom.GetStorableByID("rescaleObject");
                JSONStorableFloat scaleJSON = scaleStorableJSON.GetFloatJSONParam("scale");
                if (atomExistedAtLoad)
                {
                    float resetScale = 1f;
                    foreach (JSONClass storable in atomAtLoadJSON["storables"].AsArray)
                    {
                        if (storable["id"] == "rescaleObject")
                        {
                            resetScale = storable["scale"].AsFloat;
                            break;
                        }
                    }
                    scaleJSON.val = resetScale;
                }
                else { scaleJSON.val = 1f; }
            }

            else if (mode == "APP" && atomExistedAtLoad)
            {
                JSONArray newAtomsArray = new JSONArray();
                JSONClass newSave = new JSONClass();
                newAtomsArray.Add(atomAtLoadJSON);
                newSave["atoms"] = newAtomsArray;
                SuperController.singleton.SaveJSON(newSave, "Saves\\UIAtemp.json");
                atom.LoadAppearancePreset("Saves\\UIAtemp.json");
                FileManagerSecure.DeleteFile("Saves\\UIAtemp.json");
                
            }
            else if (mode == "APP" && !atomExistedAtLoad) atom.ResetAppearance();
        }
        private void PluginBool(Atom atom, string mode) {
            bool boolState = true;
            if (mode == "FALSE" || (mode == "TOGGLE" && GetButtonState(atom.name) == 2)) { boolState = false; }

            foreach (string storableName in atom.GetStorableIDs()) {
                if (storableName.EndsWith(_pluginType)) {
                    JSONStorable storable = atom.GetStorableByID(storableName);
                    if (storable != null && storable.GetBoolParamNames().Contains(_targetParamNameOn)) {
                        storable.SetBoolParamValue(_targetParamNameOn, boolState);
                    }
                }
            }
        }
        private void PluginAction(Atom atom, string mode) {
            string actionName = "";
            PluginVariables preActionVariables;

            if (mode == "SINGLE" || GetButtonState(atom.name) == 1) {
                actionName = _targetParamNameOn;
                preActionVariables = _pluginVariablesButtonOn;
            }
            else {
                actionName = _targetParamNameOff;
                preActionVariables = _pluginVariablesButtonOff; 
            }

            
            foreach (string storableName in atom.GetStorableIDs()) {
                if (storableName.EndsWith(_pluginType)) {
                    JSONStorable storable = atom.GetStorableByID(storableName);
                    if (storable != null && storable.GetActionNames().Contains(actionName)) {
                        // Set any pre-action plugin parameters
                        foreach (KeyValuePair<string,PluginVariable> kvp in preActionVariables._pluginVarDict)
                        {
                            if (storable.GetBoolParamNames().Contains(kvp.Key))
                            {
                                JSONStorableBool boolJSON = storable.GetBoolJSONParam(kvp.Key);
                                boolJSON.val = kvp.Value._pluginBoolValue;
                            }
                            else if (storable.GetFloatParamNames().Contains(kvp.Key))
                            {
                                JSONStorableFloat floatJSON = storable.GetFloatJSONParam(kvp.Key);
                                floatJSON.val = kvp.Value._pluginFloatValue;
                            }
                            else if (storable.GetStringParamNames().Contains(kvp.Key))
                            {
                                JSONStorableString stringJSON = storable.GetStringJSONParam(kvp.Key);
                                stringJSON.val = kvp.Value._pluginStringValue;
                            }
                            else if (storable.GetStringChooserParamNames().Contains(kvp.Key))
                            {
                                JSONStorableStringChooser stringChooserJSON = storable.GetStringChooserJSONParam(kvp.Key);
                                stringChooserJSON.val = kvp.Value._pluginStringValue;
                                
                            }
                            else if (storable.GetUrlParamNames().Contains(kvp.Key))
                            {
                                JSONStorableUrl urlJSON = storable.GetUrlJSONParam(kvp.Key);
                                urlJSON.val = kvp.Value._pluginStringValue;
                            }
                            else if (storable.GetColorParamNames().Contains(kvp.Key))
                            {
                                JSONStorableColor colorJSON = storable.GetColorJSONParam(kvp.Key);
                                colorJSON.val = kvp.Value._pluginColorValue;
                            }

                        }

                        storable.CallAction(actionName);
                    }
                }
            }
        }

        private void ClothingActions(Atom atom, string mode, bool reverse) {
            foreach (string receiverName in atom.GetStorableIDs()) {
                JSONStorable receiver = atom.GetStorableByID(receiverName);

                if (receiver != null) {
                    if (receiver.storeId.Length >= 3 && receiver.storeId.Substring(receiver.storeId.Length - 3) == "Sim") {
                        JSONStorableBool unDressBool = receiver.GetBoolJSONParam("allowDetach");
                        JSONStorableBool simEnabledBool = receiver.GetBoolJSONParam("simEnabled");

                        if (mode == "UNDRESS" && unDressBool != null) { unDressBool.val = !reverse; }
                        if (mode == "RESETSIM" && simEnabledBool != null) { if (simEnabledBool.val) { receiver.CallAction("Reset"); } }
                    }
                    if (receiver.storeId == "geometry" && mode == "REMOVE") {
                        foreach (string target in receiver.GetBoolParamNames()) {
                            if (target.Length > 8 && target.Substring(0, 9) == "clothing:") {
                                JSONStorableBool boolTarget = receiver.GetBoolJSONParam(target);
                                if (boolTarget != null) { boolTarget.val = false; }
                            }
                        }
                    }
                }
            }
        }

        private bool IsPluginLoaded(JSONClass pluginMgrJSON, string pluginFileName) {
            bool loaded = false;
            for (int index = 0; index < pluginMgrJSON["plugins"].Count; ++index) {
                if (pluginMgrJSON["plugins"][index] == null) { break; }
                else if (pluginMgrJSON["plugins"][index].Value == pluginFileName) {
                    loaded = true;
                    break;
                }
            }
            return (loaded);
        }
        private int GetNextFreePluginID(JSONClass pluginMgrJSON) {
            int pluginID = -1;
            for (int index = 0; index < 1000; ++index) {
                if (pluginMgrJSON["plugins"]["plugin#" + index.ToString()] == null) {
                    pluginID = index;
                    break;
                }
            }
            return (pluginID);
        }
        private int GetPluginMgrIndex(JSONClass atomJSON) {
            int pluginMgrIndex = -1;
            int storablesIndex = 0;

            foreach (JSONClass storableJSON in atomJSON["storables"].AsArray) {
                if (storableJSON["id"].Value == "PluginManager") { pluginMgrIndex = storablesIndex; break; }
                storablesIndex++;
            }
            return (pluginMgrIndex);
        }

        private void StartLoadPlugins(Atom atom, bool merge)
        {
            Dictionary<string, bool> timelineIsPlayingDic = new Dictionary<string, bool>();
            Dictionary<string, float> timelineScrubberDic = new Dictionary<string, float>();
            Dictionary<string, string> timelineAnimationDic = new Dictionary<string, string>();

            // Check for active Timeline plugins
            foreach (string storableID in atom.GetStorableIDs())
            {
                if (storableID.StartsWith("plugin#") && storableID.EndsWith("Timeline.AtomPlugin"))
                {
                    JSONStorable pluginReceiver = atom.GetStorableByID(storableID);
                    JSONStorableFloat timelineScrubberJSON = pluginReceiver.GetFloatJSONParam("Scrubber");
                    JSONStorableBool timelineIsPlayJSON = pluginReceiver.GetBoolJSONParam("Is Playing");
                    JSONStorableStringChooser timelineAnimationJSON = pluginReceiver.GetStringChooserJSONParam("Animation");
                    if (timelineIsPlayJSON != null && timelineScrubberJSON != null && timelineAnimationJSON != null )
                    {
                        timelineIsPlayingDic.Add(storableID, timelineIsPlayJSON.val);
                        timelineScrubberDic.Add(storableID, timelineScrubberJSON.val);
                        timelineAnimationDic.Add(storableID, timelineAnimationJSON.val);
                    }
                }
            }

            int maxPlugins = 100;
            List<string> test = new List<string>();
            test.Add("Temp");
            PatreonFeatures.GridLayoutFeature(test);
            if (!Constants._patreonContentEnabled || test == null || test.Count != 7)
            {
                maxPlugins = 1;
            }

            if (merge) MergeLoadPlugins(atom, maxPlugins);
            else LoadPlugins(atom, maxPlugins);

            // Reset active Timeline plugins
            _mvrScript.StartCoroutine(RestoreTimeLineState(atom, timelineIsPlayingDic, timelineScrubberDic, timelineAnimationDic,merge));

        }
        private IEnumerator RestoreTimeLineState(Atom atom, Dictionary<string, bool> timelineIsPlayingDic, Dictionary<string, float> timelineScrubberDic, Dictionary<string, string> timelineAnimationDic, bool merge)
        {
            if (merge) yield return new WaitUntil(() => !SuperController.singleton.isLoading);
            else yield return new WaitForEndOfFrame();
            if (timelineIsPlayingDic.Count > 0)
            {
                foreach (KeyValuePair<string, bool> kvp in timelineIsPlayingDic)
                {
                    JSONStorable pluginReceiver = atom.GetStorableByID(kvp.Key);
                    if (pluginReceiver != null && timelineAnimationDic[kvp.Key]!="(Playing...)")
                    {
                        JSONStorableAction timelineStopActionJSON = pluginReceiver.GetAction("Stop");
                        if (timelineStopActionJSON != null) pluginReceiver.CallAction("Stop");

                        JSONStorableFloat timelineScrubberJSON = pluginReceiver.GetFloatJSONParam("Scrubber");
                        JSONStorableBool timelineIsPlayJSON = pluginReceiver.GetBoolJSONParam("Is Playing");
                        JSONStorableStringChooser timelineAnimationJSON = pluginReceiver.GetStringChooserJSONParam("Animation");
                        timelineAnimationJSON.val = timelineAnimationDic[kvp.Key];
                        timelineScrubberJSON.val = timelineScrubberDic[kvp.Key];
                        if (kvp.Value)
                        {
                            JSONStorableAction timelinePlayActionJSON = pluginReceiver.GetAction("Play");
                            if (timelinePlayActionJSON != null) pluginReceiver.CallAction("Play");
                        }

                    }
                }
            }
        }
        private void MergeLoadPlugins(Atom atom, int maxPlugins)
        {
            JSONClass atomsJSON = SuperController.singleton.GetSaveJSON(atom, true, true);
            JSONClass atomJSON = (JSONClass)atomsJSON["atoms"][0];
            JSONClass containingAtomJSON = atom.GetJSON();
            int pluginMgrIndex = GetPluginMgrIndex(atomJSON);
            bool editModeOn = SuperController.singleton.editModeToggle.isOn;      

            int pluginCount = 1;
            foreach (ButtonPluginLoad buttonPluginToLoad in _pluginLoadList)
            {
                if (buttonPluginToLoad._filePath != "" && buttonPluginToLoad._filePath != null && pluginCount <= maxPlugins)
                {
                    string normalFilePath = SuperController.singleton.NormalizePath(buttonPluginToLoad._filePath);
                    if (!IsPluginLoaded((JSONClass)atomJSON["storables"][pluginMgrIndex], normalFilePath) || !buttonPluginToLoad._singleInstancePerAtom)
                    {
                        int nextFreePluginID = GetNextFreePluginID((JSONClass)atomJSON["storables"][pluginMgrIndex]);
                        atomJSON["storables"][pluginMgrIndex]["plugins"]["plugin#" + nextFreePluginID.ToString()] = normalFilePath;
                        if (buttonPluginToLoad._pluginSaveData != "" && buttonPluginToLoad._pluginSaveData != null && maxPlugins > 1)
                        {
                            PatreonFeatures.MergePluginStorables(atomJSON, buttonPluginToLoad._pluginSaveJSON, nextFreePluginID);
                        }
                    }
                }
                pluginCount++;
            }

            JSONClass mergeSceneJSON = new JSONClass();
            mergeSceneJSON["atoms"][-1] = atomJSON;

            SuperController.singleton.SaveJSON(mergeSceneJSON, "Custom\\Atom\\Person\\UIAmergeTemp.json");
            SuperController.singleton.LoadMerge("Custom\\Atom\\Person\\UIAmergeTemp.json");
            FileManagerSecure.DeleteFile("Custom\\Atom\\Person\\UIAmergeTemp.json");
            SuperController.singleton.editModeToggle.isOn = editModeOn;

        }
        private void UnloadPlugins(Atom atom)
        {
            MVRPluginManager manager = atom.GetStorableByID("PluginManager") as MVRPluginManager;

            JSONClass emptyManager = new JSONClass();
            emptyManager["id"] = "PluginManager";
            manager.LateRestoreFromJSON(emptyManager);
        }
            private void LoadPlugins(Atom atom, int maxPlugins)
        {
            MVRPluginManager manager = atom.GetStorableByID("PluginManager") as MVRPluginManager;
            
            JSONClass currentPlugins = manager.GetJSON(true, true, true);
            Dictionary<string, JSONClass> pluginSettings = new Dictionary<string, JSONClass>();

            foreach (string sid in atom.GetStorableIDs())
            {
                foreach (KeyValuePair<string, JSONNode> kvp in (JSONClass)currentPlugins["plugins"])
                {
                    if (sid.StartsWith(kvp.Key + "_"))
                    {
                        JSONClass pluginData = atom.GetStorableByID(sid).GetJSON(true, true, true);
                        pluginSettings[kvp.Key + "_"] = pluginData;
                    }
                }
            }

            int pluginCount = 1;
            foreach (ButtonPluginLoad buttonPluginToLoad in _pluginLoadList)
            {
                if (buttonPluginToLoad._filePath != "" && buttonPluginToLoad._filePath != null && pluginCount <= maxPlugins)
                {
                    string normalFilePath = SuperController.singleton.NormalizePath(buttonPluginToLoad._filePath);
                    if (!IsPluginLoaded(currentPlugins, normalFilePath) || !buttonPluginToLoad._singleInstancePerAtom)
                    {
                        int nextFreePluginID = GetNextFreePluginID(currentPlugins);
                        currentPlugins["plugins"]["plugin#" + nextFreePluginID.ToString()] = normalFilePath;

                        if (buttonPluginToLoad._pluginSaveData != "" && buttonPluginToLoad._pluginSaveData != null && maxPlugins > 1)
                        {
                            PatreonFeatures.MergePluginSettings(pluginSettings, buttonPluginToLoad._pluginSaveJSON, nextFreePluginID);

                        }
                    }
                }
                pluginCount++;
            }
            manager.LateRestoreFromJSON(currentPlugins);

            foreach (KeyValuePair<string, JSONClass> kvp in pluginSettings)
            {
                string sid = atom.GetStorableIDs().FirstOrDefault(id => id.StartsWith(kvp.Key));
                JSONStorable pluginStorable = atom.GetStorableByID(sid);
                pluginStorable.RestoreFromJSON(kvp.Value);
            }
        }

    }

    public class ButtonDisplay {
        private UIAssist _mvrScript;
        public Button _selectedButton;
        public JSONStorableStringChooser _buttonCategoryJSON;
        public JSONStorableStringChooser _buttonTypeJSON;
        public JSONStorableString _buttonLabelOnJSON;
        public JSONStorableString _buttonLabelOffJSON;
        public JSONStorableStringChooser _targetCategoryJSON;
        public JSONStorableStringChooser _targetNameJSON;
        public JSONStorableStringChooser _pluginLoadSelectorJSON;
        public JSONStorableStringChooser _pluginSaveDataSelectorJSON;
        public JSONStorableStringChooser _tagItemsToRemoveSelectorJSON;
        public JSONStorableBool _pluginForceMergeLoadJSON;
        public JSONStorableBool _closeGameUIonPlayModeJSON;
        public JSONStorableBool _openGameUIonEditModeJSON;

        public JSONStorableBool _singleInstancePerAtomJSON;
        public JSONStorableBool _suppressClothingLoadJSON;
        public JSONStorableBool _suppressPersonScaleLoadJSON;
        public JSONStorableBool _autoLabelJSON;

        public JSONStorableString _filePathJSON;
        public JSONStorableString _pluginSaveDataJSON;
        public JSONStorableStringChooser _targetPluginTypeJSON;

        public JSONStorableStringChooser _targetParamNameOnJSON;
        public JSONStorableStringChooser _targetParamNameOffJSON;
        public JSONStorableBool _useButtonTextureJSON;

        private UIDynamicPopup _buttonCategoryPopup;
        private UIDynamicPopup _buttonTypePopup;
        private UIDynamicPopup _targetCategoryPopup;
        private UIDynamicPopup _targetNamePopup;
        private UIDynamicPopup _pluginLoadSelectorPopup;
        private UIDynamicPopup _pluginSaveDataSelectorPopup;
        private UIDynamicPopup _targetPluginTypePopup;
        private UIDynamicPopup _tagItemsToRemoveSelectorPopup;
        private UIDynamicTextField _filePathTextField;
        private UIDynamicTextField _buttonLabelOnTextField;
        private UIDynamicTextField _buttonLabelOffTextField;
        private UIDynamicTextField _pluginSaveDataTextField;
        private UIDynamicPopup _targetParamNameOnPopup;
        private UIDynamicPopup _targetParamNameOffPopup;
        private UIDynamicButton _selectFileButton;
        private UIDynamicButton _copyButtonButton;
        private UIDynamicToggle _singleInstancePerAtomToggle;
        private UIDynamicToggle _useButtonTextureToggle;
        private UIDynamicToggle _suppressClothingLoadToggle;
        private UIDynamicToggle _suppressPersonScaleLoadToggle;
        private UIDynamicToggle _autoLabelToggle;
        private UIDynamicToggle _pluginForceMergeLoadToggle;
        private UIDynamicToggle _closeGameUIonPlayModeToggle;
        private UIDynamicToggle _openGameUIonEditModeToggle;

        private UIDynamicButton _gridDeleteButton;
        private UIDynamic _spacer1;
        private UIDynamic _spacer1b;
        private UIDynamic _spacer2;
        private UIDynamic _spacer2b;
        private UIDynamic _spacer3;

        public ButtonSkinDisplay _buttonSkinDisplay;
        public PluginVariablesDisplay _pluginVariablesDisplayOn;
        public PluginVariablesDisplay _pluginVariablesDisplayOff;

        bool _isDisplayUpdating = false;

        public readonly string[] _ClothingTags = { "Accessory", "Bodysuit", "Bottom", "Bra", "Dress", "Glasses", "Gloves", "Hat", "Jewelry", "Mask", "Panties", "Pants", "Shirt", "Shoes", "Shorts", "Skirt", "Socks", "Stockings", "Sweater", "Top", "Underwear", "Arms", "Feet", "Full Body", "Hands", "Head", "Hip", "Legs", "Neck", "Torso" };

        public ButtonDisplay(Button button, UIAssist mvrScript) {
            try {
                _selectedButton = button;
                _mvrScript = mvrScript;

                _buttonCategoryJSON = new JSONStorableStringChooser("buttonCategory", _selectedButton.GetButtonCategoryChoices(), "", "", SwitchButonCategory);
                _buttonTypeJSON = new JSONStorableStringChooser("buttonType", null, "", "", SwitchButonType);
                _targetCategoryJSON = new JSONStorableStringChooser("atomCategory", null, "", "", SwitchTargetCategory);
                _targetNameJSON = new JSONStorableStringChooser("atomName", null, "", "", SwitchTargetName);
                _filePathJSON = new JSONStorableString("filePath", "", SwitchFilePath);
                _pluginLoadSelectorJSON = new JSONStorableStringChooser("pluginLoadSelector", null, "1", "Plugin Selector", SwitchPluginLoad);
                _targetParamNameOnJSON = new JSONStorableStringChooser("targetNameOn", null, "", "", SwitchTargetParamNameOn);
                _targetParamNameOffJSON = new JSONStorableStringChooser("targetNameOff", null, "", "", SwitchTargetParamNameOff);
                _buttonLabelOnJSON = new JSONStorableString("buttonLabelOn", "", SwitchButtonLabelOn);
                _buttonLabelOffJSON = new JSONStorableString("buttonLabelOff", "", SwitchButtonLabelOff);

                _pluginSaveDataJSON = new JSONStorableString("pluginSaveData", "");
                _singleInstancePerAtomJSON = new JSONStorableBool("singleInstancePerAtom", true, SwitchSingleInstancePerAtom);
                _pluginSaveDataSelectorJSON = new JSONStorableStringChooser("pluginSaveDataSelector", null, "", "", SwitchSaveDataPlugin);
                _targetPluginTypeJSON = new JSONStorableStringChooser("targetPluginType", null, "", "", SwitchTargetPluginType);
                _useButtonTextureJSON = new JSONStorableBool("useButtonTexture", true, SwitchUseButtonTexture);
                _suppressClothingLoadJSON = new JSONStorableBool("suppressClothingLoad", false, SwitchSurpressClothingLoad);
                _suppressPersonScaleLoadJSON = new JSONStorableBool("suppressPersonScaleLoad", false, SwitchSurpressPersonScaleLoad);
                _autoLabelJSON = new JSONStorableBool("autoLabel", true, SwitchAutoLabel);
                _pluginForceMergeLoadJSON = new JSONStorableBool("pluginForceMergeLoad", false, SwitchForceMergeLoad);
                _closeGameUIonPlayModeJSON = new JSONStorableBool("closeGameUIonPlayMode", true, SwitchCloseGameUIonPlayMode);
                _openGameUIonEditModeJSON = new JSONStorableBool("openGameUIonEditMode", true, SwitchOpenGameUIonEditMode);

                if (_selectedButton._buttonSkin != null) _buttonSkinDisplay = new ButtonSkinDisplay(_selectedButton._buttonSkin, _mvrScript, false);
                if (_selectedButton._pluginVariablesButtonOn != null) _pluginVariablesDisplayOn = new PluginVariablesDisplay(_mvrScript,_selectedButton._pluginVariablesButtonOn);
                if (_selectedButton._pluginVariablesButtonOff != null) _pluginVariablesDisplayOff = new PluginVariablesDisplay(_mvrScript, _selectedButton._pluginVariablesButtonOff);
                _tagItemsToRemoveSelectorJSON = new JSONStorableStringChooser("tagItemsToRemoveSelector", null, "Toggle Clothing Tag...", "Clothing Tags to Remove", SwitchClothingTagsToggle);

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }

        }
        private void SwitchCloseGameUIonPlayMode(bool toggle)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._closeGameUIonPlayMode = toggle;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchOpenGameUIonEditMode(bool toggle)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._openGameUIonEditMode = toggle;
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchForceMergeLoad(bool pluginForceMergeLoad)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    _selectedButton._pluginForceMergeLoad = pluginForceMergeLoad;
                    RefreshDisplay();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchClothingTagsToggle(string clothingTag)
        {
            try
            {
                if (!_isDisplayUpdating)
                {
                    string tag = clothingTag.Substring(2, clothingTag.Length-4);
                    _selectedButton.UpdateTagItemToRemove(tag);
                    RefreshDisplay();
                }

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchAutoLabel(bool switchAutoLabel) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton._autoLabel = switchAutoLabel;
                    RefreshDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchSurpressClothingLoad(bool suppressClothingLoad) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton._suppressClothingLoad = suppressClothingLoad;
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchSurpressPersonScaleLoad(bool suppressPersonScaleLoad) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton._suppressPersonScaleLoad = suppressPersonScaleLoad;
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchTargetPluginType(string targetPluginType) {
            try {
                if (!_isDisplayUpdating) {
                    if (targetPluginType != "No plugins in Scene") {
                        _selectedButton._pluginType = targetPluginType;
                        _selectedButton.UpdateButtonType(_selectedButton._buttonType);
                        _selectedButton._pluginVariablesButtonOn._pluginVarDict.Clear();
                        _selectedButton._pluginVariablesButtonOff._pluginVarDict.Clear();
                        _pluginVariablesDisplayOn._pluginVariableSelectorJSON.val = "Select Parameter...";
                        _pluginVariablesDisplayOff._pluginVariableSelectorJSON.val = "Select Parameter...";
                        RefreshDisplay();
                    }
                    else { _selectedButton._pluginType = ""; }
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchTargetParamNameOn(string targetParamChoice) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton._targetParamNameOn = targetParamChoice;
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchTargetParamNameOff(string targetParamChoice) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton._targetParamNameOff = targetParamChoice;
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchSaveDataPlugin(string pluginChoice) {
            try {
                if (!_isDisplayUpdating) {
                    if (pluginChoice == "<Clear Plugin Data>") {
                        _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._pluginSaveData = "";
                        _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._pluginSaveJSON = null;
                    }
                    else if (pluginChoice != "<None available>") {
                        string pluginName = pluginChoice.Substring(0, pluginChoice.IndexOf('['));
                        string atomName = pluginChoice.Substring(pluginChoice.IndexOf('[') + 1);
                        JSONArray pluginDataJSON = _selectedButton.GetPluginDataByName(atomName.Remove(atomName.Length - 1), pluginName);
                        string pluginData = pluginDataJSON.ToString();
                        _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._pluginSaveData = pluginData.Substring(0, Math.Min(pluginData.Length, 500));
                        _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._pluginSaveJSON = pluginDataJSON;

                    }
                    RefreshDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        private void SwitchUseButtonTexture(bool useTexture) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton._useButtonTexture = useTexture;
                    _mvrScript.RefreshUIButtonGrid();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        private void SwitchSingleInstancePerAtom(bool singleInstance) {
            try {
                if (!_isDisplayUpdating) {
                    if (_selectedButton._buttonType == "Load Plugin(s)") {
                        _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._singleInstancePerAtom = singleInstance;
                    }
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchFilePath(string filePath) {
            try {
                if (!_isDisplayUpdating) {
                    if (_selectedButton._buttonType == "Load Plugin(s)") {
                        _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._filePath = filePath;
                        _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._pluginSaveData = "";
                        _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._pluginSaveJSON = null;
                        RefreshDisplay();
                    }
                    else if (_selectedButton._buttonType.Length > 6 && _selectedButton._buttonType.Substring(_selectedButton._buttonType.Length - 6, 6) == "Preset") {
                        _selectedButton.SetPresetFile(filePath);

                    }
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchPluginLoad(string pluginLoad) {
            try {
                if (!_isDisplayUpdating) {
                    if (pluginLoad=="Add Plugin...")
                    {
                        ButtonPluginLoad newPluginLoad = new ButtonPluginLoad();
                        _selectedButton._pluginLoadList.Add(newPluginLoad);
                        _isDisplayUpdating = true;
                        _pluginLoadSelectorJSON.val = _selectedButton._pluginLoadList.Count.ToString();
                        _isDisplayUpdating = false ;
                    }
                    else if (pluginLoad.StartsWith("Remove Plugin"))
                    {
                        _selectedButton._pluginLoadList.RemoveAt(int.Parse(pluginLoad.Substring(14)) -1);
                        _isDisplayUpdating = true;
                        _pluginLoadSelectorJSON.val = "1";
                        _isDisplayUpdating = false;
                    }
                    RefreshDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchButtonLabelOn(string buttonLabel) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton._buttonLabelOn = buttonLabel;
                    RefreshDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchButtonLabelOff(string buttonLabel) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton._buttonLabelOff = buttonLabel;
                    RefreshDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchButonCategory(string buttonCategory) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton.UpdateButtonCategory(buttonCategory);
                    _selectedButton._buttonTexture = null;
                    _mvrScript._buttonGridDisplay._selectedButtonGrid.RecalcActiveButtonCount();
                    _mvrScript._buttonGridDisplay._selectedButtonGrid.RecalcGazeSelections();
                    _mvrScript.RefreshUIButtonGrid();
                    RefreshDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchButonType(string buttonType) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton.UpdateButtonType(buttonType);
                    _mvrScript._buttonGridDisplay._selectedButtonGrid.RecalcActiveButtonCount();
                    _mvrScript._buttonGridDisplay._selectedButtonGrid.RecalcGazeSelections();
                    _mvrScript.RefreshUIButtonGrid();
                    RefreshDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchTargetCategory(string atomCategory) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton.UpdateAtomCategory(atomCategory);
                    _mvrScript._buttonGridDisplay._selectedButtonGrid.RecalcGazeSelections();
                    RefreshDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchTargetName(string atomName) {
            try {
                if (!_isDisplayUpdating) {
                    _selectedButton.UpdateAtom(atomName);
                    _mvrScript._buttonGridDisplay._selectedButtonGrid.RecalcGazeSelections();
                    RefreshDisplay();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void RefreshDisplay() {
            CloseAllPopup(null);
            RefreshDisplayValues();
            if (_buttonCategoryPopup != null) { _mvrScript.RemovePopup(_buttonCategoryPopup); }
            if (_buttonTypePopup != null) { _mvrScript.RemovePopup(_buttonTypePopup); }
            if (_targetCategoryPopup != null) { _mvrScript.RemovePopup(_targetCategoryPopup); }
            if (_targetNamePopup != null) { _mvrScript.RemovePopup(_targetNamePopup); }
            if (_tagItemsToRemoveSelectorJSON != null) { _mvrScript.RemovePopup(_tagItemsToRemoveSelectorJSON); }
            if (_pluginLoadSelectorPopup != null) { _mvrScript.RemovePopup(_pluginLoadSelectorPopup); }
            if (_filePathTextField != null) { _mvrScript.RemoveTextField(_filePathTextField); }
            if (_selectFileButton != null) { _mvrScript.RemoveButton(_selectFileButton); }
            if (_copyButtonButton != null) { _mvrScript.RemoveButton(_copyButtonButton); }

            if (_buttonLabelOnTextField != null) { _mvrScript.RemoveTextField(_buttonLabelOnTextField); }
            if (_buttonLabelOffTextField != null) { _mvrScript.RemoveTextField(_buttonLabelOffTextField); }
            if (_singleInstancePerAtomToggle != null) { _mvrScript.RemoveToggle(_singleInstancePerAtomToggle); }
            if (_pluginSaveDataSelectorPopup != null) { _mvrScript.RemovePopup(_pluginSaveDataSelectorPopup); }
            if (_pluginForceMergeLoadToggle != null) { _mvrScript.RemoveToggle(_pluginForceMergeLoadToggle); }
            if (_pluginSaveDataTextField != null) { _mvrScript.RemoveTextField(_pluginSaveDataTextField); }
            if (_useButtonTextureToggle != null) { _mvrScript.RemoveToggle(_useButtonTextureToggle); }
            if (_targetParamNameOnPopup != null) { _mvrScript.RemovePopup(_targetParamNameOnPopup); }
            if (_targetParamNameOffPopup != null) { _mvrScript.RemovePopup(_targetParamNameOffPopup); }
            if (_pluginSaveDataSelectorPopup != null) { _mvrScript.RemovePopup(_pluginSaveDataSelectorPopup); }
            if (_targetPluginTypePopup != null) { _mvrScript.RemovePopup(_targetPluginTypePopup); }
            if (_suppressClothingLoadToggle != null) { _mvrScript.RemoveToggle(_suppressClothingLoadToggle); }
            if (_suppressPersonScaleLoadToggle != null) { _mvrScript.RemoveToggle(_suppressPersonScaleLoadToggle); }
            if (_suppressPersonScaleLoadToggle != null) { _mvrScript.RemoveToggle(_suppressPersonScaleLoadToggle); }
            if (_autoLabelToggle != null) { _mvrScript.RemoveToggle(_autoLabelToggle); }
            if (_closeGameUIonPlayModeToggle != null) { _mvrScript.RemoveToggle(_closeGameUIonPlayModeToggle); }
            if (_openGameUIonEditModeToggle != null) { _mvrScript.RemoveToggle(_openGameUIonEditModeToggle); }

            if (_spacer1 != null) { _mvrScript.RemoveSpacer(_spacer1); }
            if (_spacer1b != null) { _mvrScript.RemoveSpacer(_spacer1b); }
            if (_spacer2 != null) { _mvrScript.RemoveSpacer(_spacer2); }
            if (_spacer2b != null) { _mvrScript.RemoveSpacer(_spacer2b); }
            if (_spacer3 != null) { _mvrScript.RemoveSpacer(_spacer3); }
            if (_gridDeleteButton != null) { _mvrScript.RemoveButton(_gridDeleteButton); }
            if (_buttonSkinDisplay != null) _buttonSkinDisplay.ClearDisplay();
            if (_pluginVariablesDisplayOn != null) _pluginVariablesDisplayOn.ClearDisplay();
            if (_pluginVariablesDisplayOff != null) _pluginVariablesDisplayOff.ClearDisplay();

            _buttonCategoryPopup = _mvrScript.CreateSetupPopup(_buttonCategoryJSON, true, 500f, ButtonCategoryPopupOpen, "Button " + _selectedButton._buttonRef.ToString() + " Category");
            _buttonTypePopup = _mvrScript.CreateSetupPopup(_buttonTypeJSON, true, 500f, UpdateButtonTypeChoices, "Button " + _selectedButton._buttonRef.ToString() + " Operation");

            if (_selectedButton._buttonType != "Blank") {
                _autoLabelToggle = _mvrScript.CreateToggle(_autoLabelJSON, true);
                _autoLabelToggle.label = "Auto Label Button " + _selectedButton._buttonRef.ToString();
                if (!_selectedButton._autoLabel) {
                    _buttonLabelOnTextField = _mvrScript.CreateTextInput(_buttonLabelOnJSON, true);
                    if (_selectedButton._maxButtonStates > 1) {
                        _buttonLabelOffTextField = _mvrScript.CreateTextInput(_buttonLabelOffJSON, true);
                    }
                }

                if (_selectedButton._buttonType == "Toggle Play/Edit Mode")
                {
                    _spacer1 = _mvrScript.CreateSpacer(true);
                    _spacer1.height = 10;
                    _closeGameUIonPlayModeToggle = _mvrScript.CreateToggle(_closeGameUIonPlayModeJSON, true);
                    _closeGameUIonPlayModeToggle.label = "Close Game UI for Play Mode";
                    _openGameUIonEditModeToggle = _mvrScript.CreateToggle(_openGameUIonEditModeJSON, true);
                    _openGameUIonEditModeToggle.label = "Open Game UI for Edit Mode";
                }

                if (_selectedButton.IsAtomTargetable())
                {
                    _spacer1 = _mvrScript.CreateSpacer(true);
                    _spacer1.height = 10;
                    _targetCategoryPopup = _mvrScript.CreateSetupPopup(_targetCategoryJSON, true, 500f, UpdateAtomCategoryChoices, "Target Category");
                    _targetNamePopup = _mvrScript.CreateSetupPopup(_targetNameJSON, true, 500f, UpdateAtomChoices, "Button Target");
                }
                SuperController.singleton.GetAtoms();

                if (_selectedButton._buttonCategory == "Plugins" && _selectedButton._buttonType != "Load Plugin(s)" && _selectedButton._buttonType != "Unload All Plugins") {
                    _spacer1b = _mvrScript.CreateSpacer(true);
                    _spacer1b.height = 10;
                    _targetPluginTypePopup = _mvrScript.CreateSetupPopup(_targetPluginTypeJSON, true, 300f, UpdatePluginTypeChoices, "Plugin Type");
                    if (_selectedButton._pluginType != "") {
                        if (_selectedButton._buttonType == "Plugin Action Toggle" || _selectedButton._buttonType == "Plugin Action") {
                            if (_selectedButton._buttonType == "Plugin Action Toggle")
                            {
                                _targetParamNameOnPopup = _mvrScript.CreateSetupPopup(_targetParamNameOnJSON, true, 300f, UpdatePluginParamOnSelectorChoices, "Button ON Plugin Action");
                                if (_targetParamNameOnJSON.val !="No Actions available") _pluginVariablesDisplayOn.RefreshDisplay("Button ON ");
                            }
                            else
                            {
                                _targetParamNameOnPopup = _mvrScript.CreateSetupPopup(_targetParamNameOnJSON, true, 300f, UpdatePluginParamOnSelectorChoices, "Select Plugin Action");
                                if (_targetParamNameOnJSON.val != "No Actions available") _pluginVariablesDisplayOn.RefreshDisplay("");
                            }
                           
                            
                        }

                        else if (_selectedButton._buttonType == "Plugin Bool Toggle" || _selectedButton._buttonType == "Set Plugin Bool True" || _selectedButton._buttonType == "Set Plugin Bool False") {
                            _targetParamNameOnPopup = _mvrScript.CreateSetupPopup(_targetParamNameOnJSON, true, 300f, UpdatePluginParamOnSelectorChoices, "Select Plugin Bool");
                        }
                        if (_selectedButton._buttonType == "Plugin Action Toggle") {
                            _targetParamNameOffPopup = _mvrScript.CreateSetupPopup(_targetParamNameOffJSON, true, 300f, UpdatePluginParamOffSelectorChoices, "Button OFF Plugin Action");
                            if (_targetParamNameOffJSON.val != "No Actions available") _pluginVariablesDisplayOff.RefreshDisplay("Button OFF ");
                        }
                    }
                }

                if (_selectedButton._buttonType == "Load Plugin(s)") {
                    _spacer1b = _mvrScript.CreateSpacer(true);
                    _spacer1b.height = 10;

                    if (Constants._patreonContentEnabled) {
                        _pluginLoadSelectorPopup = _mvrScript.CreateSetupPopup(_pluginLoadSelectorJSON, true, 500f, PluginLoadSelectorPopupOpen, null);
                        _pluginLoadSelectorPopup.labelTextColor = Color.blue;
                    }

                    _selectFileButton = _mvrScript.CreateButton("Select Plugin File", true);
                    List<ShortCut> shortcuts = FileManagerSecure.GetShortCutsForDirectory("Custom\\Scripts");
                    _selectFileButton.button.onClick.AddListener(() => { SuperController.singleton.GetMediaPathDialog(FileSelected, "cs|cslist", "Custom\\Scripts", false, true, false, null, false, shortcuts); });
                    _filePathTextField = _mvrScript.CreateTextField(_filePathJSON, true);
                    _filePathTextField.height = 1;
                    _pluginForceMergeLoadToggle = _mvrScript.CreateToggle(_pluginForceMergeLoadJSON, true);
                    _pluginForceMergeLoadToggle.label = "Force Merge Load";

                    if (_selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._filePath != "") {
                        _singleInstancePerAtomToggle = _mvrScript.CreateToggle(_singleInstancePerAtomJSON, true);
                        _singleInstancePerAtomToggle.label = "Limit to 1 plugin instance per Atom";
                        if (Constants._patreonContentEnabled && _selectedButton._targetName != "CoreControl") {
                            _pluginSaveDataSelectorPopup = _mvrScript.CreateSetupPopup(_pluginSaveDataSelectorJSON, true, 300f, UpdateSaveDataSelectorChoices, "Inital plugin data on Load");

                            _pluginSaveDataTextField = _mvrScript.CreateTextField(_pluginSaveDataJSON, true);
                        }
                    }
                }
                if (_selectedButton._buttonType.Substring(_selectedButton._buttonType.Length - 6, 6) == "Preset") {
                    _spacer1b = _mvrScript.CreateSpacer(true);
                    _spacer1b.height = 10;
                    _selectFileButton = _mvrScript.CreateButton("Select Preset File", true);
                    List<ShortCut> shortcuts = FileManagerSecure.GetShortCutsForDirectory(GetDefaultPresetFolder());
                    _selectFileButton.button.onClick.AddListener(() => { SuperController.singleton.GetMediaPathDialog(FileSelected, "vap", GetDefaultPresetFolder(), false, true, false, null, false, shortcuts); });
                    _filePathTextField = _mvrScript.CreateTextField(_filePathJSON, true);
                    _filePathTextField.height = 1;
                    if (Constants._patreonContentEnabled)
                    {
                        if (_selectedButton._buttonType == "Load App Preset")
                        {
                            _suppressClothingLoadToggle = _mvrScript.CreateToggle(_suppressClothingLoadJSON, true);
                            _suppressClothingLoadToggle.label = "Suppress Clothing Load";
                            _suppressPersonScaleLoadToggle = _mvrScript.CreateToggle(_suppressPersonScaleLoadJSON, true);
                            _suppressPersonScaleLoadToggle.label = "Suppress Person Scale Load";
                        }
                        _useButtonTextureToggle = _mvrScript.CreateToggle(_useButtonTextureJSON, true);
                        _useButtonTextureToggle.label = "Use Preset Thumbnail image";
                        if (_selectedButton._buttonType == "Merge Clothing Preset")
                        {
                            _tagItemsToRemoveSelectorPopup = _mvrScript.CreateSetupPopup(_tagItemsToRemoveSelectorJSON, true, 600f, UpdateTagItemChoices, null);
                        }
                    }
                }
                if (Constants._patreonContentEnabled) { _buttonSkinDisplay.RefreshDisplay(); }
            }
            Button prevButton = _mvrScript._buttonGridDisplay._selectedButtonGrid.GetButtonBefore(_selectedButton._buttonRef);
            if (prevButton != null) {
                _spacer2 = _mvrScript.CreateSpacer(true);
                _spacer2.height = 10;
                _copyButtonButton = _mvrScript.CreateButton("Copy all Button " + prevButton._buttonRef + " values to Button " + _selectedButton._buttonRef, true);
                _copyButtonButton.buttonColor = Color.yellow;
                _copyButtonButton.textColor = Color.black;
                _copyButtonButton.button.onClick.AddListener(() => {
                    _selectedButton.CopyFrom(prevButton);
                    _mvrScript._buttonGridDisplay._selectedButtonGrid.RecalcActiveButtonCount();
                    _mvrScript._buttonGridDisplay._selectedButtonGrid.RecalcGazeSelections();
                    _mvrScript.RefreshUIButtonGrid();
                    RefreshDisplay(); });
            }


            if (_mvrScript._buttonGridsList.Count > 1) {
                _spacer2b = _mvrScript.CreateSpacer(true);
                _spacer2b.height = 25;
                _gridDeleteButton = _mvrScript.CreateButton("Delete Screen " + _mvrScript._buttonGridSelectorJSON.val, true);
                _gridDeleteButton.buttonColor = Color.red;
                _gridDeleteButton.textColor = Color.white;

                _gridDeleteButton.button.onClick.AddListener(() => { _mvrScript.DeleteCurrentScreen(); });
            }
            _spacer3 = _mvrScript.CreateSpacer(true);
            _spacer3.height = 400;
            
        }
        public void ButtonCategoryPopupOpen() { _mvrScript.CloseAllRightUIPopups(_buttonCategoryPopup); }
        public void PluginLoadSelectorPopupOpen() {
            _mvrScript.CloseAllRightUIPopups(_pluginLoadSelectorPopup);
            List<string> popupChoices = new List<string>();

            for (int i = 1; i <= _selectedButton._pluginLoadList.Count; i++)
            {
                popupChoices.Add(i.ToString());
            }

            popupChoices.Add("Add Plugin...");
            if (_selectedButton._pluginLoadList.Count>1) popupChoices.Add("Remove Plugin "+ _pluginLoadSelectorJSON.val);
            _pluginLoadSelectorJSON.choices = popupChoices;
        }

        public void CloseAllPopup(UIDynamicPopup exception) {
            if (_buttonCategoryPopup != null && _buttonCategoryPopup != exception) { _buttonCategoryPopup.popup.visible = false; }
            if (_buttonTypePopup != null && _buttonTypePopup != exception) { _buttonTypePopup.popup.visible = false; }
            if (_targetCategoryPopup != null && _targetCategoryPopup != exception) { _targetCategoryPopup.popup.visible = false; }
            if (_targetNamePopup != null && _targetNamePopup != exception) { _targetNamePopup.popup.visible = false; }
            if (_pluginLoadSelectorPopup != null && _pluginLoadSelectorPopup != exception) { _pluginLoadSelectorPopup.popup.visible = false; }
            if (_pluginSaveDataSelectorPopup != null && _pluginSaveDataSelectorPopup != exception) { _pluginSaveDataSelectorPopup.popup.visible = false; }
            if (_targetParamNameOnPopup != null && _targetParamNameOnPopup != exception) { _targetParamNameOnPopup.popup.visible = false; }
            if (_targetParamNameOffPopup != null && _targetParamNameOffPopup != exception) { _targetParamNameOffPopup.popup.visible = false; }
            if (_tagItemsToRemoveSelectorPopup != null && _tagItemsToRemoveSelectorPopup != exception) { _tagItemsToRemoveSelectorPopup.popup.visible = false; }

            
            if (_targetPluginTypePopup != null && _targetPluginTypePopup != exception) { _targetPluginTypePopup.popup.visible = false; }
            if (_pluginSaveDataSelectorPopup != null && _pluginSaveDataSelectorPopup != exception) { _pluginSaveDataSelectorPopup.popup.visible = false; }
            if (_buttonSkinDisplay != null && _buttonSkinDisplay._buttonSkinParamSelectorPopup != null && _buttonSkinDisplay._buttonSkinParamSelectorPopup != exception) { _buttonSkinDisplay._buttonSkinParamSelectorPopup.popup.visible = false; }
            if (_buttonSkinDisplay != null && _buttonSkinDisplay._skinStringChooserParamPopup != null && _buttonSkinDisplay._skinStringChooserParamPopup != exception) { _buttonSkinDisplay._skinStringChooserParamPopup.popup.visible = false; }
            if (_pluginVariablesDisplayOn != null) _pluginVariablesDisplayOn.CloseAllPopup(exception);
            if (_pluginVariablesDisplayOff != null) _pluginVariablesDisplayOff.CloseAllPopup(exception);

        }
        protected string GetDefaultPresetFolder() {
            string presetFolder = "";
            switch (_selectedButton._buttonType) {
                case "Load App Preset":
                    presetFolder = "Custom\\Atom\\Person\\Appearance";
                    break;
                case "Load Pose Preset":
                    presetFolder = "Custom\\Atom\\Person\\Pose";
                    break;
                case "Load Hair Preset":
                    presetFolder = "Custom\\Atom\\Person\\Hair";
                    break;
                case "Load Glute Preset":
                    presetFolder = "Custom\\Atom\\Person\\GlutePhysics";
                    break;
                case "Load Breast Preset":
                    presetFolder = "Custom\\Atom\\Person\\BreastPhysics";
                    break;
                case "Load General Preset":
                    presetFolder = "Custom\\Atom\\Person\\General";
                    break;
                case "Load Skin Preset":
                    presetFolder = "Custom\\Atom\\Person\\Skin";
                    break;
                case "Load Morph Preset":
                    presetFolder = "Custom\\Atom\\Person\\Morphs";
                    break;
                case "Load Clothing Preset":
                case "Merge Clothing Preset":
                case "Remove Clothing Preset":
                case "Undress Cloth Preset":
                case "Reset Sim Cloth Preset":
                    presetFolder = "Custom\\Atom\\Person\\Clothing";
                    break;
            }
            return (presetFolder);
        }
        protected void FileSelected(string fileName) {
            try {
                if (fileName != "") {
                    if (fileName.IndexOf(':') == 1) {
                        _filePathJSON.val = fileName.Substring(Constants._vamFilePath.Length, fileName.Length - Constants._vamFilePath.Length); }
                    else { _filePathJSON.val = fileName; }

                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        private void UpdateTagItemChoices()
        {
            try
            {
                _mvrScript.CloseAllRightUIPopups(_tagItemsToRemoveSelectorPopup);
                List<string> popupChoices = new List<string>();
                foreach (string tag in _ClothingTags)
                {
                    if (_selectedButton._tagsToRemoveOnMerge.Contains(tag))
                    {
                        popupChoices.Add("+ "+tag+" +");
                    }
                    else popupChoices.Add("  "+tag+"  ");

                }


                _tagItemsToRemoveSelectorJSON.choices = popupChoices;
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        private void UpdatePluginTypeChoices() {
            try {
                _mvrScript.CloseAllRightUIPopups(_targetPluginTypePopup);
                List<string> popupChoices = new List<string>();

                foreach (Atom atom in SuperController.singleton.GetAtoms()) {
                    foreach (string storableName in atom.GetStorableIDs()) {
                        if (storableName.StartsWith("plugin#")) {
                            string pluginType = storableName.Substring(storableName.IndexOf('_') + 1);
                            if (!popupChoices.Contains(pluginType)) { popupChoices.Add(pluginType); }
                        }
                    }
                }

                if (popupChoices.Count == 0) { popupChoices.Add("No plugins in Scene"); }
                _targetPluginTypeJSON.choices = popupChoices;
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdatePluginParamOnSelectorChoices() {
            try {
                _mvrScript.CloseAllRightUIPopups(_targetParamNameOnPopup);
                if (_buttonTypeJSON.val.Contains("Bool")) { _targetParamNameOnJSON.choices = _selectedButton.GetTargetPluginParamNameChoices("BOOL"); }
                else { _targetParamNameOnJSON.choices = _selectedButton.GetTargetPluginParamNameChoices("ACTION"); }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdatePluginParamOffSelectorChoices() {
            try {
                _mvrScript.CloseAllRightUIPopups(_targetParamNameOffPopup);
                if (_buttonTypeJSON.val.Contains("Bool")) { _targetParamNameOffJSON.choices = _selectedButton.GetTargetPluginParamNameChoices("BOOL"); }
                else { _targetParamNameOffJSON.choices = _selectedButton.GetTargetPluginParamNameChoices("ACTION"); }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdateSaveDataSelectorChoices() {
            try {
                _mvrScript.CloseAllRightUIPopups(_pluginSaveDataSelectorPopup);
                List<string> pluginChoices = new List<string>();
                foreach (Atom atom in SuperController.singleton.GetAtoms()) {
                    List<string> pluginRefsInAtom = new List<string>();
                    pluginRefsInAtom = _selectedButton.GetPluginRefsOfType(atom, _filePathJSON.val);
                    foreach (string pluginRef in pluginRefsInAtom) {
                        pluginChoices.Add(pluginRef + "[" + atom.name + "]");
                    }
                }
                if (_pluginSaveDataJSON.val != "" && _pluginSaveDataJSON.val != null) { pluginChoices.Add("<Clear Plugin Data>"); }
                if (pluginChoices.Count == 0) { pluginChoices.Add("<None available>"); }
                _pluginSaveDataSelectorJSON.choices = pluginChoices;
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        private void UpdateButtonTypeChoices() {
            try {
                _mvrScript.CloseAllRightUIPopups(_buttonTypePopup);
                _buttonTypeJSON.choices = _selectedButton.GetButtonTypeChoices();
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdateAtomCategoryChoices() {
            try {
                _mvrScript.CloseAllRightUIPopups(_targetCategoryPopup);
                _targetCategoryJSON.choices = _selectedButton.GetAtomCategoryChoices();
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void UpdateAtomChoices() {
            try {
                _mvrScript.CloseAllRightUIPopups(_targetNamePopup);
                _targetNameJSON.choices = _selectedButton.GetTargetChoices();
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void RefreshDisplayValues() {
            _isDisplayUpdating = true;
            _buttonCategoryJSON.val = _selectedButton._buttonCategory;
            _buttonTypeJSON.val = _selectedButton._buttonType;
            if (_selectedButton._buttonLabelOn == "") {
                if (_selectedButton._maxButtonStates > 1) { _buttonLabelOnJSON.val = "<Button " + _selectedButton._buttonRef.ToString() + " Toggle On Label>"; }
                else { _buttonLabelOnJSON.val = "<Button " + _selectedButton._buttonRef.ToString() + " Label>"; }
            }
            else { _buttonLabelOnJSON.val = _selectedButton._buttonLabelOn; }
            if (_selectedButton._buttonLabelOff == "") { _buttonLabelOffJSON.val = "<Button " + _selectedButton._buttonRef.ToString() + " Toggle Off Label>"; }
            else { _buttonLabelOffJSON.val = _selectedButton._buttonLabelOff; }
            _targetCategoryJSON.val = _selectedButton._targetCategory;
            _targetNameJSON.val = _selectedButton._targetName;
            if (_selectedButton._buttonType == "Load Plugin(s)") {
                if (int.Parse(_pluginLoadSelectorJSON.val) > _selectedButton._pluginLoadList.Count) _pluginLoadSelectorJSON.val = _selectedButton._pluginLoadList.Count.ToString();
                _filePathJSON.val = _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._filePath;
                _pluginSaveDataJSON.val = _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._pluginSaveData;
                _singleInstancePerAtomJSON.val = _selectedButton._pluginLoadList[int.Parse(_pluginLoadSelectorJSON.val) - 1]._singleInstancePerAtom;
            }
            else {
                if (_selectedButton._buttonType.Length > 6 && _selectedButton._buttonType.Substring(_selectedButton._buttonType.Length - 6, 6) == "Preset") {
                    _filePathJSON.val = _selectedButton._presetFilePath;
                }
            }
            if (_pluginSaveDataSelectorPopup != null) { _pluginSaveDataSelectorPopup.popup.visible = false; }
            _pluginSaveDataSelectorJSON.val = "Select...";
            _targetPluginTypeJSON.val = _selectedButton._pluginType;
            _pluginForceMergeLoadJSON.val = _selectedButton._pluginForceMergeLoad;

            _targetParamNameOnJSON.val = _selectedButton._targetParamNameOn;
            _targetParamNameOffJSON.val = _selectedButton._targetParamNameOff;
            _suppressClothingLoadJSON.val = _selectedButton._suppressClothingLoad;
            _suppressPersonScaleLoadJSON.val = _selectedButton._suppressPersonScaleLoad;
            _autoLabelJSON.val = _selectedButton._autoLabel;
            _useButtonTextureJSON.val = _selectedButton._useButtonTexture;

            _tagItemsToRemoveSelectorJSON.val = "Toggle Clothing Tag...";

            _isDisplayUpdating = false;
        }
    }

    public class ButtonPluginLoad {
        public string _filePath = "";
        public bool _singleInstancePerAtom = true;
        public string _pluginSaveData = "";
        public JSONArray _pluginSaveJSON;
        
        public JSONClass GetJSON()
        {
            JSONClass node = new JSONClass();
            node["filePath"] = _filePath;
            node["singleInstancePerAtom"].AsBool = _singleInstancePerAtom;
            if (_pluginSaveJSON!=null) node["pluginSaveJSON"] = _pluginSaveJSON;
            return node;
        }
        public void CopyFrom(ButtonPluginLoad bpl) {
            _filePath = bpl._filePath;
            _singleInstancePerAtom = bpl._singleInstancePerAtom;
            _pluginSaveData = bpl._pluginSaveData;
            _pluginSaveJSON = bpl._pluginSaveJSON;
        }

        public void LoadJSON(JSONClass pluginLoadJSON) {

            _pluginSaveJSON = new JSONArray();

            if (pluginLoadJSON["filePath"] != null) _filePath = pluginLoadJSON["filePath"];
            if (pluginLoadJSON["pluginSaveData"] != null)
            {
                string tempJSONString = pluginLoadJSON["pluginSaveData"];
                if (tempJSONString == "")
                {
                    _pluginSaveJSON = null;
                    _pluginSaveData = "";
                }
                else
                {
                    _pluginSaveJSON = (JSONArray)JSON.Parse("[" + tempJSONString + "]");
                    _pluginSaveData = tempJSONString.Substring(0, Math.Min(tempJSONString.Length, 500));
                }

            }
            else
            {
                if (pluginLoadJSON["pluginSaveJSON"] != null)
                {
                    _pluginSaveJSON = (JSONArray)pluginLoadJSON["pluginSaveJSON"];
                    _pluginSaveData = _pluginSaveJSON.ToString().Substring(0, Math.Min(_pluginSaveJSON.ToString().Length, 500));
                }

            }
            if (pluginLoadJSON["singleInstancePerAtom"] != null) _singleInstancePerAtom = pluginLoadJSON["singleInstancePerAtom"].AsBool;

        }

    }
    public class PluginVariables
    {
        public Dictionary<string, PluginVariable> _pluginVarDict;
        private Button _parentButton;

        public PluginVariables (Button parent)
        {
            try
            {
                _pluginVarDict = new Dictionary<string, PluginVariable>();
                _parentButton = parent;
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void CopyFrom(PluginVariables pv)
        {
            if (pv != null)
            {
                _pluginVarDict = new Dictionary<string, PluginVariable>();
                foreach (KeyValuePair<string, PluginVariable> kvp in pv._pluginVarDict) _pluginVarDict.Add(kvp.Key, kvp.Value);
            }

        }
        public JSONClass GetJSON()
        {
            JSONClass node = new JSONClass();
            JSONArray pluginVariablesArray = new JSONArray();

            foreach (KeyValuePair<string, PluginVariable> kvp in _pluginVarDict)
            {
                JSONClass variableNode = new JSONClass();
                variableNode["pluginVarName"] = kvp.Key;
                variableNode["pluginVarType"] = kvp.Value._pluginVarType;

                switch (kvp.Value._pluginVarType)
                {
                    case "BOOL":
                        variableNode["pluginVarBoolValue"].AsBool = kvp.Value._pluginBoolValue;
                        break;
                    case "FLOAT":
                        variableNode["pluginVarFloatValue"].AsFloat = kvp.Value._pluginFloatValue;
                        break;
                    case "STRING":
                    case "STRINGCHOOSER":
                    case "URL":
                        variableNode["pluginVarStringValue"] = kvp.Value._pluginStringValue;
                        break;
                    case "COLOR":
                        variableNode["pluginVarColorHValue"].AsFloat = kvp.Value._pluginColorValue.H ;
                        variableNode["pluginVarColorSValue"].AsFloat = kvp.Value._pluginColorValue.S;
                        variableNode["pluginVarColorVValue"].AsFloat = kvp.Value._pluginColorValue.V;

                        break;
                }

                pluginVariablesArray.Add(variableNode);
            }

            node["pluginVariables"] = pluginVariablesArray;

            return node;
        }

        public void LoadJSON(JSONClass pluginVarJSON)
        {
            _pluginVarDict = new Dictionary<string, PluginVariable>();
            
            if (pluginVarJSON["pluginVariables"] != null)
            {
                JSONArray pluginVariablesArray = pluginVarJSON["pluginVariables"].AsArray;
                for (int i = 0; i < pluginVariablesArray.Count; i++)
                {
                    JSONClass variableNodeJSON = pluginVariablesArray[i].AsObject;
                    PluginVariable pluginVariable = new PluginVariable() ;
                    pluginVariable._pluginVarType = variableNodeJSON["pluginVarType"];

                    switch (pluginVariable._pluginVarType)
                    {
                        case "BOOL":
                            pluginVariable._pluginBoolValue = variableNodeJSON["pluginVarBoolValue"].AsBool;
                            break;
                        case "FLOAT":
                            pluginVariable._pluginFloatValue = variableNodeJSON["pluginVarFloatValue"].AsFloat;
                            break;
                        case "STRING":
                        case "STRINGCHOOSER":
                        case "URL":
                            pluginVariable._pluginStringValue = variableNodeJSON["pluginVarStringValue"];
                            break;
                        case "COLOR":
                            pluginVariable._pluginColorValue = new HSVColor();
                            pluginVariable._pluginColorValue.H = variableNodeJSON["pluginVarColorHValue"].AsFloat;
                            pluginVariable._pluginColorValue.S = variableNodeJSON["pluginVarColorSValue"].AsFloat;
                            pluginVariable._pluginColorValue.V = variableNodeJSON["pluginVarColorVValue"].AsFloat;
                            break;
                    }

                    _pluginVarDict.Add(variableNodeJSON["pluginVarName"], pluginVariable);
                }
            }
        }
        public List<string> GetPluginParamChoices()
        {
            List<string> paramChoices = new List<string>();
            List<string> displayParamChoices = new List<string>();
            paramChoices = _parentButton.GetTargetPluginParamNameChoices("ALL");

            foreach (string choice in paramChoices)
            {
 //               SuperController.LogMessage(choice);
                if (_pluginVarDict.ContainsKey(choice))
                {
                    displayParamChoices.Add("+ " + choice + " +");
                }
                else displayParamChoices.Add("  " + choice + "  ");

            }

            return (displayParamChoices);
        }

        public string GetVarType(string varName)
        {
            string varType = "NONE";

            if (_pluginVarDict.ContainsKey(varName))
            {
                varType = _pluginVarDict[varName]._pluginVarType;
            }

            return (varType);
        }
        public bool GetVarSet(string varName)
        {
            return (_pluginVarDict.ContainsKey(varName));
        }

        public string GetVarTypeFromPlugin (string varName)
        {
            string varType = _parentButton.GetVarTypeFromPlugin(varName);

            return (varType);
        }
        public float GetMinFloatValue(string varName)
        {
            float minFloat = _parentButton.GetMinFloatValue(varName);
            if (_pluginVarDict.ContainsKey(varName))
            {
                float varValue = _pluginVarDict[varName]._pluginFloatValue;
                if (varValue < minFloat) minFloat = varValue;
            }
            return (minFloat);
        }
        public float GetMaxFloatValue(string varName)
        {
            float maxFloat = _parentButton.GetMaxFloatValue(varName);
            if (_pluginVarDict.ContainsKey(varName))
            {
                float varValue = _pluginVarDict[varName]._pluginFloatValue;
                if (varValue > maxFloat) maxFloat = varValue;
            }
            return (maxFloat);
        }
        public List<string> GetStringChooserChoices(string varName)
        {
            List<string> choices = _parentButton.GetStringChooserChoices(varName);
            if (choices.Count < 1) choices.Add("NO VALUES AVAILABLE");
            return (choices);
        }
    }
    public class PluginVariablesDisplay
    {
        private UIAssist _mvrScript;
        public PluginVariables _selectButtonPluginVariables;

        bool _isDisplayUpdating = false;

        public JSONStorableStringChooser _pluginVariableSelectorJSON;
        public JSONStorableBool _pluginVariableSetJSON;
        public JSONStorableFloat _pluginVariableFloatValueJSON;
        public JSONStorableString _pluginVariableStringValueJSON;
        public JSONStorableStringChooser _pluginVariableStringChooserValueJSON;
        public JSONStorableBool _pluginVariableBoolValueJSON;
        public JSONStorableColor _pluginVariableColorValueJSON;

        private UIDynamicPopup _pluginVariableSelectorPopup;
        private UIDynamicToggle _pluginVariableSetToggle;
        private UIDynamicToggle _pluginVariableBoolValueToggle;

        private UIDynamicSlider _pluginVariableFloatValueSlider;
        private UIDynamicColorPicker _pluginVariableColorValueColorPicker;
        private UIDynamicTextField _pluginVariableStringValueTextField;
        private UIDynamicPopup _pluginVariableStringValuePopup;

        public PluginVariablesDisplay( UIAssist mvrScript, PluginVariables pluginVariables)
        {
            try
            {
                _selectButtonPluginVariables = pluginVariables;
                _mvrScript = mvrScript;

                HSVColor hsvColor; hsvColor.H = 0f; hsvColor.S = 0f; hsvColor.V = 0f;

                _pluginVariableSelectorJSON = new JSONStorableStringChooser("pluginVariableChooser", null, "Select Parameter...", "", SwitchPluginVariable);
                _pluginVariableSetJSON = new JSONStorableBool("pluginVariableSet",false, SwitchPluginVariableSet);
                _pluginVariableFloatValueJSON = new JSONStorableFloat("pluginVariableFloat", 0f, SwitchFloatValue, 0f, 1f, true);
                _pluginVariableStringValueJSON = new JSONStorableString("pluginVariableString","",SwitchStringValue);
                _pluginVariableStringChooserValueJSON = new JSONStorableStringChooser("pluginVariableStringChooser", null, "", "", SwitchStringValue);
                _pluginVariableBoolValueJSON = new JSONStorableBool("pluginVariableBoolValue", true, SwitchBoolValue);
                _pluginVariableColorValueJSON = new JSONStorableColor("pluginVariableColorValue", hsvColor, SwitchColorValue);

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }

        }
        private void SwitchPluginVariable(string stringVal)
        {
            try
            {
                _mvrScript.CloseAllRightUIPopups(_pluginVariableSelectorPopup);
                if (!_isDisplayUpdating)
                {
                    _mvrScript.RefreshRightUIDisplay();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchPluginVariableSet(bool setVal)
        {
            try
            {
                string pluginVariable = _pluginVariableSelectorJSON.val.Substring(2, _pluginVariableSelectorJSON.val.Length - 4);
                if (setVal && !_selectButtonPluginVariables._pluginVarDict.ContainsKey(pluginVariable) )
                {
                    PluginVariable pluginVar = new PluginVariable();
                    pluginVar._pluginVarType = _selectButtonPluginVariables.GetVarTypeFromPlugin(pluginVariable);

                    _selectButtonPluginVariables._pluginVarDict.Add(pluginVariable, pluginVar);
                }
                else if (!setVal && _selectButtonPluginVariables._pluginVarDict.ContainsKey(pluginVariable))
                {
                    _selectButtonPluginVariables._pluginVarDict.Remove(pluginVariable);
                }
                if (!_isDisplayUpdating)
                {
                    _mvrScript.RefreshRightUIDisplay();
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchFloatValue(float floatVal)
        {
            try
            {
                string pluginVariable = _pluginVariableSelectorJSON.val.Substring(2, _pluginVariableSelectorJSON.val.Length - 4);
                if (!_isDisplayUpdating)
                {
                    if (_selectButtonPluginVariables._pluginVarDict.ContainsKey(pluginVariable))
                    {
                        _selectButtonPluginVariables._pluginVarDict[pluginVariable]._pluginFloatValue = floatVal;
                    }
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchStringValue(string stringVal)
        {
            try
            {
                if (_pluginVariableStringValuePopup != null) _mvrScript.CloseAllRightUIPopups(_pluginVariableStringValuePopup);
                string pluginVariable = _pluginVariableSelectorJSON.val.Substring(2, _pluginVariableSelectorJSON.val.Length - 4);
                if (!_isDisplayUpdating)
                {
                    if (_selectButtonPluginVariables._pluginVarDict.ContainsKey(pluginVariable))
                    {
                        _selectButtonPluginVariables._pluginVarDict[pluginVariable]._pluginStringValue = stringVal;
                    }
                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        private void SwitchBoolValue(bool boolVal)
        {
            try
            {
                string pluginVariable = _pluginVariableSelectorJSON.val.Substring(2, _pluginVariableSelectorJSON.val.Length - 4);
                if (!_isDisplayUpdating)
                {
                    if (_selectButtonPluginVariables._pluginVarDict.ContainsKey(pluginVariable))
                    {
                        _selectButtonPluginVariables._pluginVarDict[pluginVariable]._pluginBoolValue = boolVal;
                    }

                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchColorValue(JSONStorableColor colorVal)
        {
            try
            {
                string pluginVariable = _pluginVariableSelectorJSON.val.Substring(2, _pluginVariableSelectorJSON.val.Length - 4);
                if (!_isDisplayUpdating)
                {
                    if (_selectButtonPluginVariables._pluginVarDict.ContainsKey(pluginVariable))
                    {
                        _selectButtonPluginVariables._pluginVarDict[pluginVariable]._pluginColorValue = colorVal.val;
                    }

                }
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void CloseAllPopup(UIDynamicPopup exception)
        {
            if (_pluginVariableSelectorPopup != null && _pluginVariableSelectorPopup != exception) { _pluginVariableSelectorPopup.popup.visible = false; }
            if (_pluginVariableStringValuePopup != null && _pluginVariableStringValuePopup != exception) { _pluginVariableStringValuePopup.popup.visible = false; }
        }
        public void ClearDisplay()
        {
            if (_pluginVariableSelectorPopup != null) { _mvrScript.RemovePopup(_pluginVariableSelectorPopup); }
            if (_pluginVariableStringValuePopup != null) { _mvrScript.RemovePopup(_pluginVariableStringValuePopup); }
            if (_pluginVariableStringValueTextField != null) { _mvrScript.RemoveTextField(_pluginVariableStringValueTextField); }
            if (_pluginVariableFloatValueSlider != null) { _mvrScript.RemoveSlider(_pluginVariableFloatValueSlider); }
            if (_pluginVariableColorValueColorPicker != null) { _mvrScript.RemoveColorPicker(_pluginVariableColorValueColorPicker); }
            if (_pluginVariableSetToggle != null) { _mvrScript.RemoveToggle(_pluginVariableSetToggle); }

            if (_pluginVariableBoolValueToggle != null) { _mvrScript.RemoveToggle(_pluginVariableBoolValueToggle); }
            
        }
        public void RefreshDisplay(string labelPrefix)
        {
            ClearDisplay();
            _isDisplayUpdating = true;

            _pluginVariableSelectorPopup = _mvrScript.CreateSetupPopup(_pluginVariableSelectorJSON, true, 500f, SkinParamSelectorPopupOpen, labelPrefix+"Plugin Settings");
            _pluginVariableSelectorPopup.labelTextColor = Color.blue;
            if (_pluginVariableSelectorJSON.val!="Select Parameter...")
            {
                string pluginVariable = _pluginVariableSelectorJSON.val.Substring(2, _pluginVariableSelectorJSON.val.Length - 4);
                bool isSet = _selectButtonPluginVariables.GetVarSet(pluginVariable);
                _pluginVariableSetJSON.val = isSet;
                _pluginVariableSetToggle = _mvrScript.CreateToggle(_pluginVariableSetJSON, true);
                _pluginVariableSetToggle.label = "Initialise Setting value pre-Action";
                if (isSet)
                {
                    string varType = _selectButtonPluginVariables.GetVarType(pluginVariable);
                    
                    switch (varType)
                    {
                        case "FLOAT":                           
                            _pluginVariableFloatValueJSON.min = _selectButtonPluginVariables.GetMinFloatValue(pluginVariable);
                            _pluginVariableFloatValueJSON.max = _selectButtonPluginVariables.GetMaxFloatValue(pluginVariable);
                            _pluginVariableFloatValueJSON.val = _selectButtonPluginVariables._pluginVarDict[pluginVariable]._pluginFloatValue;
                            _pluginVariableFloatValueJSON.constrained = false;
                            _pluginVariableFloatValueSlider = _mvrScript.CreateSlider(_pluginVariableFloatValueJSON, true);
                            _pluginVariableFloatValueSlider.label = pluginVariable;
                            break;
                        case "BOOL":
                            _pluginVariableBoolValueJSON.val = _selectButtonPluginVariables._pluginVarDict[pluginVariable]._pluginBoolValue;
                            
                            _pluginVariableBoolValueToggle = _mvrScript.CreateToggle(_pluginVariableBoolValueJSON, true);
                            _pluginVariableBoolValueToggle.label = pluginVariable;
                            break;
                        case "STRINGCHOOSER":
                            _pluginVariableStringChooserValueJSON.val = _selectButtonPluginVariables._pluginVarDict[pluginVariable]._pluginStringValue;
                            _pluginVariableStringChooserValueJSON.choices = _selectButtonPluginVariables.GetStringChooserChoices(pluginVariable);
                            if (!_pluginVariableStringChooserValueJSON.choices.Contains(_pluginVariableStringChooserValueJSON.val))
                            {
                                _pluginVariableStringChooserValueJSON.val = _pluginVariableStringChooserValueJSON.choices[0];
                                _selectButtonPluginVariables._pluginVarDict[pluginVariable]._pluginStringValue = _pluginVariableStringChooserValueJSON.val;
                            }
                            _pluginVariableStringValuePopup = _mvrScript.CreateSetupPopup(_pluginVariableStringChooserValueJSON,true);
                            _pluginVariableStringValuePopup.label = "Set '"+pluginVariable+"'";
                            break;
                        case "STRING":
                        case "URL":
                            _pluginVariableStringValueJSON.val = _selectButtonPluginVariables._pluginVarDict[pluginVariable]._pluginStringValue;
                           
                            _pluginVariableStringValueTextField = _mvrScript.CreateTextInput(_pluginVariableStringValueJSON, true);
                            break;
                        case "COLOR":
                            _pluginVariableColorValueJSON.val = _selectButtonPluginVariables._pluginVarDict[pluginVariable]._pluginColorValue;
                            _pluginVariableColorValueColorPicker = _mvrScript.CreateColorPicker(_pluginVariableColorValueJSON, true);
                            break;
                    }
                }
            }


            _isDisplayUpdating = false;
        }
        private void SkinParamSelectorPopupOpen()
        {
            _mvrScript.CloseAllRightUIPopups(_pluginVariableSelectorPopup);

            _pluginVariableSelectorJSON.choices = _selectButtonPluginVariables.GetPluginParamChoices();
        }
    }
        public class PluginVariable
    {
        public string _pluginVarType = "NONE";

        public float _pluginFloatValue =0;
        public bool _pluginBoolValue = false;
        public string _pluginStringValue ="";
        public HSVColor _pluginColorValue;
    }
    public class ButtonSkin {
        public float _minFontSize = 4f;
        public float _maxFontSize = 25f;
        public float _minSmallFontSize = 4f;
        public float _maxSmallFontSize = 15f;
        public float _minMediumFontSize = 4f;
        public float _maxMediumFontSize = 25f;
        public float _minLargeFontSize = 4f;
        public float _maxLargeFontSize = 35f;
        public bool _fontSizeDefault = true;
        public HSVColor _textOnColor;
        public bool _textOnColorDefault = true;
        public HSVColor _textOffColor;
        public bool _textOffColorDefault = true;
        public string _textAlignment = "MiddleCenter";
        public bool _textAlignmentDefault = true;
        public HSVColor _buttonOnColor;
        public bool _buttonOnColorDefault = true;
        public HSVColor _buttonOffColor;
        public bool _buttonOffColorDefault = true;
        public string _customButtonOnTextureURL = "";
        public bool _customButtonOnTextureDefault = true;
        public string _customButtonOffTextureURL = "";
        public bool _customButtonOffTextureDefault = true;
        public string _buttonFont = "Arial";
        public bool _buttonFontDefault = true;
        public int _maxButtonStates;
        public string _defaultGazeTarget = "Last viewed Female";
        public Texture2D _buttonOnTexture;
        public Texture2D _buttonOffTexture;
        private UIAssist _mvrScript;
        public HSVColor _buttonGazeSelectColor;

        public ButtonSkin(int buttonStates, UIAssist mvrScript) {
            try {
                _mvrScript = mvrScript;
                _maxButtonStates = buttonStates;
                _buttonGazeSelectColor = new HSVColor();

                Color gazeButtonColor = new Color(128f / 255f, 142f / 255f, 216f / 255f, 1f);
                Color.RGBToHSV(gazeButtonColor, out _buttonGazeSelectColor.H, out _buttonGazeSelectColor.S, out _buttonGazeSelectColor.V);
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        public void CopyFrom(ButtonSkin bs) {
            _minFontSize = bs._minFontSize;
            _maxFontSize = bs._maxFontSize;
            _minSmallFontSize = bs._minSmallFontSize;
            _maxSmallFontSize = bs._maxSmallFontSize;
            _minMediumFontSize = bs._minMediumFontSize;
            _maxMediumFontSize = bs._maxMediumFontSize;
            _minLargeFontSize = bs._minLargeFontSize;
            _maxLargeFontSize = bs._maxLargeFontSize;
            _fontSizeDefault = bs._fontSizeDefault;

            _textOnColor.H = bs._textOnColor.H;
            _textOnColor.S = bs._textOnColor.S;
            _textOnColor.V = bs._textOnColor.V;
            _textOnColorDefault = bs._textOnColorDefault;

            _textOffColor.H = bs._textOffColor.H;
            _textOffColor.S = bs._textOffColor.S;
            _textOffColor.V = bs._textOffColor.V;
            _textOffColorDefault = bs._textOffColorDefault;

            _textAlignment = bs._textAlignment;
            _textAlignmentDefault = bs._textAlignmentDefault;

            _buttonOnColor.H = bs._buttonOnColor.H;
            _buttonOnColor.S = bs._buttonOnColor.S;
            _buttonOnColor.V = bs._buttonOnColor.V;
            _buttonOnColorDefault = bs._buttonOnColorDefault;

            _buttonOffColor.H = bs._buttonOffColor.H;
            _buttonOffColor.S = bs._buttonOffColor.S;
            _buttonOffColor.V = bs._buttonOffColor.V;
            _buttonOffColorDefault = bs._buttonOffColorDefault;

            _customButtonOnTextureURL = bs._customButtonOnTextureURL;
            _customButtonOnTextureDefault = bs._customButtonOnTextureDefault;
            _customButtonOffTextureURL = bs._customButtonOffTextureURL;
            _customButtonOffTextureDefault = bs._customButtonOffTextureDefault;
            _buttonFont = bs._buttonFont;
            _buttonFontDefault = bs._buttonFontDefault;
            _maxButtonStates = bs._maxButtonStates;
            _buttonOnTexture = bs._buttonOnTexture;
            _buttonOffTexture = bs._buttonOffTexture;
            _buttonGazeSelectColor = bs._buttonGazeSelectColor;
            _defaultGazeTarget = bs._defaultGazeTarget;

        }

        public JSONClass GetJSON()
        {
            JSONClass node = new JSONClass();
            node["minFontSize"].AsFloat = _minFontSize;
            node["maxFontSize"].AsFloat = _maxFontSize;
            node["minSmallFontSize"].AsFloat = _minSmallFontSize;
            node["maxSmallFontSize"].AsFloat = _maxSmallFontSize;
            node["minMediumFontSize"].AsFloat = _minMediumFontSize;
            node["maxMediumFontSize"].AsFloat = _maxMediumFontSize;
            node["minLargeFontSize"].AsFloat = _minLargeFontSize;
            node["maxLargeFontSize"].AsFloat = _maxLargeFontSize;

            node["fontSizeDefault"].AsBool = _fontSizeDefault;
            node["textOnColorDefault"].AsBool = _textOnColorDefault;
            node["textOffColorDefault"].AsBool = _textOffColorDefault;
            node["textAlignmentDefault"].AsBool = _textAlignmentDefault;
            node["buttonOnColorDefault"].AsBool = _buttonOnColorDefault;
            node["buttonOffColorDefault"].AsBool = _buttonOffColorDefault;
            node["customButtonOnTextureDefault"].AsBool = _customButtonOnTextureDefault;
            node["customButtonOffTextureDefault"].AsBool = _customButtonOffTextureDefault;
            node["buttonFontDefault"].AsBool = _buttonFontDefault;

            node["textOnColor"]["H"].AsFloat = _textOnColor.H;
            node["textOnColor"]["S"].AsFloat = _textOnColor.S;
            node["textOnColor"]["V"].AsFloat = _textOnColor.V;

            node["textOffColor"]["H"].AsFloat = _textOffColor.H;
            node["textOffColor"]["S"].AsFloat = _textOffColor.S;
            node["textOffColor"]["V"].AsFloat = _textOffColor.V;

            node["buttonOnColor"]["H"].AsFloat = _buttonOnColor.H;
            node["buttonOnColor"]["S"].AsFloat = _buttonOnColor.S;
            node["buttonOnColor"]["V"].AsFloat = _buttonOnColor.V;

            node["buttonOffColor"]["H"].AsFloat = _buttonOffColor.H;
            node["buttonOffColor"]["S"].AsFloat = _buttonOffColor.S;
            node["buttonOffColor"]["V"].AsFloat = _buttonOffColor.V;

            node["textAlignment"] = _textAlignment;
            node["customButtonOnTextureURL"] = _customButtonOnTextureURL;
            node["customButtonOffTextureURL"] = _customButtonOffTextureURL;
            node["buttonFont"] = _buttonFont;
            node["defaultGazeTarget"] = _defaultGazeTarget;

            node["maxButtonStates"].AsInt = _maxButtonStates;

            node["gazeSelectColor"]["H"].AsFloat = _buttonGazeSelectColor.H;
            node["gazeSelectColor"]["S"].AsFloat = _buttonGazeSelectColor.S;
            node["gazeSelectColor"]["V"].AsFloat = _buttonGazeSelectColor.V;

            return node;
        }
        public void LoadJSON(JSONClass buttonSkinJSON) {
            if (buttonSkinJSON["minFontSize"] != null) _minFontSize = buttonSkinJSON["minFontSize"].AsFloat;
            if (buttonSkinJSON["maxFontSize"] != null) _maxFontSize = buttonSkinJSON["maxFontSize"].AsFloat;
            if (buttonSkinJSON["minSmallFontSize"] != null) _minSmallFontSize = buttonSkinJSON["minSmallFontSize"].AsFloat;
            if (buttonSkinJSON["maxSmallFontSize"] != null) _maxSmallFontSize = buttonSkinJSON["maxSmallFontSize"].AsFloat;
            if (buttonSkinJSON["minMediumFontSize"] != null) _minMediumFontSize = buttonSkinJSON["minMediumFontSize"].AsFloat;
            if (buttonSkinJSON["maxMediumFontSize"] != null) _maxMediumFontSize = buttonSkinJSON["maxMediumFontSize"].AsFloat;
            if (buttonSkinJSON["minLargeFontSize"] != null) _minLargeFontSize = buttonSkinJSON["minLargeFontSize"].AsFloat;
            if (buttonSkinJSON["maxLargeFontSize"] != null) _maxLargeFontSize = buttonSkinJSON["maxLargeFontSize"].AsFloat;

            if (buttonSkinJSON["fontSizeDefault"] != null) _fontSizeDefault = buttonSkinJSON["fontSizeDefault"].AsBool;
            if (buttonSkinJSON["textOnColorDefault"] != null) _textOnColorDefault = buttonSkinJSON["textOnColorDefault"].AsBool;
            if (buttonSkinJSON["textOffColorDefault"] != null) _textOffColorDefault = buttonSkinJSON["textOffColorDefault"].AsBool;
            if (buttonSkinJSON["textAlignmentDefault"] != null) _textAlignmentDefault = buttonSkinJSON["textAlignmentDefault"].AsBool;
            if (buttonSkinJSON["buttonOnColorDefault"] != null) _buttonOnColorDefault = buttonSkinJSON["buttonOnColorDefault"].AsBool;
            if (buttonSkinJSON["buttonOffColorDefault"] != null) _buttonOffColorDefault = buttonSkinJSON["buttonOffColorDefault"].AsBool;
            if (buttonSkinJSON["customButtonOnTextureDefault"] != null) _customButtonOnTextureDefault = buttonSkinJSON["customButtonOnTextureDefault"].AsBool;
            if (buttonSkinJSON["customButtonOffTextureDefault"] != null) _customButtonOffTextureDefault = buttonSkinJSON["customButtonOffTextureDefault"].AsBool;
            if (buttonSkinJSON["buttonFontDefault"] != null) _buttonFontDefault = buttonSkinJSON["buttonFontDefault"].AsBool;

            if (buttonSkinJSON["textOnColor"] != null) _textOnColor.H = buttonSkinJSON["textOnColor"]["H"].AsFloat;
            if (buttonSkinJSON["textOnColor"] != null) _textOnColor.S = buttonSkinJSON["textOnColor"]["S"].AsFloat;
            if (buttonSkinJSON["textOnColor"] != null) _textOnColor.V = buttonSkinJSON["textOnColor"]["V"].AsFloat;

            if (buttonSkinJSON["textOffColor"] != null) _textOffColor.H = buttonSkinJSON["textOffColor"]["H"].AsFloat;
            if (buttonSkinJSON["textOffColor"] != null) _textOffColor.S = buttonSkinJSON["textOffColor"]["S"].AsFloat;
            if (buttonSkinJSON["textOffColor"] != null) _textOffColor.V = buttonSkinJSON["textOffColor"]["V"].AsFloat;

            if (buttonSkinJSON["buttonOnColor"] != null) _buttonOnColor.H = buttonSkinJSON["buttonOnColor"]["H"].AsFloat;
            if (buttonSkinJSON["buttonOnColor"] != null) _buttonOnColor.S = buttonSkinJSON["buttonOnColor"]["S"].AsFloat;
            if (buttonSkinJSON["buttonOnColor"] != null) _buttonOnColor.V = buttonSkinJSON["buttonOnColor"]["V"].AsFloat;

            if (buttonSkinJSON["buttonOffColor"] != null) _buttonOffColor.H = buttonSkinJSON["buttonOffColor"]["H"].AsFloat;
            if (buttonSkinJSON["buttonOffColor"] != null) _buttonOffColor.S = buttonSkinJSON["buttonOffColor"]["S"].AsFloat;
            if (buttonSkinJSON["buttonOffColor"] != null) _buttonOffColor.V = buttonSkinJSON["buttonOffColor"]["V"].AsFloat;

            if (buttonSkinJSON["gazeSelectColor"] != null) _buttonGazeSelectColor.H = buttonSkinJSON["gazeSelectColor"]["H"].AsFloat;
            if (buttonSkinJSON["gazeSelectColor"] != null) _buttonGazeSelectColor.S = buttonSkinJSON["gazeSelectColor"]["S"].AsFloat;
            if (buttonSkinJSON["gazeSelectColor"] != null) _buttonGazeSelectColor.V = buttonSkinJSON["gazeSelectColor"]["V"].AsFloat;

            if (buttonSkinJSON["textAlignment"] != null) _textAlignment = buttonSkinJSON["textAlignment"];
            if (buttonSkinJSON["customButtonOnTextureURL"] != null) _customButtonOnTextureURL = buttonSkinJSON["customButtonOnTextureURL"];
            if (buttonSkinJSON["customButtonOffTextureURL"] != null) _customButtonOffTextureURL = buttonSkinJSON["customButtonOffTextureURL"];
            if (buttonSkinJSON["buttonFont"] != null) _buttonFont = buttonSkinJSON["buttonFont"];
            if (buttonSkinJSON["defaultGazeTarget"] != null) _defaultGazeTarget = buttonSkinJSON["defaultGazeTarget"];

            if (buttonSkinJSON["maxButtonStates"] != null) _maxButtonStates = buttonSkinJSON["maxButtonStates"].AsInt;

            if (_customButtonOnTextureURL != "") QueueLoadTexture(_customButtonOnTextureURL, "ON");
            else { _buttonOnTexture = null; }
            if (_customButtonOffTextureURL != "") QueueLoadTexture(_customButtonOffTextureURL, "OFF");
            else { _buttonOffTexture = null; }
        }

        public List<string> GetButtonSkinParamChoices(bool defaultSkinChoices = false) {
            List<string> choices = new List<string>();
            if (defaultSkinChoices) {
                choices.Add("Button(Small) Font Size");
                choices.Add("Button(Medium) Font Size");
                choices.Add("Button(Large) Font Size");
            }
            else choices.Add("Button Font Size");
            choices.Add("Button Font");
            choices.Add("Button Text Alignment");
            if (_maxButtonStates == 1) { choices.Add("Button Text Color"); choices.Add("Button Color"); choices.Add("Button Thumbnail"); }
            else { choices.Add("Button Text Color (On)"); choices.Add("Button Text Color (Off)"); choices.Add("Button Color (On)"); choices.Add("Button Color (Off)"); choices.Add("Button Thumbnail (On)"); choices.Add("Button Thumbnail (Off)"); }
            if (defaultSkinChoices) choices.Add("Gaze Select Button Color");
            _buttonOnColor.V = 0.84f;
            _buttonOffColor.V = 0.4f;
            if (defaultSkinChoices) choices.Add("Default Gaze Target");
            return (choices);
        }
        public void QueueLoadTexture(string url, string loadType)
        {
            if (ImageLoaderThreaded.singleton == null)
                return;
            if (string.IsNullOrEmpty(url)) {
                if (loadType == "ON") _buttonOnTexture = null;
                else _buttonOffTexture = null;
                _mvrScript.RefreshUIButtonGrid();
                return;
            }

            ImageLoaderThreaded.QueuedImage queuedImage = new ImageLoaderThreaded.QueuedImage();
            queuedImage.imgPath = url;
            queuedImage.forceReload = true;
            queuedImage.skipCache = true;
            queuedImage.compress = true;
            queuedImage.createMipMaps = false;
            queuedImage.isNormalMap = false;
            queuedImage.linear = false;
            queuedImage.createAlphaFromGrayscale = false;
            queuedImage.createNormalFromBump = false;
            queuedImage.bumpStrength = 0f;
            queuedImage.isThumbnail = false;
            queuedImage.fillBackground = false;
            queuedImage.invert = false;
            switch (loadType) {
                case "ON":
                    queuedImage.callback = (ImageLoaderThreaded.QueuedImage qi) =>
                    {
                        Texture2D tex = qi.tex;
                        _buttonOnTexture = tex;
                        _mvrScript.RefreshUIButtonGrid();
                    };
                    break;
                case "OFF":
                    queuedImage.callback = (ImageLoaderThreaded.QueuedImage qi) =>
                    {
                        Texture2D tex = qi.tex;
                        _buttonOffTexture = tex;
                        _mvrScript.RefreshUIButtonGrid();
                    };
                    break;
            }

            ImageLoaderThreaded.singleton.QueueImage(queuedImage);
        }
    }

    public class ButtonSkinDisplay {
        private UIAssist _mvrScript;
        public ButtonSkin _selectedButtonSkin;

        bool _isDefaultSkin;
        bool _isDisplayUpdating = false;

        public JSONStorableStringChooser _buttonSkinParamSelectorJSON;
        public JSONStorableFloat _skinFloatParam1JSON;
        public JSONStorableFloat _skinFloatParam2JSON;
        public JSONStorableColor _skinColorParamJSON;
        public JSONStorableStringChooser _skinStringChooserParamJSON;
        public JSONStorableString _skinStringParamJSON;
        public JSONStorableBool _skinSetToCurrentProfileDefaultJSON;

        public UIDynamicPopup _buttonSkinParamSelectorPopup;
        public UIDynamicPopup _skinStringChooserParamPopup;
        private UIDynamicSlider _skinFloatParam1Slider;
        private UIDynamicSlider _skinFloatParam2Slider;
        private UIDynamicColorPicker _skinColorParamColorPicker;
        private UIDynamicToggle _skinSetToCurrentProfileDefaultToggle;

        private UIDynamicTextField _skinStringParamTextField;
        private UIDynamicButton _skinMirrorOnButton;
        private UIDynamicButton _skinSelectTextureButton;
        private UIDynamicButton _skinClearTextureButton;

        private UIDynamic _spacer1;
        private UIDynamic _spacer2;

        public ButtonSkinDisplay(ButtonSkin buttonSkin, UIAssist mvrScript, bool defaultSkin = false) {
            try {
                _selectedButtonSkin = buttonSkin;
                _mvrScript = mvrScript;
                _isDefaultSkin = defaultSkin;
                string initalSkinParamValue = "Button Font Size";
                if (defaultSkin) initalSkinParamValue = "Button(Small) Font Size";
                _buttonSkinParamSelectorJSON = new JSONStorableStringChooser("buttonSkinParamSelect", _selectedButtonSkin.GetButtonSkinParamChoices(), initalSkinParamValue, "", SwitchSkinParam);
                HSVColor hsvColor; hsvColor.H = 0f; hsvColor.S = 0f; hsvColor.V = 0f;
                _skinColorParamJSON = new JSONStorableColor("skinColorParam", hsvColor, SwitchColorParam);
                _skinSetToCurrentProfileDefaultJSON = new JSONStorableBool("skinSetToDefault", true, SwitchSetToDefault);
                _skinFloatParam1JSON = new JSONStorableFloat("skinFloatParam1", 8f, SwitchFloat1, 2f, 40f, true);
                _skinFloatParam2JSON = new JSONStorableFloat("skinFloatParam2", 18f, SwitchFloat2, 2f, 40f, true);
                _skinStringChooserParamJSON = new JSONStorableStringChooser("skinStringChooser", null, "", "", SwitchStringChooser);
                _skinStringParamJSON = new JSONStorableString("skinString", "");

            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }

        }
        private void SwitchStringChooser(string stringVal) {
            try {


                if (!_isDisplayUpdating) {
                    if (_buttonSkinParamSelectorJSON.val == "Button Font") {
                        _selectedButtonSkin._buttonFont = stringVal;
                    }
                    else if (_buttonSkinParamSelectorJSON.val == "Button Text Alignment") {
                        _selectedButtonSkin._textAlignment = stringVal;
                    }
                    else if (_buttonSkinParamSelectorJSON.val == "Default Gaze Target")
                    {
                        _selectedButtonSkin._defaultGazeTarget = stringVal;
                    }

                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchFloat1(float float1val) {
            try {
                if (!_isDisplayUpdating) {
                    if (_buttonSkinParamSelectorJSON.val == "Button Font Size") _selectedButtonSkin._minFontSize = float1val;
                    else if (_buttonSkinParamSelectorJSON.val == "Button(Small) Font Size") _selectedButtonSkin._minSmallFontSize = float1val;
                    else if (_buttonSkinParamSelectorJSON.val == "Button(Medium) Font Size") _selectedButtonSkin._minMediumFontSize = float1val;
                    else if (_buttonSkinParamSelectorJSON.val == "Button(Large) Font Size") _selectedButtonSkin._minLargeFontSize = float1val;

                    _isDisplayUpdating = true;
                    _skinFloatParam1JSON.val = Mathf.Round(float1val);
                    _isDisplayUpdating = false;

                    if (_skinFloatParam1JSON.val > _skinFloatParam2JSON.val) _skinFloatParam2JSON.val = Mathf.Round(float1val);

                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchFloat2(float float2val) {
            try {
                if (!_isDisplayUpdating) {
                    if (_buttonSkinParamSelectorJSON.val == "Button Font Size") _selectedButtonSkin._maxFontSize = float2val;
                    else if (_buttonSkinParamSelectorJSON.val == "Button(Small) Font Size") _selectedButtonSkin._maxSmallFontSize = float2val;
                    else if (_buttonSkinParamSelectorJSON.val == "Button(Medium) Font Size") _selectedButtonSkin._maxMediumFontSize = float2val;
                    else if (_buttonSkinParamSelectorJSON.val == "Button(Large) Font Size") _selectedButtonSkin._maxLargeFontSize = float2val;

                    _isDisplayUpdating = true;
                    _skinFloatParam2JSON.val = Mathf.Round(float2val);
                    _isDisplayUpdating = false;

                    if (_skinFloatParam2JSON.val < _skinFloatParam1JSON.val) _skinFloatParam1JSON.val = Mathf.Round(float2val);

                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchSetToDefault(bool setToDefault) {
            try {
                if (!_isDisplayUpdating) {
                    switch (_buttonSkinParamSelectorJSON.val) {
                        case "Button Text Color":
                        case "Button Text Color (On)":
                            _selectedButtonSkin._textOnColorDefault = _skinSetToCurrentProfileDefaultJSON.val;
                            break;
                        case "Button Text Color (Off)":
                            _selectedButtonSkin._textOffColorDefault = _skinSetToCurrentProfileDefaultJSON.val;
                            break;
                        case "Button Color":
                        case "Button Color (On)":
                            _selectedButtonSkin._buttonOnColorDefault = _skinSetToCurrentProfileDefaultJSON.val;
                            break;
                        case "Button Color (Off)":
                            _selectedButtonSkin._buttonOffColorDefault = _skinSetToCurrentProfileDefaultJSON.val;
                            break;
                        case "Button Font Size":
                            _selectedButtonSkin._fontSizeDefault = _skinSetToCurrentProfileDefaultJSON.val;
                            break;
                        case "Button Font":
                            _selectedButtonSkin._buttonFontDefault = _skinSetToCurrentProfileDefaultJSON.val;
                            break;
                        case "Button Text Alignment":
                            _selectedButtonSkin._textAlignmentDefault = _skinSetToCurrentProfileDefaultJSON.val;
                            break;
                        case "Button Thumbnail (On)":
                        case "Button Thumbnail":
                            _selectedButtonSkin._customButtonOnTextureDefault = _skinSetToCurrentProfileDefaultJSON.val;
                            break;
                        case "Button Thumbnail (Off)":
                            _selectedButtonSkin._customButtonOffTextureDefault = _skinSetToCurrentProfileDefaultJSON.val;
                            break;
                    }
                    if (_isDefaultSkin) { _mvrScript.RefreshLeftUIDisplay(); }
                    else { _mvrScript.RefreshRightUIDisplay(); }
                    _mvrScript.RefreshUIButtonGrid();
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchColorParam(JSONStorableColor color) {
            try {
                if (!_isDisplayUpdating) {
                    switch (_buttonSkinParamSelectorJSON.val) {
                        case "Button Text Color":
                        case "Button Text Color (On)":
                            _selectedButtonSkin._textOnColor = color.val;
                            break;
                        case "Button Text Color (Off)":
                            _selectedButtonSkin._textOffColor = color.val;
                            break;
                        case "Button Color":
                        case "Button Color (On)":
                            _selectedButtonSkin._buttonOnColor = color.val;
                            break;
                        case "Button Color (Off)":
                            _selectedButtonSkin._buttonOffColor = color.val;
                            break;
                        case "Gaze Select Button Color":
                            _selectedButtonSkin._buttonGazeSelectColor = color.val;
                            break;
                    }
                    if (_isDefaultSkin) { _mvrScript.RefreshLeftUIDisplay(); }
                    else { _mvrScript.RefreshRightUIDisplay(); }
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchSkinParam(string skinParam) {
            try {
                if (!_isDisplayUpdating) {
                    if (_isDefaultSkin) { _mvrScript.RefreshLeftUIDisplay(); }
                    else { _mvrScript.RefreshRightUIDisplay(); }
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void TextureSelected(string filePath) {
            try {
                if (filePath != "") {
                    if (_buttonSkinParamSelectorJSON.val == "Button Thumbnail" || _buttonSkinParamSelectorJSON.val == "Button Thumbnail (On)") {
                        _selectedButtonSkin._customButtonOnTextureURL = filePath;
                        _selectedButtonSkin.QueueLoadTexture(filePath, "ON");
                    }
                    else if (_buttonSkinParamSelectorJSON.val == "Button Thumbnail (Off)") {
                        _selectedButtonSkin._customButtonOffTextureURL = filePath;
                        _selectedButtonSkin.QueueLoadTexture(filePath, "OFF");
                    }
                    if (_isDefaultSkin) { _mvrScript.RefreshLeftUIDisplay(); }
                    else { _mvrScript.RefreshRightUIDisplay(); }
                }
            } catch (Exception e) { SuperController.LogError("Exception caught: " + e); }

        }
        public void ClearDisplay() {
            if (_buttonSkinParamSelectorPopup != null) { _mvrScript.RemovePopup(_buttonSkinParamSelectorPopup); }
            if (_skinStringChooserParamPopup != null) { _mvrScript.RemovePopup(_skinStringChooserParamPopup); }
            if (_skinStringParamTextField != null) { _mvrScript.RemoveTextField(_skinStringParamTextField); }
            if (_skinFloatParam1Slider != null) { _mvrScript.RemoveSlider(_skinFloatParam1Slider); }
            if (_skinFloatParam2Slider != null) { _mvrScript.RemoveSlider(_skinFloatParam2Slider); }
            if (_skinColorParamColorPicker != null) { _mvrScript.RemoveColorPicker(_skinColorParamColorPicker); }
            if (_skinStringParamTextField != null) { _mvrScript.RemoveTextField(_skinStringParamTextField); }
            if (_skinSetToCurrentProfileDefaultToggle != null) { _mvrScript.RemoveToggle(_skinSetToCurrentProfileDefaultToggle); }
            if (_skinMirrorOnButton != null) { _mvrScript.RemoveButton(_skinMirrorOnButton); }
            if (_skinSelectTextureButton != null) { _mvrScript.RemoveButton(_skinSelectTextureButton); }
            if (_skinClearTextureButton != null) { _mvrScript.RemoveButton(_skinClearTextureButton); }

            if (_spacer1 != null) { _mvrScript.RemoveSpacer(_spacer1); }
            if (_spacer2 != null) { _mvrScript.RemoveSpacer(_spacer2); }
        }
        public void RefreshDisplay() {
            ClearDisplay();
            _isDisplayUpdating = true;

            if (_isDefaultSkin) {
                _buttonSkinParamSelectorPopup = _mvrScript.CreateSetupPopup(_buttonSkinParamSelectorJSON, !_isDefaultSkin, 500f, SkinParamSelectorPopupOpen, "Button Default Skin Param");
            }
            else {
                _spacer1 = _mvrScript.CreateSpacer(true);
                _spacer1.height = 10;
                _buttonSkinParamSelectorPopup = _mvrScript.CreateSetupPopup(_buttonSkinParamSelectorJSON, !_isDefaultSkin, 500f, SkinParamSelectorPopupOpen, "Button Skin Parameter");
                _skinSetToCurrentProfileDefaultToggle = _mvrScript.CreateToggle(_skinSetToCurrentProfileDefaultJSON, !_isDefaultSkin);
                _skinSetToCurrentProfileDefaultToggle.label = "Set Skin Param to default value";
                switch (_buttonSkinParamSelectorJSON.val) {
                    case "Button Text Color":
                    case "Button Text Color (On)":
                        _skinSetToCurrentProfileDefaultJSON.val = _selectedButtonSkin._textOnColorDefault;
                        break;
                    case "Button Text Color (Off)":
                        _skinSetToCurrentProfileDefaultJSON.val = _selectedButtonSkin._textOffColorDefault;
                        break;
                    case "Button Color":
                    case "Button Color (On)":
                        _skinSetToCurrentProfileDefaultJSON.val = _selectedButtonSkin._buttonOnColorDefault;
                        break;
                    case "Button Color (Off)":
                        _skinSetToCurrentProfileDefaultJSON.val = _selectedButtonSkin._buttonOffColorDefault;
                        break;

                    case "Button Font Size":
                        _skinSetToCurrentProfileDefaultJSON.val = _selectedButtonSkin._fontSizeDefault;
                        break;
                    case "Button Font":
                        _skinSetToCurrentProfileDefaultJSON.val = _selectedButtonSkin._buttonFontDefault;
                        break;
                    case "Button Text Alignment":
                        _skinSetToCurrentProfileDefaultJSON.val = _selectedButtonSkin._textAlignmentDefault;
                        break;
                    case "Button Thumbnail (On)":
                    case "Button Thumbnail":
                        _skinSetToCurrentProfileDefaultJSON.val = _selectedButtonSkin._customButtonOnTextureDefault;
                        break;
                    case "Button Thumbnail (Off)":
                        _skinSetToCurrentProfileDefaultJSON.val = _selectedButtonSkin._customButtonOffTextureDefault;
                        break;
                }
            }
            _buttonSkinParamSelectorPopup.labelTextColor = Color.blue;
            if (_isDefaultSkin || _skinSetToCurrentProfileDefaultJSON.val == false) {
                switch (_buttonSkinParamSelectorJSON.val) {
                    case "Button Text Color":
                    case "Button Text Color (On)":
                        _skinColorParamColorPicker = _mvrScript.CreateColorPicker(_skinColorParamJSON, !_isDefaultSkin);
                        _skinColorParamJSON.val = _selectedButtonSkin._textOnColor;
                        break;
                    case "Button Text Color (Off)":
                        _skinColorParamColorPicker = _mvrScript.CreateColorPicker(_skinColorParamJSON, !_isDefaultSkin);
                        _skinColorParamJSON.val = _selectedButtonSkin._textOffColor;
                        if (_selectedButtonSkin._textOffColor.H != _selectedButtonSkin._textOnColor.H || _selectedButtonSkin._textOffColor.S != _selectedButtonSkin._textOnColor.S || _selectedButtonSkin._textOffColor.V != _selectedButtonSkin._textOnColor.V) {
                            _skinMirrorOnButton = _mvrScript.CreateButton("Copy from ON Color", !_isDefaultSkin);
                            _skinMirrorOnButton.button.onClick.AddListener(() => {
                                if (!_isDisplayUpdating) {
                                    _selectedButtonSkin._textOffColor = _selectedButtonSkin._textOnColor;
                                    if (_isDefaultSkin) { _mvrScript.RefreshLeftUIDisplay(); }
                                    else { _mvrScript.RefreshRightUIDisplay(); }
                                }
                            });
                        }
                        break;
                    case "Button Color":
                    case "Button Color (On)":
                        _skinColorParamColorPicker = _mvrScript.CreateColorPicker(_skinColorParamJSON, !_isDefaultSkin);
                        _skinColorParamJSON.val = _selectedButtonSkin._buttonOnColor;
                        break;
                    case "Button Color (Off)":
                        _skinColorParamColorPicker = _mvrScript.CreateColorPicker(_skinColorParamJSON, !_isDefaultSkin);
                        _skinColorParamJSON.val = _selectedButtonSkin._buttonOffColor;
                        if (_selectedButtonSkin._buttonOffColor.H != _selectedButtonSkin._buttonOnColor.H || _selectedButtonSkin._buttonOffColor.S != _selectedButtonSkin._buttonOnColor.S || _selectedButtonSkin._buttonOffColor.V != _selectedButtonSkin._buttonOnColor.V) {
                            _skinMirrorOnButton = _mvrScript.CreateButton("Copy from ON Color", !_isDefaultSkin);
                            _skinMirrorOnButton.button.onClick.AddListener(() => {
                                if (!_isDisplayUpdating) {
                                    _selectedButtonSkin._buttonOffColor = _selectedButtonSkin._buttonOnColor;
                                    if (_isDefaultSkin) { _mvrScript.RefreshLeftUIDisplay(); }
                                    else { _mvrScript.RefreshRightUIDisplay(); }
                                }
                            });
                        }
                        break;
                    case "Gaze Select Button Color":
                        _skinColorParamColorPicker = _mvrScript.CreateColorPicker(_skinColorParamJSON, !_isDefaultSkin);
                        HSVColor buttonGazeSelectDefaultColor = new HSVColor();

                        Color gazeButtonColor = new Color(128f / 255f, 142f / 255f, 216f / 255f, 1f);
                        Color.RGBToHSV(gazeButtonColor, out buttonGazeSelectDefaultColor.H, out buttonGazeSelectDefaultColor.S, out buttonGazeSelectDefaultColor.V);
                        _skinColorParamJSON.val = _selectedButtonSkin._buttonGazeSelectColor;
                        _skinColorParamJSON.defaultVal = buttonGazeSelectDefaultColor;
                        break;
                    case "Button Font Size":
                    case "Button(Small) Font Size":
                    case "Button(Medium) Font Size":
                    case "Button(Large) Font Size":
                        _skinFloatParam1Slider = _mvrScript.CreateSlider(_skinFloatParam1JSON, !_isDefaultSkin);
                        _skinFloatParam2Slider = _mvrScript.CreateSlider(_skinFloatParam2JSON, !_isDefaultSkin);
                        _skinFloatParam1Slider.label = "Min Font Size";
                        _skinFloatParam2Slider.label = "Max Font Size";
                        if (_buttonSkinParamSelectorJSON.val == "Button Font Size") {
                            _skinFloatParam1JSON.val = Mathf.Round(_selectedButtonSkin._minFontSize);
                            _skinFloatParam2JSON.val = Mathf.Round(_selectedButtonSkin._maxFontSize);
                        }
                        else if (_buttonSkinParamSelectorJSON.val == "Button(Small) Font Size") {
                            _skinFloatParam1JSON.val = Mathf.Round(_selectedButtonSkin._minSmallFontSize);
                            _skinFloatParam2JSON.val = Mathf.Round(_selectedButtonSkin._maxSmallFontSize);
                        }
                        else if (_buttonSkinParamSelectorJSON.val == "Button(Medium) Font Size") {
                            _skinFloatParam1JSON.val = Mathf.Round(_selectedButtonSkin._minMediumFontSize);
                            _skinFloatParam2JSON.val = Mathf.Round(_selectedButtonSkin._maxMediumFontSize);
                        }
                        else if (_buttonSkinParamSelectorJSON.val == "Button(Large) Font Size") {
                            _skinFloatParam1JSON.val = Mathf.Round(_selectedButtonSkin._minLargeFontSize);
                            _skinFloatParam2JSON.val = Mathf.Round(_selectedButtonSkin._maxLargeFontSize);
                        }
                        break;
                    case "Default Gaze Target":
                        _skinStringChooserParamPopup = _mvrScript.CreateSetupPopup(_skinStringChooserParamJSON, !_isDefaultSkin, 250f, SkinStringChooserPopupOpen, "Gaze Target",false);
                        _skinStringChooserParamJSON.val = _selectedButtonSkin._defaultGazeTarget;
                        break;
                    case "Button Font":
                        _skinStringChooserParamPopup = _mvrScript.CreateSetupPopup(_skinStringChooserParamJSON, !_isDefaultSkin, 500f, SkinStringChooserPopupOpen, "Select Font");
                        _skinStringChooserParamJSON.val = _selectedButtonSkin._buttonFont;
                        break;
                    case "Button Text Alignment":
                        _skinStringChooserParamPopup = _mvrScript.CreateSetupPopup(_skinStringChooserParamJSON, !_isDefaultSkin, 500f, SkinStringChooserPopupOpen, "Select Alignment");
                        _skinStringChooserParamJSON.val = _selectedButtonSkin._textAlignment;
                        break;
                    case "Button Thumbnail (On)":
                    case "Button Thumbnail":
                        _skinStringParamTextField = _mvrScript.CreateTextField(_skinStringParamJSON, !_isDefaultSkin);
                        _skinStringParamTextField.height = 1;
                        _skinStringParamJSON.val = _selectedButtonSkin._customButtonOnTextureURL;
                        _skinSelectTextureButton = _mvrScript.CreateButton("Select Thumbnail Image", !_isDefaultSkin);
                        _skinSelectTextureButton.button.onClick.AddListener(() => { SuperController.singleton.GetMediaPathDialog(TextureSelected, "jpg|png", "", false, true, false); });
                        if (_skinStringParamJSON.val != null && _skinStringParamJSON.val != "") {
                            _skinClearTextureButton = _mvrScript.CreateButton("Clear Thumbnail", !_isDefaultSkin);

                            _skinClearTextureButton.button.onClick.AddListener(() => {
                                if (!_isDisplayUpdating) {
                                    _selectedButtonSkin._customButtonOnTextureURL = "";
                                    _selectedButtonSkin.QueueLoadTexture("", "ON");
                                    if (_isDefaultSkin) { _mvrScript.RefreshLeftUIDisplay(); }
                                    else { _mvrScript.RefreshRightUIDisplay(); }
                                }
                            }
                        ); }
                        break;
                    case "Button Thumbnail (Off)":
                        _skinStringParamTextField = _mvrScript.CreateTextField(_skinStringParamJSON, !_isDefaultSkin);
                        _skinStringParamTextField.height = 1;
                        _skinStringParamJSON.val = _selectedButtonSkin._customButtonOffTextureURL;
                        _skinSelectTextureButton = _mvrScript.CreateButton("Select Thumbnail Image", !_isDefaultSkin);
                        _skinSelectTextureButton.button.onClick.AddListener(() => { SuperController.singleton.GetMediaPathDialog(TextureSelected, "jpg|png", "", false, true, false); });
                        if (_skinStringParamJSON.val != "" && _skinStringParamJSON.val != null) {
                            _skinClearTextureButton = _mvrScript.CreateButton("Clear Thumbnail", !_isDefaultSkin);

                            _skinClearTextureButton.button.onClick.AddListener(() => {
                                if (!_isDisplayUpdating) {
                                    _selectedButtonSkin._customButtonOffTextureURL = "";
                                    _selectedButtonSkin.QueueLoadTexture("", "OFF");
                                    if (_isDefaultSkin) { _mvrScript.RefreshLeftUIDisplay(); }
                                    else { _mvrScript.RefreshRightUIDisplay(); }
                                }
                            }
                        ); }
                        if (_selectedButtonSkin._customButtonOffTextureURL != _selectedButtonSkin._customButtonOnTextureURL) {
                            _skinMirrorOnButton = _mvrScript.CreateButton("Copy from ON Thumbnail", !_isDefaultSkin);
                            _skinMirrorOnButton.button.onClick.AddListener(() => {
                                if (!_isDisplayUpdating)
                                { _selectedButtonSkin._customButtonOffTextureURL = _selectedButtonSkin._customButtonOnTextureURL;
                                    _selectedButtonSkin.QueueLoadTexture(_selectedButtonSkin._customButtonOffTextureURL, "OFF");
                                    if (_isDefaultSkin) { _mvrScript.RefreshLeftUIDisplay(); }
                                    else { _mvrScript.RefreshRightUIDisplay(); }
                                }
                            });
                        }
                        break;

                }
                if (_skinColorParamColorPicker != null) _skinColorParamColorPicker.label = "Color";
            }
            _isDisplayUpdating = false;
        }

        private void SkinParamSelectorPopupOpen() {
            _mvrScript.CloseAllRightUIPopups(_buttonSkinParamSelectorPopup);
            _mvrScript.CloseAllLeftUIPopups(_buttonSkinParamSelectorPopup);

            _buttonSkinParamSelectorJSON.choices = _selectedButtonSkin.GetButtonSkinParamChoices(_isDefaultSkin);
        }
        private void SkinStringChooserPopupOpen() {
            _mvrScript.CloseAllRightUIPopups(_skinStringChooserParamPopup);
            _mvrScript.CloseAllLeftUIPopups(_buttonSkinParamSelectorPopup);
            List<string> popupChoices = new List<string>();
            if (_buttonSkinParamSelectorJSON.val == "Button Font") {
                popupChoices.Add("Arial");
                popupChoices.Add("chintzy");
                popupChoices.Add("DroidSansMono");
                popupChoices.Add("Laffayette_Comic_Pro");
            }
            else if (_buttonSkinParamSelectorJSON.val == "Button Text Alignment") {
                popupChoices.Add("MiddleCenter");
                popupChoices.Add("LowerCenter");
                popupChoices.Add("UpperCenter");
                popupChoices.Add("MiddleLeft");
                popupChoices.Add("LowerLeft");
                popupChoices.Add("UpperLeft");
                popupChoices.Add("MiddleRight");
                popupChoices.Add("LowerRight");
                popupChoices.Add("UpperRight");
            }
            else if (_buttonSkinParamSelectorJSON.val == "Default Gaze Target")
            {
                popupChoices.Add("Last viewed Person");
                popupChoices.Add("Last viewed Female");
                popupChoices.Add("Last viewed Male");
            }
            _skinStringChooserParamJSON.choices = popupChoices;
        }

    }

    public class ToolBox
    {

        public ToolBox(UIAssist mvrScript)
        {
            _mvrScript = mvrScript;
            _heelAdjustTool = new HeelAdjustTool(mvrScript);
            _pluginLoaderTool = new PluginLoaderTool(mvrScript);
        }

        public void RefreshToolboxDisplay(string toolboxSelection)
        {
            if (_toolBoxToggle1 != null) _mvrScript.RemoveToggle(_toolBoxToggle1);
            if (_toolBoxToggle2 != null) _mvrScript.RemoveToggle(_toolBoxToggle2);
            if (_toolBoxPopup1 != null) _mvrScript.RemovePopup(_toolBoxPopup1);
            if (_toolBoxPopup2 != null) _mvrScript.RemovePopup(_toolBoxPopup2);

            switch (toolboxSelection)
            {
                case "Auto Heel Adjust":
                    _toolBoxToggle1 = _mvrScript.CreateToggle(_heelAdjustTool._heelAdjustActiveJSON);
                    _toolBoxToggle1.label = "Auto Load Prestigitis Heel Adjusts";
                    if (_heelAdjustTool._heelAdjustActiveJSON.val)
                    {
                        _toolBoxToggle2 = _mvrScript.CreateToggle(_heelAdjustTool._heelAdjustRaisePeopleJSON);
                        _toolBoxToggle2.label = "Auto Raise Person by Heel Height";
                    }
                    break;
                case "Auto Plugin Loader":
                    _toolBoxPopup1 = _mvrScript.CreateSetupPopup(_pluginLoaderTool._atomTypeJSON);

                    break;

            }
        }

        public JSONClass GetJSON()
        {
            JSONClass node = new JSONClass();
            node["heelAdjustActive"].AsBool = _heelAdjustTool._heelAdjustActiveJSON.val;
            node["heelAdjustRaisePerson"].AsBool = _heelAdjustTool._heelAdjustRaisePeopleJSON.val;
            return (node);
        }

        public void LoadJSON(JSONClass toolboxJSON)
        {
            if (toolboxJSON["heelAdjustActive"] != null) _heelAdjustTool._heelAdjustActiveJSON.val = toolboxJSON["heelAdjustActive"].AsBool;
            if (toolboxJSON["heelAdjustRaisePerson"] != null) _heelAdjustTool._heelAdjustRaisePeopleJSON.val = toolboxJSON["heelAdjustRaisePerson"].AsBool;
        }

        private UIAssist _mvrScript;
        private UIDynamicToggle _toolBoxToggle1;
        private UIDynamicToggle _toolBoxToggle2;
        private UIDynamicPopup _toolBoxPopup1;
        private UIDynamicPopup _toolBoxPopup2;

        public HeelAdjustTool _heelAdjustTool;
        public PluginLoaderTool _pluginLoaderTool;
    }
    public class PluginLoaderTool
    {
        public PluginLoaderTool(UIAssist mvrScript)
        {
            _mvrScript = mvrScript;
            List<string> atomTypeChoices = new List<string>();
            atomTypeChoices.Add("People");
            atomTypeChoices.Add("Females");
            atomTypeChoices.Add("Males");
            atomTypeChoices.Add("Custom Unity Assets");


            _atomTypeJSON = new JSONStorableStringChooser("atomType", atomTypeChoices, "Females","Atom types",SwitchAtomType);


        }

        public void SwitchAtomType(string atomType)
        {
 //           if (!_mvrScript)
            {
 //               RefreshLeftUIDisplay();
            }
        }

        UIAssist _mvrScript;
        public JSONStorableStringChooser _atomTypeJSON;



    }

    public class HeelAdjustTool
    {
        public HeelAdjustTool(UIAssist mvrScript)
        {
            _mvrScript = mvrScript;
            _heelAdjustActiveJSON = new JSONStorableBool("heelAdjustActive", true, SwitchHeelAdjustActive);
            _mvrScript.RegisterBool(_heelAdjustActiveJSON);
            _heelAdjustActiveJSON.isRestorable = false;
            _heelAdjustActiveJSON.isStorable = false;
            _heelAdjustRaisePeopleJSON = new JSONStorableBool("heelAdjustRaisePerson", false, SwitchHeelAdjustRaisePerson);
            _mvrScript.RegisterBool(_heelAdjustRaisePeopleJSON);
            _heelAdjustRaisePeopleJSON.isRestorable = false;
            _heelAdjustRaisePeopleJSON.isStorable = false;

        }
        private void SwitchHeelAdjustActive(bool haActive)
        {
            try
            {
                if (haActive)
                {
                    if (_heelAdjustRaisePeopleJSON.val)
                    {
                        // If Raising people is now being switched to active, then disable the temporary suppression.
                        foreach (KeyValuePair<string, HASetting> kvp in _atomHASettings)
                        {
                            kvp.Value._suppressControlNodeHeightReset = false;
                        }
                    }
                    if (!isHAFolderPresent())
                    {
                        if (!_mvrScript._isRestoring) SuperController.LogMessage("UIAssist: Auto heel adjust requires HeelAdjust configs to be created by prestigitis plugin and placed in 'Saves\\HeelAdjust' folder.");
                    }                
                }
                else
                {
                    foreach (Atom atom in SuperController.singleton.GetAtoms())
                    {
                        if (atom.type == "Person")FootPoseReset(atom);
                    }
                }

                JSONStorable otherUIA = _mvrScript.GetOtherActiveUIAPlugin();
                if (otherUIA != null)
                {
                    JSONStorableBool otherUIAHeelAdjustActiveJSON = otherUIA.GetBoolJSONParam("heelAdjustActive");
                    if (otherUIAHeelAdjustActiveJSON != null) otherUIAHeelAdjustActiveJSON.val = haActive;
                    
                }
                _mvrScript.RefreshLeftUIDisplay();
            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }
        private void SwitchHeelAdjustRaisePerson(bool haRaisePerson)
        {
            try
            {
                if (haRaisePerson)
                {
                    foreach (KeyValuePair<string,HASetting> kvp in _atomHASettings)
                    {
                        kvp.Value._suppressControlNodeHeightReset = false;
                    }
                }
                foreach (Atom atom in SuperController.singleton.GetAtoms())
                {
                    if (atom.type == "Person")
                    {
                        if (haRaisePerson) UpdateAtomHeightOffset(atom);
                        else ResetAtomHeightOffset(atom);
                    }
                }

            }
            catch (Exception e) { SuperController.LogError("Exception caught: " + e); }
        }

        public void UpdateExistingHAFiles()
        {
            _existingHAFiles = new List<string>();
            if (FileManagerSecure.DirectoryExists(SuperController.singleton.savesDir + Constants._HeelAdjustSubfolderName))
            {
                foreach (string file in SuperController.singleton.GetFilesAtPath(SuperController.singleton.savesDir + Constants._HeelAdjustSubfolderName))
                {
                    string[] pathsplit = file.Split(new string[] { "\\" }, StringSplitOptions.None);
                    _existingHAFiles.Add(pathsplit.Last<String>().Replace("." + Constants._HeelAdjustFileExtension, ""));
                }
            }
        }
        private Transform CreateSingleShoeCollider(Atom atom, string tranformHierarchy, string colliderNamePreFix, Vector3 baseRotation, float colliderSize)
        {
            Transform limbTransform = atom.gameObject.transform.Find(tranformHierarchy);

            GameObject rescaleGO = new GameObject();
            rescaleGO.transform.position = limbTransform.position;
            rescaleGO.transform.rotation = limbTransform.rotation;
            rescaleGO.transform.parent = limbTransform;
            rescaleGO.transform.localEulerAngles = baseRotation;
            rescaleGO.name = colliderNamePreFix + "ColliderRescaleUIA";
            rescaleGO.transform.localScale = new Vector3(1f, 0.01f, 1f);

            GameObject colliderGO = GameObject.CreatePrimitive(PrimitiveType.Cube);
            colliderGO.transform.position = rescaleGO.transform.position;
            colliderGO.transform.parent = rescaleGO.transform;
            colliderGO.transform.localScale = new Vector3(colliderSize, Constants._ShoeColliderBaseScale, colliderSize);
            colliderGO.name = colliderNamePreFix + "ColliderUIA";
            colliderGO.GetComponent<Collider>().enabled = true;
            colliderGO.transform.localPosition = new Vector3(0f, -Constants._ShoeColliderBaseScale / 2, 0f);
            colliderGO.transform.localEulerAngles = new Vector3(0f, 0.0f, 0.0f);
            colliderGO.GetComponent<Renderer>().enabled = false;
            colliderGO.GetComponent<Renderer>().material.color = Color.green;

            return (rescaleGO.transform);
        }
        private void CreateHAColliders(Atom atom)
        {

           // if there are any colliders active from the HeelAdjust plugin - then remove them (needed because HA plugin doesnt tidy up OnDestroy)
            RemoveCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot", "rShoeHeel", false);
            RemoveCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot", "lShoeHeel", false);
            RemoveCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rToe", "rShoeToe", false);
            RemoveCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot/lToe", "lShoeToe", false);

            Transform rShoeToeRescaleTran = CreateSingleShoeCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rToe", "rShoeToe", _ToeColliderBaseRotationR, 0.05f);
            Transform lShoeToeRescaleTran = CreateSingleShoeCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot/lToe", "lShoeToe", _ToeColliderBaseRotationL, 0.05f);
            Transform rShoeHeelRescaleTran = CreateSingleShoeCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot", "rShoeHeel", _HeelColliderBaseRotationR, 0.02f);
            Transform lShoeHeelRescaleTran = CreateSingleShoeCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot", "lShoeHeel", _HeelColliderBaseRotationL, 0.02f);

        }

        private void ApplySingleColiderHeightRotation(Transform transform, float colliderHeight, Vector3 colliderRotation, Vector3 baseColliderRotation, float colliderUpdateDelay)
        {
            _mvrScript.StartCoroutine(DelayedColliderUpdate(transform,colliderHeight, colliderUpdateDelay));
            transform.transform.localEulerAngles = new Vector3(-colliderRotation.x + baseColliderRotation.x, baseColliderRotation.y, baseColliderRotation.z);
        }
        private IEnumerator DelayedColliderUpdate(Transform transform, float colliderHeight, float colliderUpdateDelay)
        {
            if (colliderUpdateDelay == 0.001f) yield return new WaitForEndOfFrame();
            else yield return new WaitForSeconds(colliderUpdateDelay);
            transform.transform.localScale = new Vector3(1f, colliderHeight, 1f);
        }
        private void ApplyHAColliderHeightsRotations(Atom atom, HASetting haSetting,bool vamBuiltInHeel, float colliderUpdateDelay)
        {
            float toeColliderHeight = 0.0001f;
            float heelColliderHeight = 0.0001f;
            if (!vamBuiltInHeel)
            {
                toeColliderHeight = haSetting._toeColliderHeight;
                heelColliderHeight = haSetting._heelColliderHeight;
            }

            Transform rToeColliderRescale = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rToe/rShoeToeColliderRescaleUIA");
            ApplySingleColiderHeightRotation(rToeColliderRescale, toeColliderHeight, haSetting._toeRotation, _ToeColliderBaseRotationR, colliderUpdateDelay);

            Transform rHeelColliderRescale = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rShoeHeelColliderRescaleUIA");
            ApplySingleColiderHeightRotation(rHeelColliderRescale, heelColliderHeight, haSetting._footRotation, _HeelColliderBaseRotationR, colliderUpdateDelay);

            Transform lToeColliderRescale = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot/lToe/lShoeToeColliderRescaleUIA");
            ApplySingleColiderHeightRotation(lToeColliderRescale, toeColliderHeight, haSetting._toeRotation, _ToeColliderBaseRotationL, colliderUpdateDelay);

            Transform lHeelColliderRescale = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot/lShoeHeelColliderRescaleUIA");
            ApplySingleColiderHeightRotation(lHeelColliderRescale, heelColliderHeight, haSetting._footRotation, _HeelColliderBaseRotationL, colliderUpdateDelay);

            Transform rHeelCollider = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rShoeHeelColliderRescaleUIA/rShoeHeelColliderUIA");
            Transform lHeelCollider = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot/lShoeHeelColliderRescaleUIA/lShoeHeelColliderUIA");

            rHeelCollider.transform.localPosition = new Vector3(0f, -Constants._ShoeColliderBaseScale / 2, haSetting._heelColliderBackOffset);
            lHeelCollider.transform.localPosition = new Vector3(0f, -Constants._ShoeColliderBaseScale / 2, haSetting._heelColliderBackOffset);

            //adjust foot pose to the given angles
            Transform rfoot = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot");
            Transform rtoe = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rToe");
            Transform lfoot = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot");
            Transform ltoe = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot/lToe");

            Vector3 rf = new Vector3(haSetting._footRotation.x, haSetting._footRotation.y, haSetting._footRotation.z);
            Vector3 rt = new Vector3(haSetting._toeRotation.x - haSetting._footRotation.x, haSetting._toeRotation.y, haSetting._toeRotation.z);
            Vector3 lf = new Vector3(haSetting._footRotation.x, -haSetting._footRotation.y, -haSetting._footRotation.z);
            Vector3 lt = new Vector3(haSetting._toeRotation.x - haSetting._footRotation.x, -haSetting._toeRotation.y, -haSetting._toeRotation.z);

            if (rfoot.GetComponent<DAZBone>().baseJointRotation != rf || rtoe.GetComponent<DAZBone>().baseJointRotation != rt || lfoot.GetComponent<DAZBone>().baseJointRotation != lf || ltoe.GetComponent<DAZBone>().baseJointRotation != lt)
            {
                if (_savedRotationStatesDict.ContainsKey(atom.name)) _savedRotationStatesDict[atom.name]._timeSaved = Time.unscaledTime;
                else
                {
                    ToeFootRotationStates savedRotationStates = GetCurrentRotationStates(atom);
                    _savedRotationStatesDict.Add(atom.name, savedRotationStates);
                }

                ResetRotationStates(atom, null);
                rfoot.GetComponent<DAZBone>().baseJointRotation = rf;
                rtoe.GetComponent<DAZBone>().baseJointRotation = rt;
                lfoot.GetComponent<DAZBone>().baseJointRotation = lf;
                ltoe.GetComponent<DAZBone>().baseJointRotation = lt;
            }
        }
        private ToeFootRotationStates GetCurrentRotationStates(Atom atom)
        {
            ToeFootRotationStates rotationStates = new ToeFootRotationStates();
            var rf = atom.gameObject.transform.Find("reParentObject/object/rescaleObject/JointControl/rToeControl").GetComponent<FreeControllerV3>();
            var rt = atom.gameObject.transform.Find("reParentObject/object/rescaleObject/JointControl/rFootControl").GetComponent<FreeControllerV3>();
            var lf = atom.gameObject.transform.Find("reParentObject/object/rescaleObject/JointControl/lToeControl").GetComponent<FreeControllerV3>();
            var lt = atom.gameObject.transform.Find("reParentObject/object/rescaleObject/JointControl/lFootControl").GetComponent<FreeControllerV3>();

            rotationStates._rFootRotationState = rf.currentRotationState;
            rotationStates._rToeRotationState = rt.currentRotationState;
            rotationStates._lFootRotationState = lf.currentRotationState;
            rotationStates._lToeRotationState = lt.currentRotationState;
            rotationStates._timeSaved = Time.unscaledTime;

            return (rotationStates);
        }

        public void ResetRotationStates(Atom atom, ToeFootRotationStates rotationStates)
        {
            if (atom != null)
            {
                var rf = atom.gameObject.transform.Find("reParentObject/object/rescaleObject/JointControl/rToeControl").GetComponent<FreeControllerV3>();
                var rt = atom.gameObject.transform.Find("reParentObject/object/rescaleObject/JointControl/rFootControl").GetComponent<FreeControllerV3>();
                var lf = atom.gameObject.transform.Find("reParentObject/object/rescaleObject/JointControl/lToeControl").GetComponent<FreeControllerV3>();
                var lt = atom.gameObject.transform.Find("reParentObject/object/rescaleObject/JointControl/lFootControl").GetComponent<FreeControllerV3>();

                if (rotationStates == null)
                {
                    rf.currentRotationState = FreeControllerV3.RotationState.Off;
                    rt.currentRotationState = FreeControllerV3.RotationState.Off;
                    lf.currentRotationState = FreeControllerV3.RotationState.Off;
                    lt.currentRotationState = FreeControllerV3.RotationState.Off;
                }
                else
                {
                    rf.currentRotationState = rotationStates._rFootRotationState;
                    rt.currentRotationState = rotationStates._rToeRotationState;
                    lf.currentRotationState = rotationStates._lFootRotationState;
                    lt.currentRotationState = rotationStates._lToeRotationState;
                }
            }
        }
        public void ResetAllSavedAtomRotationStates(bool setOff = false)
        {
            List<string> atomsToRemoveFromDictionary = new List<string>();
            foreach (KeyValuePair<string, ToeFootRotationStates> kvp in _savedRotationStatesDict)
            {
                if (Time.unscaledTime - kvp.Value._timeSaved > Constants._FCV3ResetDelay)
                {
                    Atom atom = SuperController.singleton.GetAtomByUid(kvp.Key);
                    ResetRotationStates(atom, kvp.Value);
                    atomsToRemoveFromDictionary.Add(kvp.Key);
                }

            }
            foreach (string atomName in atomsToRemoveFromDictionary) _savedRotationStatesDict.Remove(atomName);
        }
        private void AjustLinkedAPHeights(string personUID, float heightAdjust)
        {
            if (heightAdjust > 0.001f || heightAdjust < -0.001f)
            {
                foreach (Atom atom in SuperController.singleton.GetAtoms())
                {
                    if (atom.type == "AnimationPattern")
                    {
                        JSONStorable storable = atom.GetStorableByID("AnimatedObject");
                        if (storable != null)
                        {
                            JSONClass apAnimatiedObjectJSON = storable.GetJSON();
                            string reciever = apAnimatiedObjectJSON["receiver"];
                            if (reciever.Contains(":") && reciever.Length > reciever.IndexOf(':'))
                            {
                                string atomName = reciever.Substring(0, reciever.IndexOf(':'));
                                string nodeName = reciever.Substring(reciever.IndexOf(':') + 1);
                                if (atomName == personUID && (nodeName == "control" || nodeName == "lToeControl" || nodeName == "rToeControl" || nodeName == "lFootControl" || nodeName == "rFootControl"))
                                {
                                    FreeControllerV3 fcAP = atom.freeControllers.First(fc => fc.name == "control");
                                    fcAP.transform.position += Vector3.up * heightAdjust;
                                }
                            }
                        }
                    }
                }
            }
        }
        private void ResetAtomHeightOffset(Atom atom)
        {
            if (_atomHASettings.ContainsKey(atom.name) && !_atomHASettings[atom.name]._suppressControlNodeHeightReset)
            {
               FreeControllerV3 atomControlFCV3 = atom.freeControllers.First(fc => fc.name == "control");
                atomControlFCV3.transform.position -= Vector3.up * _atomHASettings[atom.name]._controlNodeHeightOffset * Constants._ShoeColliderBaseScale;
                AjustLinkedAPHeights(atom.name, -(_atomHASettings[atom.name]._controlNodeHeightOffset * Constants._ShoeColliderBaseScale));
               _atomHASettings[atom.name]._controlNodeHeightOffset = 0f;
            }
        }

        private void UpdateAtomHeightOffset(Atom atom)
        {
            if (_atomHASettings.ContainsKey(atom.name))
            {
                HASetting haSetting = _atomHASettings[atom.name];
                float heightAdjustment = 0f;
                heightAdjustment = haSetting._heelColliderHeight - (0.064f / Constants._ShoeColliderBaseScale) - haSetting._controlNodeHeightOffset;
                haSetting._controlNodeHeightOffset = haSetting._heelColliderHeight - (0.064f / Constants._ShoeColliderBaseScale);

                FreeControllerV3 atomControlFCV3 = atom.freeControllers.First(fc => fc.name == "control");

                atomControlFCV3.transform.position += Vector3.up * heightAdjustment * Constants._ShoeColliderBaseScale;
                AjustLinkedAPHeights(atom.name, heightAdjustment * Constants._ShoeColliderBaseScale);
            }
        }
        private void LoadHASetting (HASetting haSetting, string path)
        {
            string loadedSettings = SuperController.singleton.ReadFileIntoString(path);
            string[] loadedSettingsLines = loadedSettings.Split(new string[] { "\r\n", "\n" }, StringSplitOptions.None);
            string[] footVals = loadedSettingsLines[0].Split(new string[] { "," }, StringSplitOptions.None);
            string[] toeVals = loadedSettingsLines[1].Split(new string[] { "," }, StringSplitOptions.None);
            string[] heelColliderVals = loadedSettingsLines[2].Split(new string[] { "," }, StringSplitOptions.None);
            string[] toeColliderVals = loadedSettingsLines[3].Split(new string[] { "," }, StringSplitOptions.None);
            haSetting._footRotation.x = float.Parse(footVals[0]);
            haSetting._footRotation.y = float.Parse(footVals[1]);
            haSetting._footRotation.z = float.Parse(footVals[2]);
            haSetting._toeRotation.x = float.Parse(toeVals[0]);
            haSetting._toeRotation.y = float.Parse(toeVals[1]);
            haSetting._toeRotation.z = float.Parse(toeVals[2]);
            haSetting._heelColliderHeight = float.Parse(heelColliderVals[0]);
            haSetting._heelColliderBackOffset = float.Parse(heelColliderVals[1]);
            haSetting._toeColliderHeight = float.Parse(toeColliderVals[0]);
        }
        private void ApplyHA(Atom atom, string path)
        {
            HASetting haSetting;
            if (_atomHASettings.ContainsKey(atom.name)) haSetting = _atomHASettings[atom.name];
            else
            {
                haSetting = new HASetting();
                _atomHASettings.Add(atom.name, haSetting);
            }
            LoadHASetting(haSetting, path);

            float colliderUpdateDelay = 0.001f;
            if (_heelAdjustRaisePeopleJSON.val)
            {
                UpdateAtomHeightOffset(atom);
                colliderUpdateDelay = 0.3f;
            }
            

            Transform rt = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rShoeHeelColliderRescaleUIA");
            if (rt == null)
            {
                CreateHAColliders(atom);
                rt = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rShoeHeelColliderRescaleUIA");
            }

            if (rt.transform.localScale.y > haSetting._heelColliderHeight)colliderUpdateDelay = 0.001f;
            
            if (path.EndsWith("MeshedVR-HarliHeels.heeladjust")|| path.EndsWith("MeshedVR-CasualDenimShoes.heeladjust")) ApplyHAColliderHeightsRotations(atom, haSetting,true, colliderUpdateDelay);
            else ApplyHAColliderHeightsRotations(atom, haSetting,false, colliderUpdateDelay);
        }
        private void RemoveCollider(Atom atom, string limbHierarchy, string prefix, bool UIA = true)
        {
            string colliderRescaleName = prefix + "ColliderRescale";
            string colliderName = prefix + "Collider";
            if (UIA)
            {
                colliderRescaleName = colliderRescaleName + "UIA";
                colliderName = colliderName + "UIA";
            }

            Transform collider = atom.gameObject.transform.Find(limbHierarchy + "/" + colliderRescaleName + "/" + colliderName);
            if (collider != null) {
                collider.parent = null;
                GameObject.Destroy(collider.gameObject);
            }

            Transform colliderRescale = atom.gameObject.transform.Find(limbHierarchy + "/" + colliderRescaleName);
            if (colliderRescale != null)
            {
                colliderRescale.parent = null;
                GameObject.Destroy(collider.gameObject);
            }
        }

        public void FootPoseReset(Atom atom)
        {
            ResetAtomHeightOffset(atom);
            Transform rt = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rToe/rShoeToeColliderRescaleUIA");
            if (rt != null)
            {
                if (_savedRotationStatesDict.ContainsKey(atom.name)) _savedRotationStatesDict[atom.name]._timeSaved = Time.unscaledTime;
                else
                {
                    ToeFootRotationStates savedRotationStates = GetCurrentRotationStates(atom);
                    _savedRotationStatesDict.Add(atom.name, savedRotationStates);
                }
                ResetRotationStates(atom, null);
                //reset foot and toe rotations to their defaults
                Vector3 rfootBaseRotation = new Vector3(-15f, 0f, 0f);
                Vector3 rtoeBaseRotation = new Vector3(0f, 0f, 0f);
                Vector3 lfootBaseRotation = new Vector3(-15f, 0f, 0f);
                Vector3 ltoeBaseRotation = new Vector3(0f, 0f, 0f);

                var rfoot = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot");
                var rtoe = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rToe");
                var lfoot = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot");
                var ltoe = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot/lToe");

                rfoot.GetComponent<DAZBone>().baseJointRotation = rfootBaseRotation;
                rtoe.GetComponent<DAZBone>().baseJointRotation = rtoeBaseRotation;
                lfoot.GetComponent<DAZBone>().baseJointRotation = lfootBaseRotation;
                ltoe.GetComponent<DAZBone>().baseJointRotation = ltoeBaseRotation;

                RemoveCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot", "rShoeHeel");
                RemoveCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot", "lShoeHeel");
                RemoveCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rToe", "rShoeToe");
                RemoveCollider(atom, "rescale2/PhysicsModel/Genesis2Female/hip/pelvis/lThigh/lShin/lFoot/lToe", "lShoeToe");
            }
            if (_atomHASettings.ContainsKey(atom.name)) _atomHASettings.Remove(atom.name);
        }
        public bool isPrestigitisHALoaded(Atom atom)
        {
            bool foundHeelAdjustPlugin = false;
            foreach (string storableName in atom.GetStorableIDs())
            {
                if (storableName.StartsWith("plugin#"))
                {
                    string pluginType = storableName.Substring(storableName.IndexOf('_') + 1);
                    if (pluginType.StartsWith("prestigitis.HeelAdjust"))
                    {
                        foundHeelAdjustPlugin = true;
                        break;
                    }
                }
            }

            return (foundHeelAdjustPlugin);
        }
        public bool isHAFolderPresent()
        {
            bool folderExists = FileManagerSecure.DirectoryExists(SuperController.singleton.savesDir + Constants._HeelAdjustSubfolderName);
            return (folderExists);
        }
        public void LoadFootPosesForClothing(Atom atom, List<string> activePersonClothing)
        {
            Transform rt = atom.gameObject.transform.Find("rescale2/PhysicsModel/Genesis2Female/hip/pelvis/rThigh/rShin/rFoot/rToe/rShoeToeColliderRescale");
            if (rt != null && isPrestigitisHALoaded(atom)) return;
            if (!isHAFolderPresent()) return;

            //get intersection of active clothing and files
            List<string> loadableProfiles = activePersonClothing.Intersect(_existingHAFiles).ToList();
            if (loadableProfiles.Count < 1) //intersection list is empty
            {
                FootPoseReset(atom);
            }
            else
            {
                //a heeladjust file was found for one of the active clothing items, load the first one
                ApplyHA(atom, SuperController.singleton.savesDir + "/" + Constants._HeelAdjustSubfolderName + "/" + loadableProfiles.First().ToString() + "." + Constants._HeelAdjustFileExtension);
            }
        }
        public void AtomNameUpdate(string oldName, string newName)
        {
            if (_savedRotationStatesDict.ContainsKey(oldName))
            {
                _savedRotationStatesDict.Add(newName, _savedRotationStatesDict[oldName]);
                _savedRotationStatesDict.Remove(oldName);
            }
            if (_atomHASettings.ContainsKey(oldName))
            {
                _atomHASettings.Add(newName, _atomHASettings[oldName]);
                _atomHASettings.Remove(oldName);
            }
        }
        public JSONClass GetJSON(JSONClass jc)
        {
            if (_atomHASettings.Count > 0)
            {
                JSONArray atomHASettingsJSON = new JSONArray();
                foreach (KeyValuePair<string,HASetting> haSetting in _atomHASettings)
                {
                    atomHASettingsJSON.Add(haSetting.Value.GetJSON(haSetting.Key));
                }
                jc["HeelAdjustSettings"] = atomHASettingsJSON;
            }
           return (jc);
        }

        public void RestoreJSON(JSONClass jc)
        {
            //Assumption is that a scene is loading
            if (_mvrScript._isPrimaryUIAJSON.val)
            {
                //Refresh HASettings dictionary with HASettings stored in Scene save file
                _atomHASettings = new Dictionary<string, HASetting>();
                JSONArray atomHASettingsJSON = jc["HeelAdjustSettings"].AsArray;

                if (atomHASettingsJSON != null && atomHASettingsJSON.Count > 0)
                {
                    for (int i = 0; i < atomHASettingsJSON.Count; i++)
                    {
                        HASetting atomHASetting = new HASetting();
                        JSONClass haSettingJSON = atomHASettingsJSON[i].AsObject;
                        atomHASetting.RestoreJSON(haSettingJSON);
                        _atomHASettings.Add(haSettingJSON["atomName"], atomHASetting);
                    }
                }
                // Handle HeelAdjust clothing items that were not height adjusted when the scene was created
                UpdateExistingHAFiles();
                
                foreach (Atom atom in SuperController.singleton.GetAtoms())
                {
                    if (atom.type == "Person" && !_atomHASettings.ContainsKey(atom.name))
                    {
                        // If the scene contains heel adjusted files but was not saved when UIAssis was active then we create an HASetting entry but dont apply it (assumption is that scene was saved with the person atom at correct heel height already)
                        ActiveClothingLists activeClothingLists = new ActiveClothingLists();
                        activeClothingLists.UpdateActiveClothing(atom);
                        List<string> loadableProfiles = activeClothingLists._activeHAClothingItems.Intersect(_existingHAFiles).ToList();
                        if (loadableProfiles.Count > 0)
                        {
                            HASetting haSetting = new HASetting();
                            LoadHASetting(haSetting, SuperController.singleton.savesDir + "/" + Constants._HeelAdjustSubfolderName + "/" + loadableProfiles.First().ToString() + "." + Constants._HeelAdjustFileExtension);
                            haSetting._controlNodeHeightOffset = haSetting._heelColliderHeight - (0.064f / Constants._ShoeColliderBaseScale);
                            // Heel Adjust raising people is not active, then we dont want to lower people - so we temporarily suppress this until HeelAdjust raising people is activated
                            if (!_heelAdjustActiveJSON.val || !_heelAdjustRaisePeopleJSON.val) haSetting._suppressControlNodeHeightReset = true;
                         
                       
                            _atomHASettings.Add(atom.name, haSetting);
                        }
                    }
                }

                if (!_heelAdjustActiveJSON.val)
                {
                    foreach (Atom atom in SuperController.singleton.GetAtoms())
                    {
                        if (atom.type == "Person")
                        {
                            FootPoseReset(atom);
                        }
                    }
                }
                else if (!_heelAdjustRaisePeopleJSON.val)
                {
                    foreach (Atom atom in SuperController.singleton.GetAtoms())
                    {
                        if (atom.type == "Person")
                        {
                            ResetAtomHeightOffset(atom);
                        }
                    }
                }
            }

        }


        private UIAssist _mvrScript;
        public JSONStorableBool _heelAdjustActiveJSON;
        public JSONStorableBool _heelAdjustRaisePeopleJSON;
        
        private List<string> _existingHAFiles;
        private Dictionary<string, ToeFootRotationStates> _savedRotationStatesDict = new Dictionary<string, ToeFootRotationStates>();
        private Dictionary<string, HASetting> _atomHASettings = new Dictionary<string, HASetting>();
        private Vector3 _ToeColliderBaseRotationR = new Vector3(-16.7f, 0f, -4.3f);
        private Vector3 _ToeColliderBaseRotationL = new Vector3(-16.7f, 0f, 4.3f);
        private Vector3 _HeelColliderBaseRotationR = new Vector3(-20f, 0f, 0f); //rFoot world rotation: Vector3(13.1f, 9.2f, 1.3f);
        private Vector3 _HeelColliderBaseRotationL = new Vector3(-20f, 0f, 0f); //lFoot world rotation: Vector3(13.1f, 351.4f, 359.4f);
    }
    public class HASetting
    {
        public Vector3 _footRotation;
        public Vector3 _toeRotation;
        public float _heelColliderHeight;
        public float _toeColliderHeight;
        public float _heelColliderBackOffset;
        public float _controlNodeHeightOffset =0f;
        public bool _suppressControlNodeHeightReset = false;

        public JSONClass GetJSON(string atomName)
        {
            JSONClass jc = new JSONClass();

            jc["atomName"] = atomName;
            jc["footRotationX"].AsFloat = _footRotation.x;
            jc["footRotationY"].AsFloat = _footRotation.y;
            jc["footRotationZ"].AsFloat = _footRotation.z;

            jc["toeRotationX"].AsFloat = _toeRotation.x;
            jc["toeRotationY"].AsFloat = _toeRotation.y;
            jc["toeRotationZ"].AsFloat = _toeRotation.z;

            jc["heelColliderHeight"].AsFloat = _heelColliderHeight;
            jc["toeColliderHeight"].AsFloat = _toeColliderHeight;
            jc["heelColliderBackOffset"].AsFloat = _heelColliderBackOffset;
            jc["controlNodeHeightOffset"].AsFloat = _controlNodeHeightOffset;
            jc["suppressControlNodeHeightReset"].AsBool = _suppressControlNodeHeightReset;

            return (jc);
        }
        public void RestoreJSON(JSONClass jc)
        {
            _footRotation.x= jc["footRotationX"].AsFloat;
            _footRotation.y = jc["footRotationY"].AsFloat;
            _footRotation.z = jc["footRotationZ"].AsFloat;

            _toeRotation.x = jc["toeRotationX"].AsFloat ;
            _toeRotation.y = jc["toeRotationY"].AsFloat ;
            _toeRotation.z = jc["toeRotationZ"].AsFloat ;

            _heelColliderHeight =jc["heelColliderHeight"].AsFloat;
            _toeColliderHeight= jc["toeColliderHeight"].AsFloat;
            _heelColliderBackOffset =jc["heelColliderBackOffset"].AsFloat;
            _controlNodeHeightOffset = jc["controlNodeHeightOffset"].AsFloat;
            _suppressControlNodeHeightReset = jc["suppressControlNodeHeightReset"].AsBool;
        }
    }

    public class ToeFootRotationStates
    {
        public FreeControllerV3.RotationState _lToeRotationState;
        public FreeControllerV3.RotationState _lFootRotationState;

        public FreeControllerV3.RotationState _rToeRotationState;
        public FreeControllerV3.RotationState _rFootRotationState;
        public float _timeSaved;
    }
    public class WristUI {
        public WristUI(MVRScript plugin) {
            GameObject canvasObject = new GameObject();
            _canvas = canvasObject.AddComponent<Canvas>();
            _canvas.pixelPerfect = false;
            _canvas.renderMode = RenderMode.WorldSpace;
            SuperController.singleton.AddCanvas(_canvas);

            CanvasScaler cs = canvasObject.AddComponent<CanvasScaler>();
            cs.dynamicPixelsPerUnit = 3f;
            cs.scaleFactor = 80.0f;

            GraphicRaycaster rayCaster = canvasObject.AddComponent<GraphicRaycaster>();

            _canvas.transform.localPosition = new Vector3(0.26f, -0.14f, 0.0f);
            _canvas.transform.localScale = new Vector3(0.002f, 0.002f, 0.002f);

            _mvrUIAssist = plugin;
        }

        public UIDynamicButton CreateUIButtonInCanvas(string label, float x, float y, float z = 0, float w = 0, float h = 0) {
            Transform transform = UnityEngine.Object.Instantiate(_mvrUIAssist.manager.configurableButtonPrefab.transform);

            transform.SetParent(_canvas.transform, false);
            transform.gameObject.SetActive(true);

            Font arialFont = (Font)Resources.GetBuiltinResource(typeof(Font), "Arial.ttf");
            UIDynamicButton uiButton = transform.GetComponent<UIDynamicButton>();

            uiButton.label = label;
            uiButton.buttonText.fontSize = 8;
            uiButton.buttonText.font = arialFont;

            if (w > 0) uiButton.button.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, w);
            if (h > 0) uiButton.button.GetComponent<RectTransform>().SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, h);

            transform.Translate(Vector3.down * y, Space.Self);
            transform.Translate(Vector3.right * x, Space.Self);
            transform.Translate(Vector3.forward * z, Space.Self);

            return uiButton;
        }

        public UIDynamicTextField CreateTextField(float x, float y, float z, float width, float height) {
            Transform textField = GameObject.Instantiate<Transform>(_mvrUIAssist.manager.configurableTextFieldPrefab);

            textField.SetParent(_canvas.transform, false);
            RectTransform textFieldRectTransform = textField.GetComponent<RectTransform>();
            textFieldRectTransform.anchoredPosition = new Vector2(width / 2, 0);
            textFieldRectTransform.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
            textFieldRectTransform.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);

            UIDynamicTextField uiTextField = textField.GetComponent<UIDynamicTextField>();

            RectTransform rectT = uiTextField.UItext.GetComponent<RectTransform>();
            rectT.anchoredPosition = new Vector2(0, 0);
            rectT.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width * 0.8f);
            rectT.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height * 0.8f);

            textField.Translate(Vector3.down * y, Space.Self);
            textField.Translate(Vector3.right * x, Space.Self);
            textField.Translate(Vector3.forward * z, Space.Self);

            return uiTextField;
        }

        public void OnDestroy() {
            if (SuperController.singleton != null) SuperController.singleton.RemoveCanvas(_canvas);
            _canvas.transform.SetParent(null, false);

            if (_canvas.gameObject != null) GameObject.Destroy(_canvas.gameObject);
        }

        public void Update(Vector3 vrPosition, Quaternion vrRotation, string uiDisplayMode, bool buttonCanvas = false, bool rightHand = false, float verticalHUDOffset = 0f, float horizontalHUDOffset = 0f, float depthHUDOffset = 0f) {
            if (uiDisplayMode == "HUD" || uiDisplayMode == "VRHUD")
            {
                if (uiDisplayMode == "HUD") depthHUDOffset = 2f;
                Transform cameraT = SuperController.singleton.lookCamera.transform;
                Vector3 facingPos = cameraT.position + cameraT.forward * 10000000.0f;
                Vector3 endPos = Camera.main.ViewportToWorldPoint(new Vector3(horizontalHUDOffset, verticalHUDOffset, depthHUDOffset));

                _canvas.transform.LookAt(facingPos, cameraT.up);
                _canvas.transform.position = endPos;
            }
            else {
                _canvas.transform.position = vrPosition;
                _canvas.transform.rotation = vrRotation;
                if (uiDisplayMode == "WRIST" && buttonCanvas) _canvas.transform.Translate(0f, -0.05f, 0f, Space.Self);
                float inverseMultiplier = 0f;
                if (rightHand) inverseMultiplier = -1f;
                _canvas.transform.Rotate(180 * inverseMultiplier, 90 + (180 * 0), 188 + (196 * inverseMultiplier));
                if (buttonCanvas) {
                    _canvas.transform.Rotate(-50, 0, 0);
                    _canvas.transform.position = _canvas.transform.position + (_canvas.transform.up * 0.02f);
                }
                float inverseOffetMultiplier = 1f;
                if (rightHand) inverseOffetMultiplier = -1f;
                _canvas.transform.position = _canvas.transform.position + (_canvas.transform.up * 0.02f) + (_canvas.transform.forward * -0.045f) + (_canvas.transform.right * (-0.16f * inverseOffetMultiplier));
            }
        }

        public Canvas _canvas;
        MVRScript _mvrUIAssist;
    }

    static class Constants {
		public const bool _patreonContentEnabled = false;
        public const float _TimeBetweenClothingUpdates = 0.1f;
        public const float _TimeBetweenHAFileUpdates = 1f;
        public const string _HeelAdjustSubfolderName = "HeelAdjust";
        public const string _HeelAdjustFileExtension = "heeladjust";
        public static string _vamFilePath = Application.dataPath.Substring(0, Application.dataPath.Length - 8).Replace('/', '\\');
        public const float _ShoeColliderBaseScale = 0.01f;
        public const float _FCV3ResetDelay = 0.8f;
        
    }
	
	static class PatreonFeatures {
		public static void GridLayoutFeature (List<string> buttonGridLayoutChoices){
			buttonGridLayoutChoices =null;
		}
		public static void ClothingPresetActions (UIAssist mvrScript, List<string> mergeClothingGeometryIDs,Atom atom, string mode, string presetFile, bool reverse){
           
		}

		public static void PopulateMergeClothingGeometryIDs (UIAssist mvrScript,List<string> mergeClothingGeometryIDs, string presetFile, Atom atom){
			
		}

        public static void MergePluginStorables (JSONClass atomJSON, JSONArray newPluginStorables, int pluginID)
        {

        }
        public static void MergePluginSettings(Dictionary<string, JSONClass> pluginSettings, JSONArray newPluginStorables, int pluginID)
        {

        }

        public static void RemoveTagClothingItems(Atom atom, Dictionary<string, List<string>> activeClothingTags, List<String> tagsToRemove)
        {


        }
        public static string AppPresetSuppressLoad(UIAssist mvrScript, Atom atom, string presetFilePath, bool suppressClothingLoad, bool suppressPersonScaleLoad){
			return(null);
		}

	}
}
